// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************

#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <assert.h>

#include <getopt.h>

#include <fstream>
#include <list>
#include <sstream>
#include <string>
#include <tuple>

#include "legacy/ExpManager.h"
#include "aevol_version.h"
#include "parameters/ParamLoader.h"
#include "parameters/ParamReader.h"

using namespace aevol;

// Helper functions
void print_help(char* prog_path);
auto interpret_cmd_line_options(int argc, char* argv[]) -> std::tuple<std::string, std::string>;
auto read_sequence_file(const std::string& chromosome_file_name) -> std::list<std::string>;

int main(int argc, char* argv[]) {
  const auto [param_file_name, chromosome_file_name] = interpret_cmd_line_options(argc, argv);

  // Print the hash and date of the commit used to compile this executable
  printf("Running aevol version %s\n", version_string);

  // Read the parameter file
  ParamValues param_values = ParamReader::read_file(param_file_name.c_str());

  // Initialize the experiment manager
  exp_manager = new ExpManager();

  // Initialize the simulation from the parameter file (and the provided sequences if any)
  if (chromosome_file_name.length() != 0) {
    auto chromosomes = read_sequence_file(chromosome_file_name);
    std::cout << "Loading chromosome from text file " << chromosome_file_name << " (" << chromosomes.front().length()
              << " base pairs)" << std::endl;
    ParamLoader::load(param_values, exp_manager, chromosomes, true);
  }
  else {
    ParamLoader::load(param_values, exp_manager, true);
  }

  // Save the experiment
  exp_manager->Save(true);

  delete exp_manager;
}

void print_help(char* prog_path) {
  // Get the program file-name in prog_name (strip prog_path of the path)
  char* prog_name; // No new, it will point to somewhere inside prog_path
  if ((prog_name = strrchr(prog_path, '/'))) {
    prog_name++;
  }
  else {
    prog_name = prog_path;
  }

  printf("******************************************************************************\n");
  printf("*                                                                            *\n");
  printf("*                        aevol - Artificial Evolution                        *\n");
  printf("*                                                                            *\n");
  printf("* Aevol is a simulation platform that allows one to let populations of       *\n");
  printf("* digital organisms evolve in different conditions and study experimentally  *\n");
  printf("* the mechanisms responsible for the structuration of the genome and the     *\n");
  printf("* transcriptome.                                                             *\n");
  printf("*                                                                            *\n");
  printf("******************************************************************************\n");
  printf("\n");
  printf("%s: create an experiment with setup as specified in PARAM_FILE.\n",
         prog_name);
  printf("\n");
  printf("Usage : %s -h or --help\n", prog_name);
  printf("   or : %s -V or --version\n", prog_name);
  printf("   or : %s [-f PARAM_FILE] [-C CHROM_FILE]\n",
         prog_name);
  printf("\nOptions\n");
  printf("  -h, --help\n\tprint this help, then exit\n");
  printf("  -V, --version\n\tprint version number, then exit\n");
  printf("  -f, --file PARAM_FILE\n");
  printf("\tspecify parameter file (default: param.in)\n");
  printf("  -C, --chromosome CHROM_FILE\n");
  printf("\tload chromosome from given text file instead of generating it\n");
}

auto interpret_cmd_line_options(int argc, char* argv[]) -> std::tuple<std::string, std::string> {
  // Command-line option variables
  std::string param_file_name = "param.in";
  std::string chromosome_file_name;

  // Define allowed options
  const char* options_list = "hVf:C:r:";
  static struct option long_options_list[] = {
      {"help",       no_argument,       nullptr, 'h'},
      {"version",    no_argument,       nullptr, 'V'},
      {"file",       required_argument, nullptr, 'f'},
      {"chromosome", required_argument, nullptr, 'C'},
      {0, 0, 0, 0}
  };

  // Get actual values of the CLI options
  int option;
  while ((option = getopt_long(argc, argv, options_list, long_options_list,
                               nullptr)) != -1) {
    switch (option) {
      case 'h' : {
        print_help(argv[0]);
        exit(EXIT_SUCCESS);
      }
      case 'V' : {
        Utils::PrintAevolVersion();
        exit(EXIT_SUCCESS);
      }
      case 'f' : {
        param_file_name = optarg;
        break;
      }
      case 'C': {
        chromosome_file_name = optarg;
        break;
      }
      default : {
        printf("error %s %d\n",optarg,option);
        // An error message is printed in getopt_long, we just need to exit
        exit(EXIT_FAILURE);
      }
    }
  }

  return std::make_tuple(param_file_name, chromosome_file_name);
}

auto read_sequence(std::ifstream& chromosome_file, const std::string& chromosome_file_name) {
  std::string chromosome;

  std::getline(chromosome_file, chromosome);
  if (not chromosome_file or chromosome.length() == 0) {
    Utils::ExitWithUsrMsg(std::string("failed to read from chromosome file ") + chromosome_file_name);
  }

  return chromosome;
}

auto read_sequence_file(const std::string& chromosome_file_name) -> std::list<std::string> {
  assert(chromosome_file_name.length() != 0);

  std::list<std::string> chromosomes;

  std::ifstream chromosome_file(chromosome_file_name);
  if (not chromosome_file) {
    Utils::ExitWithUsrMsg(std::string("failed to open source chromosome file ") + chromosome_file_name);
  }

  chromosomes.push_back(read_sequence(chromosome_file, chromosome_file_name));

  return chromosomes;
}
