// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************

// ============================================================================
//                                   Includes
// ============================================================================
#include <cerrno>
#include <cinttypes>
#include <cstdlib>
#include <cstdio>
#include <csignal>

#include <getopt.h>

#include <charconv>
#include <filesystem>
#include <string>

#ifdef _OPENMP
#include <omp.h>
#else
typedef int omp_int_t;
inline omp_int_t omp_get_thread_num() { return 0;}
inline omp_int_t omp_get_max_threads() { return 1;}
#endif

// =================================================================
//                            Project Files
// =================================================================
#include "legacy/ExpManager.h"

#include "ExpManager_7.h"
#include "macros.h"

#include "aevol_version.h"

using namespace aevol;

// Helper functions
void print_help(char* prog_path);
void interpret_cmd_line_options(int argc, char* argv[]);

// Command-line option variables
// static bool pause_on_startup = false;
static bool verbose = false;
static int64_t t0 = -1;
static int64_t t_end = -1;
static int64_t nb_steps = -1;
static int grain_size = 1;
static bool w_mrca = false;
static std::filesystem::path ui_output_dir = "ui_output_dir";
static uint32_t ui_output_frequency = 100;

#ifdef _OPENMP
static bool run_in_parallel = false;
#endif


int main(int argc, char* argv[]) {
  interpret_cmd_line_options(argc, argv);

  // Print the hash and date of the commit used to compile this executable
  printf("Running aevol version %s\n", version_string);

  // =================================================================
  //                          Load the simulation
  // =================================================================
  exp_manager = new ExpManager();

  exp_manager->load(t0, true);
  exp_manager->set_t_end(t_end);

  auto& exp_manager_7 = *(exp_manager->exp_m_7_);
  exp_manager_7.setup_ui_output(ui_output_dir, ui_output_frequency);

  // Make a numbered copy of each static input file
  // (dynamic files are saved elsewhere)
  // TODO (?)

  // =================================================================
  //                         Run the simulation
  // =================================================================
  exp_manager->run_evolution();

  delete exp_manager;
  //~ return EXIT_SUCCESS;
}


/*!
  \brief

*/
void print_help(char* prog_path) {
  // Get the program file-name in prog_name (strip prog_path of the path)
  char* prog_name; // No new, it will point to somewhere inside prog_path
  if ((prog_name = strrchr(prog_path, '/'))) {
    prog_name++;
  }
  else {
    prog_name = prog_path;
  }

	printf("******************************************************************************\n");
	printf("*                                                                            *\n");
	printf("*                        aevol - Artificial Evolution                        *\n");
	printf("*                                                                            *\n");
	printf("* Aevol is a simulation platform that allows one to let populations of       *\n");
  printf("* digital organisms evolve in different conditions and study experimentally  *\n");
  printf("* the mechanisms responsible for the structuration of the genome and the     *\n");
  printf("* transcriptome.                                                             *\n");
	printf("*                                                                            *\n");
	printf("******************************************************************************\n");
  printf("\n");
	printf("%s: run an aevol simulation.\n", prog_name);
  printf("\n");
	printf("Usage : %s -h or --help\n", prog_name);
	printf("   or : %s -V or --version\n", prog_name);
	printf("   or : %s [-b TIMESTEP] [-e TIMESTEP|-n NB_TIMESTEPS] [-p NB_THREADS] [-vw]\n",
         prog_name);
	printf("\nOptions\n");
	printf("  -h, --help\n\tprint this help, then exit\n");
	printf("  -V, --version\n\tprint version number, then exit\n");
  printf("  -b, --begin TIMESTEP\n");
  printf("\tspecify time t0 to resume simulation at (default read in last_gener.txt)\n");
  printf("  -e, --end TIMESTEP\n");
  printf("\tspecify time of the end of the simulation\n");
  printf("  -n, --nb-timesteps NB_TIMESTEPS\n");
  printf("\tspecify number of timesteps to be simulated (default 1000)\n");
  printf("  -p, --parallel NB_THREADS\n");
  printf("\trun on NB_THREADS threads (use -1 for system default)\n");
  printf("  -v, --verbose\n\tbe verbose\n");
  printf("  -w, --wait\n\tpause after loading\n");
  printf("  --ui-output-dir UI_OUTDIR\n\tdirectory in which to output data for the UI\n");
  printf("  --ui-output-frequency NB_GENER\n\tfrequency at which to output data for the UI\n");
}

void interpret_cmd_line_options(int argc, char* argv[]) {
  // Define allowed options
  const char* options_list = "hVb:e:n:vwxp:";
  int option_index = 0;
  enum long_opt_codes {
    UI_OUTPUT_DIR = 1000,
    UI_OUTPUT_FREQUENCY
  };
  static struct option long_options_list[] = {
      {"help",          no_argument,       nullptr, 'h'},
      {"version",       no_argument,       nullptr, 'V'},
      {"begin",         required_argument, nullptr, 'b'},
      {"end",           required_argument, nullptr, 'e'},
      {"nb-timesteps",  required_argument, nullptr, 'n'},
      {"verbose",       no_argument,       nullptr, 'v'},
      {"wait",          no_argument,       nullptr, 'w'},
      {"noX",           no_argument,       nullptr, 'x'},
      {"parallel",      required_argument, nullptr, 'p'},
      // long-only options
      {"ui-output-dir",       required_argument, nullptr, UI_OUTPUT_DIR},
      {"ui-output-frequency", required_argument, nullptr, UI_OUTPUT_FREQUENCY},
      {0, 0, 0, 0}
  };

  // Get actual values of the CLI options
  int option;
  while ((option = getopt_long(argc, argv, options_list, long_options_list, &option_index)) != -1) {
    switch (option) {
      case 'h' : {
        print_help(argv[0]);
        exit(EXIT_SUCCESS);
      }
      case 'V' : {
        Utils::PrintAevolVersion();
        exit(EXIT_SUCCESS);
      }
      case 'b' : {
        t0 = atol(optarg);
        break;
      }
      case 'e' : {
        if (nb_steps != -1) {
          Utils::ExitWithUsrMsg("use either option -n or -e, not both");
        }

        t_end = atol(optarg);
        w_mrca = true;
        break;
      }
      case 'n' : {
        if (t_end != -1) {
          Utils::ExitWithUsrMsg("use either option -n or -e, not both");
        }

        nb_steps = atol(optarg);
        break;
      }
      case 'v' : {
        verbose = true;
        break;
      }
      case 'w' : {
        // pause_on_startup = true;
        break;
      }
      case 'p' : {
        #ifdef _OPENMP
        run_in_parallel = true;
          if (atoi(optarg) > 0) {
            omp_set_num_threads(atoi(optarg));
          }
        #endif
        break;
      }
      case 'g' : {
        grain_size = atoi(optarg);
        break;
      }
      // Handle long-only options
      case UI_OUTPUT_DIR: {
        ui_output_dir = optarg;
        break;
      }
      case UI_OUTPUT_FREQUENCY: {
        auto [ptr, ec] = std::from_chars(optarg, optarg + strlen(optarg), ui_output_frequency);
        std::cout << int(ec) << std::endl;
        if (ec != std::errc()) {
          Utils::ExitWithUsrMsg(std::string("invalid value for option ") + long_options_list[option_index].name);
        }
        break;
      }
      default : {
        // An error message is printed in getopt_long, we just need to exit
        exit(EXIT_FAILURE);
      }
    }
  }

  // If t0 wasn't provided, use default
  if (t0 < 0) {
    t0 = OutputManager::last_gener();
  }

  // If t_end_ wasn't provided, set it according to nb_steps or use default
  // (run for 1000 timesteps)
  if (t_end < 0) {
    if (nb_steps >= 0) {
      t_end = t0 + nb_steps;
    }
    else {
      t_end = t0 + 1000;
    }
  }

  // It the user didn't ask for a parallel run, set number of threads to 1
  #ifdef _OPENMP
  if (not run_in_parallel) {
      omp_set_num_threads(1);
    }
  #endif
}
