// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************

#include "AnnotatedChromosome.h"

#include <cstring>

#include "List_Metadata.h"


namespace aevol {

AnnotatedChromosome::AnnotatedChromosome(const char* seq, Dna_7::size_type len) {
  metadata_ = new List_Metadata(this);
  dna_ = Dna_7::make_from_sequence(seq, len);
  dna_->set_chrsm(this);
}

AnnotatedChromosome::AnnotatedChromosome() {
  metadata_ = new List_Metadata(this);
}

AnnotatedChromosome* AnnotatedChromosome::make_empty() {
  return new AnnotatedChromosome();
}

AnnotatedChromosome* AnnotatedChromosome::make_from_sequence(const char* seq, Dna_7::size_type len) {
  return new AnnotatedChromosome(seq, len);
}

AnnotatedChromosome* AnnotatedChromosome::make_clone(AnnotatedChromosome* orig,
                                                     Individual_7* owner) {
  AnnotatedChromosome* annotated_chromosome = new AnnotatedChromosome();

  annotated_chromosome->dna_ = Dna_7::make_dna(orig->dna_->length());
  annotated_chromosome->dna_->set_chrsm(annotated_chromosome);
  annotated_chromosome->dna_->set_indiv(orig->dna_, owner);

  for (auto& strand: {Strand::LEADING, Strand::LAGGING}) {
    for (auto& rna: orig->promoter_list(strand)) {
      annotated_chromosome->promoter_list(strand).emplace_back(rna);
      annotated_chromosome->promoter_list(strand).back().to_compute_end_ = false;
    }
  }

  return annotated_chromosome;
}

AnnotatedChromosome::~AnnotatedChromosome() {
  Dna_7::release(std::move(dna_));
  delete metadata_;
}

PromoterList& AnnotatedChromosome::promoter_list() {
  return promoter_list_;
}

PromoterList::SingleStranded& AnnotatedChromosome::promoter_list(Strand strand) {
  return promoter_list_[strand];
}

const std::list<Rna_7*>& AnnotatedChromosome::rnas() const {
  return rna_list_.rnas();
}

const std::list<Protein_7*>& AnnotatedChromosome::proteins() const {
  return proteins_.proteins();
}

void AnnotatedChromosome::reset_stats() {
  nb_genes_activ     = 0;
  nb_genes_inhib     = 0;
  nb_func_genes      = 0;
  nb_non_func_genes  = 0;
  nb_coding_RNAs     = 0;
  nb_non_coding_RNAs = 0;

  nb_bases_in_0_CDS_ = 0;
  nb_bases_in_0_functional_CDS_ = 0;
  nb_bases_in_0_non_functional_CDS_ = 0;
  nb_bases_in_0_RNA_ = 0;
  nb_bases_in_0_coding_RNA_ = 0;
  nb_bases_in_0_non_coding_RNA_ = 0;
  nb_bases_non_essential_ = 0;
  nb_bases_non_essential_including_nf_genes_ = 0;
  nb_bases_in_neutral_regions_ = 0;
};

void AnnotatedChromosome::reset_metadata() {
    delete metadata_;
    metadata_ = new List_Metadata(this);
    for (auto& strand: {Strand::LEADING, Strand::LAGGING}) {
      promoter_list(strand).clear();
    }
}

void AnnotatedChromosome::compute_non_coding() {
  if (non_coding_computed_) return;
  non_coding_computed_ = true;

  // printf("Compute non coding\n");

  // Create a table of <genome_length> bools initialized to false (non-coding)
  int32_t genome_length = dna_->length_;

  // Including Shine-Dalgarno, spacer, START and STOP
  bool* belongs_to_CDS;
  bool* belongs_to_functional_CDS;
  bool* belongs_to_non_functional_CDS; // non-functional CDSs are those that have a null area or that lack a kind of codons (M, W or H)

  // Including Promoters and terminators
  bool* belongs_to_RNA;
  bool* belongs_to_coding_RNA;
  bool* belongs_to_non_coding_RNA;

  // Genes + prom + term (but not UTRs)
  bool* is_essential_DNA;
  bool* is_essential_DNA_including_nf_genes; // Adds non-functional genes + promoters & terminators

  bool* is_not_neutral;            // prom + term + everything in between (as opposed to neutral)


  belongs_to_CDS = new bool[genome_length];
  belongs_to_functional_CDS = new bool[genome_length];
  belongs_to_non_functional_CDS = new bool[genome_length];
  belongs_to_RNA = new bool[genome_length];
  belongs_to_coding_RNA = new bool[genome_length];
  belongs_to_non_coding_RNA = new bool[genome_length];
  is_essential_DNA = new bool[genome_length];
  is_essential_DNA_including_nf_genes = new bool[genome_length];
  is_not_neutral = new bool[genome_length];

  memset(belongs_to_CDS, 0, genome_length);
  memset(belongs_to_functional_CDS, 0, genome_length);
  memset(belongs_to_non_functional_CDS, 0, genome_length);
  memset(belongs_to_RNA, 0, genome_length);
  memset(belongs_to_coding_RNA, 0, genome_length);
  memset(belongs_to_non_coding_RNA, 0, genome_length);
  memset(is_essential_DNA, 0, genome_length);
  memset(is_essential_DNA_including_nf_genes, 0, genome_length);
  memset(is_not_neutral, 0, genome_length);


  // Parse protein lists and mark the corresponding bases as coding
  for (const auto& prot : proteins()) {
    #ifdef __REGUL
    if (!prot->signal_ && prot->is_init_) {
    #else
    if (prot->is_init_) {
    #endif
      int32_t first;
      int32_t last;

      

      switch (prot->strand_) {
        case Strand::LEADING:
        #ifdef __EUKARYOTE
          first = prot->protein_start-PROT_START_SIZE;
        #else
          first = Utils::mod(prot->protein_start-PROT_START_SIZE,genome_length);
        #endif
          last = prot->protein_end;
          break;
        case Strand::LAGGING:
        #ifdef __EUKARYOTE
          last = prot->protein_start+PROT_START_SIZE;
        #else
          last = Utils::mod(prot->protein_start+PROT_START_SIZE,genome_length);
        #endif
          first = prot->protein_end;
          break;
        default:
          assert(false); // error: should never happen
      }

      if (first <= last) {
        for (int32_t i = first; i <= last; i++) {
          belongs_to_CDS[i] = true;
          if (prot->is_functional) is_essential_DNA[i] = true;
          is_essential_DNA_including_nf_genes[i] = true;
        }
      }
      else {
        for (int32_t i = first; i < genome_length; i++) {
          belongs_to_CDS[i] = true;
          if (prot->is_functional) is_essential_DNA[i] = true;
          is_essential_DNA_including_nf_genes[i] = true;
        }
        for (int32_t i = 0; i <= last; i++) {
          belongs_to_CDS[i] = true;
          if (prot->is_functional) is_essential_DNA[i] = true;
          is_essential_DNA_including_nf_genes[i] = true;
        }
      }

      // Include the promoter and terminator to essential DNA
      // Mark everything between promoter and terminator as not neutral
      for (Rna_7* rna : prot->rna_list_) {

        if (rna->is_init_) {
          int32_t prom_first;
          int32_t prom_last;
          int32_t term_first;
          int32_t term_last;
          int32_t rna_first;
          int32_t rna_last;

          if (prot->strand_ == Strand::LEADING) {
            prom_first = rna->begin;
            prom_last = Utils::mod(prom_first + PROM_SIZE - 1, genome_length);
            term_last = rna->end;
            #ifdef __EUKARYOTE
              assert(prom_first + PROM_SIZE - 1 < genome_length && prom_first + PROM_SIZE - 1 >= 0);
              assert(term_last - TERM_SIZE + 1 < genome_length && term_last - TERM_SIZE + 1 >= 0);
            #endif
            term_first = Utils::mod(term_last - TERM_SIZE + 1, genome_length);
            rna_first = prom_first;
            rna_last = term_last;
          }
          else {
            prom_last = rna->begin;
            prom_first = Utils::mod(prom_last - PROM_SIZE + 1, genome_length);
            term_first = rna->end;
            term_last = Utils::mod(term_first + TERM_SIZE - 1, genome_length);
            #ifdef __EUKARYOTE
              assert(prom_first + PROM_SIZE - 1 < genome_length && prom_first + PROM_SIZE - 1 >= 0);
              assert(term_last - TERM_SIZE + 1 < genome_length && term_last - TERM_SIZE + 1 >= 0);
            #endif
            rna_first = term_first;
            rna_last = prom_last;
          }

          // Let us begin with "non-neutral" regions...
          if (rna_first <= rna_last) {
            for (int32_t i = rna_first;
                i <= rna_last; i++) { is_not_neutral[i] = true; }
          }
          else {
            for (int32_t i = rna_first;
                i < genome_length; i++) { is_not_neutral[i] = true; }
            for (int32_t i = 0; i <= rna_last; i++) { is_not_neutral[i] = true; }
          }

          // ...and go on with essential DNA
          if (prom_first <= prom_last) {
            for (int32_t i = prom_first; i <= prom_last; i++) {
              //~ printf("%ld ", i);
              if (prot->is_functional) is_essential_DNA[i] = true;
              is_essential_DNA_including_nf_genes[i] = true;
            }
          }
          else {
            for (int32_t i = prom_first; i < genome_length; i++) {
              //~ printf("%ld ", i);
              if (prot->is_functional) is_essential_DNA[i] = true;
              is_essential_DNA_including_nf_genes[i] = true;
            }
            for (int32_t i = 0; i <= prom_last; i++) {
              //~ printf("%ld ", i);
              if (prot->is_functional) is_essential_DNA[i] = true;
              is_essential_DNA_including_nf_genes[i] = true;
            }
          }
          //~ printf("\n");

          //~ printf("term ");
          if (term_first <= term_last) {
            for (int32_t i = term_first; i <= term_last; i++) {
              //~ printf("%ld ", i);
              if (prot->is_functional) is_essential_DNA[i] = true;
              is_essential_DNA_including_nf_genes[i] = true;
            }
          }
          else {
            for (int32_t i = term_first; i < genome_length; i++) {
              //~ printf("%ld ", i);
              if (prot->is_functional) is_essential_DNA[i] = true;
              is_essential_DNA_including_nf_genes[i] = true;
            }
            for (int32_t i = 0; i <= term_last; i++) {
              //~ printf("%ld ", i);
              if (prot->is_functional) is_essential_DNA[i] = true;
              is_essential_DNA_including_nf_genes[i] = true;
            }
          }
        }
        //~ printf("\n");
        //~ getchar();
      }


      if (prot->is_functional) {
        if (first <= last) {
          for (int32_t i = first; i <= last; i++) {
            belongs_to_functional_CDS[i] = true;
          }
        }
        else {
          for (int32_t i = first; i < genome_length; i++) {
            belongs_to_functional_CDS[i] = true;
          }
          for (int32_t i = 0; i <= last; i++) {
            belongs_to_functional_CDS[i] = true;
          }
        }
      }
      else // degenerated protein
      {
        if (first <= last) {
          for (int32_t i = first; i <= last; i++) {
            belongs_to_non_functional_CDS[i] = true;
          }
        }
        else {
          for (int32_t i = first; i < genome_length; i++) {
            belongs_to_non_functional_CDS[i] = true;
          }
          for (int32_t i = 0; i <= last; i++) {
            belongs_to_non_functional_CDS[i] = true;
          }
        }
      }
    }
  }


  // Parse RNA lists and mark the corresponding bases as coding (only for the coding RNAs)
  for (Rna_7* rna : rnas()) {
      if (rna->is_init_) {
        int32_t first;
        int32_t last;

        if (rna->strand_ == Strand::LEADING) {
          first = rna->begin;
          last = rna->end;
        }
        else { // (strand == LAGGING)
          first = rna->end;
          last = rna->begin;
        }

        assert(first < genome_length);
        assert(last < genome_length);
        
        if (first <= last) {
          for (int32_t i = first; i <= last; i++) {
            belongs_to_RNA[i] = true;
          }
        }
        else {
          for (int32_t i = first; i < genome_length; i++)
            belongs_to_RNA[i] = true;
          for (int32_t i = 0; i <= last; i++)
            belongs_to_RNA[i] = true;
        }

        bool is_coding_rna = false;
        for (const auto& prot : proteins()) {
          if (prot->is_init_) {
            for (auto i_rna : prot->rna_list_) {
              if (i_rna->begin == rna->begin && i_rna->end == rna->end && i_rna->strand_ == rna->strand_) {
                is_coding_rna = true;
              }

            }
          }
        }


        if (is_coding_rna) { // coding RNA
          if (first <= last) {
            for (int32_t i = first; i <= last; i++)
              belongs_to_coding_RNA[i] = true;
          }
          else {
            for (int32_t i = first; i < genome_length; i++)
              belongs_to_coding_RNA[i] = true;
            for (int32_t i = 0; i <= last; i++)
              belongs_to_coding_RNA[i] = true;
          }
        }
        else // non coding RNA
        {
          if (first <= last) {
            for (int32_t i = first; i <= last; i++)
              belongs_to_non_coding_RNA[i] = true;
          }
          else {
            for (int32_t i = first; i < genome_length; i++)
              belongs_to_non_coding_RNA[i] = true;
            for (int32_t i = 0; i <= last; i++)
              belongs_to_non_coding_RNA[i] = true;
          }
        }
      }
    }
  

  // Count non-coding bases
  nb_bases_in_0_CDS_ = 0;
  nb_bases_in_0_functional_CDS_ = 0;
  nb_bases_in_0_non_functional_CDS_ = 0;
  nb_bases_in_0_RNA_ = 0;
  nb_bases_in_0_coding_RNA_ = 0;
  nb_bases_in_0_non_coding_RNA_ = 0;
  nb_bases_non_essential_ = 0;
  nb_bases_non_essential_including_nf_genes_ = 0;
  nb_bases_in_neutral_regions_ = 0;

  for (int32_t i = 0; i < genome_length; i++) {
    if (belongs_to_CDS[i] == false) {
      nb_bases_in_0_CDS_++;
    }
    if (belongs_to_functional_CDS[i] == false) {
      nb_bases_in_0_functional_CDS_++;
    }
    if (belongs_to_non_functional_CDS[i] == false) {
      nb_bases_in_0_non_functional_CDS_++;
    }
    if (belongs_to_RNA[i] == false) {
      nb_bases_in_0_RNA_++;
    }
    if (belongs_to_coding_RNA[i] == false) {
      nb_bases_in_0_coding_RNA_++;
    }
    if (belongs_to_non_coding_RNA[i] == false) {
      nb_bases_in_0_non_coding_RNA_++;
    }
    if (is_essential_DNA[i] == false) {
      nb_bases_non_essential_++;
    }
    if (is_essential_DNA_including_nf_genes[i] == false) {
      nb_bases_non_essential_including_nf_genes_++;
    }
    if (is_not_neutral[i] == false) {
      nb_bases_in_neutral_regions_++;
    }
  }

  // printf("nb_bases_in_0_CDS_ %d\n",nb_bases_in_0_CDS_);

  delete[] belongs_to_CDS;
  delete[] belongs_to_functional_CDS;
  delete[] belongs_to_non_functional_CDS;
  delete[] belongs_to_RNA;
  delete[] belongs_to_coding_RNA;
  delete[] belongs_to_non_coding_RNA;
  delete[] is_essential_DNA;
  delete[] is_essential_DNA_including_nf_genes;
  delete[] is_not_neutral;
}

}
