// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************


#include "ExpManager_7.h"
#include "7/fuzzy/AbstractFuzzy.h"
#include "7/fuzzy/Discrete_Double_Fuzzy.h"
#include "7/fuzzy/Vector_Fuzzy.h"
#include "7/Stats_7.h"
#include "DnaMutator.h"
#include "legacy/ExpManager.h"
#include "Promoter.h"
#include "Protein_7.h"
#include "Rna_7.h"
#include "List_Metadata.h"
#include "7/fuzzy/FuzzyFactory_7.h"

#include <algorithm>
#include <err.h>
#include <fstream>
#include <sstream>
#include <sys/stat.h>
#include<chrono>
#include <omp.h>

#ifdef __OMP_LIST_SORT
  #if 1 == __OMP_LIST_SORT
    #include <random>       // std::default_random_engine
  #endif
#endif

namespace aevol {

#define __VECTORIZE_STRCMP



void ExpManager_7::load_legacy(ExpManager* exp_m, const ExpSetup* exp_s, const World* world) {
  assert(exp_s == exp_m->exp_s());
  assert(world == exp_m->world());
  printf("  Loading SIMD Controller...");

  exp_m_ = exp_m;

  nb_indivs_ = exp_m_->nb_indivs();

  current_individuals       = new Individual_7*[exp_m_->nb_indivs()];
  previous_individuals      = new Individual_7*[exp_m_->nb_indivs()];

  // Allocate Dna_7
  int max_size_dna = -1;
  for (int indiv_id = 0; indiv_id < exp_m_->nb_indivs(); indiv_id++) {
    int x = indiv_id / world->height();
    int y = indiv_id % world->height();

    max_size_dna = max_size_dna < world->grid(x, y)->individual()->genetic_unit_seq_length(0) ?
                   world->grid(x, y)->individual()->genetic_unit_seq_length(0) : max_size_dna;
  }

  #if 0 == __DNA_FACTORY_ALG
    Dna_7::reset_dna_factory(
        new DnaFactory(DnaFactory_Policy::FIRST, exp_m_->nb_indivs() * 3, max_size_dna, nb_indivs_));
  #elif 1 == __DNA_FACTORY_ALG
    Dna_7::reset_dna_factory(
        new DnaFactory(DnaFactory_Policy::FIRSTFIT, exp_m_->nb_indivs() * 3, max_size_dna, nb_indivs_));
  #elif 2 == __DNA_FACTORY_ALG
    Dna_7::reset_dna_factory(
        new DnaFactory(DnaFactory_Policy::LOCAL_GLOBAL_FIT, exp_m_->nb_indivs() * 3, max_size_dna, nb_indivs_));
  #elif 3 == __DNA_FACTORY_ALG
    Dna_7::reset_dna_factory(
         new DnaFactory(DnaFactory_Policy::ALLOCATE, exp_m_->nb_indivs() * 3, max_size_dna, nb_indivs_));
  #endif

  fuzzy_factory_ = new FuzzyFactory_7(exp_s->get_fuzzy_flavor(),exp_m_->nb_indivs()*4,
                        world->phenotypic_target_handler()->sampling(),nb_indivs_);

  for (int indiv_id = 0; indiv_id < exp_m_->nb_indivs(); indiv_id++) {
    int x = indiv_id / exp_m_->world()->height();
    int y = indiv_id % exp_m_->world()->height();

    current_individuals[indiv_id] = Individual_7::make_from_sequence(
        exp_m_->world()->indiv_at(0, 0)->w_max(),
        fuzzy_factory_,
        world->grid(x, y)->individual()->genetic_unit(0).dna()->data(),
        world->grid(x, y)->individual()->genetic_unit(0).dna()->length());
    current_individuals[indiv_id]->annotated_chromosome_->dna_->set_parent_length(
        world->grid(x, y)->individual()->genetic_unit(0).dna()->length());
    current_individuals[indiv_id]->parent_id = indiv_id;
    previous_individuals[indiv_id] = current_individuals[indiv_id];
  }

  dna_mutator_array_ = new DnaMutator*[exp_m_->grid_height() * exp_m_->grid_width()];
    for (int i = 0; i < exp_m_->grid_height() * exp_m_->grid_width(); i++) {
        dna_mutator_array_[i] = nullptr;
    }

#ifdef __REGUL
  printf("VAR Method %d\n",world->phenotypic_target_handler()->var_method());
  phenotypic_target_handler_ = new SIMD_PhenotypicTargetHandler_R(
        std::dynamic_pointer_cast<PhenotypicTargetHandler_R>(world->phenotypic_target_handler()),
        exp_m->exp_s(),fuzzy_factory_);
#else
  target = fuzzy_factory_->get_fuzzy();
  // copy from legacy fuzzy
  const auto& legacy_fuzzy = world->phenotypic_target_handler()->phenotypic_target().fuzzy();
  switch (fuzzy_factory_->flavor()) {
    case FuzzyFlavor::VECTOR: {
      auto casted_target = static_cast<Vector_Fuzzy*>(target);
      casted_target->points().clear();
      for (auto p: legacy_fuzzy->points())
        casted_target->points().insert(p);
      break;
    }
    case FuzzyFlavor::DISCRETE_DOUBLE_TABLE: {
      auto casted_target = static_cast<Discrete_Double_Fuzzy*>(target);
      double len = static_cast<double>(casted_target->length());
      for (int i = 0; i < casted_target->length(); i++) {
        double tmp = legacy_fuzzy->y(((double)i) / len);
        casted_target->points_[i] = tmp;
      }
      break;
    }
  }
#endif

  #ifdef WITH_PERF_TRACES
    std::ofstream perf_traces_file_;

    perf_traces_file_.open("sm_perf_traces.csv", std::ofstream::trunc);
    perf_traces_file_ << "generation,runtime,io" << std::endl;
    perf_traces_file_.close();
  #endif


  grid_width_ = exp_m_->grid_width();
  grid_height_ = exp_m_->grid_height();

  printf(" OK\n");
  #ifdef WITH_PERF_TRACES_PER_INDIV
  apply_mutation = new long[exp_m_->nb_indivs()];
  compute_rna_ = new long[exp_m_->nb_indivs()];
  start_protein_ = new long[exp_m_->nb_indivs()];
  compute_protein_ = new long[exp_m_->nb_indivs()];
  translate_protein_ = new long[exp_m_->nb_indivs()];
  compute_phenotype_ = new long[exp_m_->nb_indivs()];
  compute_fitness_ = new long[exp_m_->nb_indivs()];
  total_ = new long[exp_m_->nb_indivs()];


  allocate_individual_start_ = new long[exp_m_->nb_indivs()];
  apply_mutation_start_ = new long[exp_m_->nb_indivs()];
  compute_rna_start_ = new long[exp_m_->nb_indivs()];
  start_protein_start_ = new long[exp_m_->nb_indivs()];
  compute_protein_start_ = new long[exp_m_->nb_indivs()];
  translate_protein_start_ = new long[exp_m_->nb_indivs()];
  compute_phenotype_start_ = new long[exp_m_->nb_indivs()];
  compute_fitness_start_ = new long[exp_m_->nb_indivs()];

  allocate_individual_stop_ = new long[exp_m_->nb_indivs()];
  apply_mutation_stop_ = new long[exp_m_->nb_indivs()];
  compute_rna_stop_ = new long[exp_m_->nb_indivs()];
  start_protein_stop_ = new long[exp_m_->nb_indivs()];
  compute_protein_stop_ = new long[exp_m_->nb_indivs()];
  translate_protein_stop_ = new long[exp_m_->nb_indivs()];
  compute_phenotype_stop_ = new long[exp_m_->nb_indivs()];
  compute_fitness_stop_ = new long[exp_m_->nb_indivs()];

  total_start_ = new long[exp_m_->nb_indivs()];
  total_stop_ = new long[exp_m_->nb_indivs()];
  omp_tid_ = new long[exp_m_->nb_indivs()];
  #endif
}

void ExpManager_7::selection(int indiv_id) {
  if (exp_m_->sel()->selection_scope() == SCOPE_LOCAL) {
  int32_t selection_scope_x = exp_m_->sel()->selection_scope_x();
  int32_t selection_scope_y = exp_m_->sel()->selection_scope_y();

  int16_t neighborhood_size = selection_scope_x * selection_scope_y;
  FitnessFunction fitness_function = exp_m_->sel()->fitness_func();

  double *  local_fit_array   = new double[(selection_scope_x) * (selection_scope_y)];
  double *  local_meta_array   = new double[(selection_scope_x) * (selection_scope_y)];
  double *  probs             = new double[(selection_scope_x) * (selection_scope_y)];
  int16_t   count             = 0;
  double    sum_local_fit     = 0.0;

  int32_t x = indiv_id / grid_height_;
  int32_t y = indiv_id % grid_height_;
  int cur_x,cur_y;

#ifdef __REGUL
  double** fitness_sum_local_tab_;
  int32_t fitness_function_scope_x = exp_m_->sel()->fitness_function_scope_x();
  int32_t fitness_function_scope_y = exp_m_->sel()->fitness_function_scope_y();
#endif

  if (fitness_function == FITNESS_LOCAL_SUM) {
#ifdef __REGUL
    fitness_sum_local_tab_ = new double*[fitness_function_scope_x*fitness_function_scope_y];
        for (int tab_id = 0; tab_id < fitness_function_scope_x*fitness_function_scope_y; tab_id++)
          fitness_sum_local_tab_[tab_id] = new double[phenotypic_target_handler_->nb_env_];

        for (int env_id = 0; env_id < phenotypic_target_handler_->nb_env_; env_id++) {

          int tab_id = 0;
          fitness_sum_local_tab_[tab_id][env_id] = 0;

          for (int8_t i = -(selection_scope_x/2) ; i <= (selection_scope_x/2) ; i++) {
            for (int8_t j = -(selection_scope_y/2) ; j <= (selection_scope_y/2) ; j++) {
              cur_x = (x + i + grid_width_) % grid_width_;
              cur_y = (y + j + grid_height_) % grid_height_;

              int16_t new_x,new_y;
              for (int8_t ii = -(fitness_function_scope_x/2); ii <= (fitness_function_scope_x/2); ii++) {
                for (int8_t jj = -(fitness_function_scope_y/2); jj <= (fitness_function_scope_y/2); jj++) {
                  //TODO: Check values HERE !

                  new_x = (cur_x + ii + grid_width_) % grid_width_;
                  new_y = (cur_y + jj + grid_height_) % grid_height_;

                  fitness_sum_local_tab_[tab_id][env_id] +=
                      previous_individuals[new_x * grid_height_ + new_y]->fitness_by_env_id_[env_id];
                }
              }
              tab_id++;
            }
          }
        }
#else
    printf("Fitness local sum is not supported for Aevol (only R-Aevol)\n");
    exit(-1);
#endif
  }

  #ifdef __REGUL
    int tab_id = 0;
  #endif

  for (int8_t i = -(selection_scope_x/2) ; i <= (selection_scope_x/2) ; i++) {
    for (int8_t j = -(selection_scope_y/2) ; j <= (selection_scope_y/2) ; j++) {
      cur_x = (x + i + grid_width_) % grid_width_;
      cur_y = (y + j + grid_height_) % grid_height_;

      if (fitness_function == FITNESS_EXP) {
        local_fit_array[count] = previous_individuals[cur_x * grid_height_ + cur_y]->fitness_;
      } else if (fitness_function == FITNESS_GLOBAL_SUM) {
#ifdef __REGUL
        double composed_fitness = 0;
            for (int env_id = 0; env_id < phenotypic_target_handler_->nb_env_;
                 env_id++) {
              composed_fitness +=
                  previous_individuals[cur_x * grid_height_ + cur_y]->fitness_by_env_id_[env_id] /
                  fitness_sum_tab_[env_id];
            }
            composed_fitness /= phenotypic_target_handler_->nb_env_;
            local_fit_array[count] = composed_fitness;
#else
        printf("Fitness global sum is not supported for Aevol (only R-Aevol)\n");
        exit(-1);
#endif
      } else if (fitness_function == FITNESS_LOCAL_SUM) {
#ifdef __REGUL
        double composed_fitness = 0;
            for (int env_id = 0; env_id < phenotypic_target_handler_->nb_env_;
                 env_id++) {
              composed_fitness +=
                  previous_individuals[cur_x * grid_height_ + cur_y]->fitness_by_env_id_[env_id] /
                  fitness_sum_local_tab_[tab_id][env_id];
            }
            composed_fitness /= phenotypic_target_handler_->nb_env_;
            local_fit_array[count] = composed_fitness;
#else
        printf("Fitness local sum is not supported for Aevol (only R-Aevol)\n");
        exit(-1);
#endif
      }

      sum_local_fit += local_fit_array[count];
      count++;
      #ifdef __REGUL
        tab_id++;
      #endif
    }
  }

  if (fitness_function == FITNESS_LOCAL_SUM) {
#ifdef __REGUL
    for (int tab_id = 0; tab_id < fitness_function_scope_x * fitness_function_scope_y; tab_id++)
          delete[] fitness_sum_local_tab_[tab_id];
        delete[] fitness_sum_local_tab_;
#else
    printf("Fitness local sum is not supported for Aevol (only R-Aevol)\n");
    exit(-1);
#endif
  }

  for(int16_t i = 0 ; i < neighborhood_size ; i++) {
    probs[i] = local_fit_array[i]/sum_local_fit;
  }
  int16_t found_org = exp_m_->world()->grid(x,y)->reprod_prng_simd_->roulette_random(probs, neighborhood_size);

  int32_t x_offset,y_offset;

  x_offset = (found_org / selection_scope_x) - 1;
  y_offset = (found_org % selection_scope_x) - 1;

  delete [] local_fit_array;
  delete [] local_meta_array;
  delete [] probs;

  exp_m_->next_generation_reproducer_[indiv_id] = ((x+x_offset+grid_width_)  % grid_width_)*grid_height_+
                                          ((y+y_offset+grid_height_) % grid_height_);
  }
}

void ExpManager_7::do_mutation(int indiv_id) {
  int32_t dna_length = -1;

  int32_t x = indiv_id / exp_m_->world()->height();
  int32_t y = indiv_id % exp_m_->world()->height();
  delete dna_mutator_array_[indiv_id];

  dna_length = previous_individuals[exp_m_->next_generation_reproducer_[indiv_id]]->annotated_chromosome_->dna_->length();

  dna_mutator_array_[indiv_id] = new DnaMutator(
      exp_m_->world()->grid(x, y)->mut_prng(),
      dna_length,
      exp_m_->exp_s()->mut_params()->duplication_rate(),
      exp_m_->exp_s()->mut_params()->deletion_rate(),
      exp_m_->exp_s()->mut_params()->translocation_rate(),
      exp_m_->exp_s()->mut_params()->inversion_rate(),
      exp_m_->exp_s()->mut_params()->point_mutation_rate(),
      exp_m_->exp_s()->mut_params()->small_insertion_rate(),
      exp_m_->exp_s()->mut_params()->small_deletion_rate(),
      exp_m_->exp_s()->mut_params()->max_indel_size(),
      exp_m_->exp_s()->min_genome_length(),
      exp_m_->exp_s()->max_genome_length(),
      indiv_id,
      x,
      y);
  dna_mutator_array_[indiv_id]->generate_mutations();

  if (dna_mutator_array_[indiv_id]->hasMutate()) {
    dna_mutator_array_[indiv_id]->generate_all_mutations(dna_mutator_array_[indiv_id]->length_);
    
    #pragma omp critical
    {
    mutant_list_.push_back(indiv_id);
    }
  } else {
    int32_t parent_id = exp_m_->next_generation_reproducer_[indiv_id];

    #ifdef __REGUL
    bool first_to_add = false;
    #pragma omp critical
    {
      first_to_add = previous_individuals[parent_id]->first_to_add;
      previous_individuals[parent_id]->first_to_add = false;
    }
    
    if (first_to_add)
        previous_individuals[parent_id]->added_id = indiv_id;

    
    if (first_to_add && phenotypic_target_handler_->hasChanged_) {
      #pragma omp critical
      {
      mutant_list_.push_back(indiv_id);
      }
    }
    #endif

    #pragma omp atomic
    nb_clones_++;

    current_individuals[indiv_id] = previous_individuals[parent_id];


    #pragma omp atomic
    current_individuals[indiv_id]->usage_count_++;


    #pragma omp critical
    {
    current_individuals[indiv_id]->last_id = indiv_id;
    }

    if (exp_m_->record_tree()) {
      int x = indiv_id / exp_m_->world()->height();
      int y = indiv_id % exp_m_->world()->height();
      NewIndivEvent *eindiv = new NewIndivEvent(
            current_individuals[indiv_id],
            previous_individuals[exp_m_->next_generation_reproducer_[indiv_id]],
            x, y,indiv_id,
            exp_m_->next_generation_reproducer_[indiv_id]);
      exp_m_->tree()->update_new_indiv(eindiv);
      delete eindiv;
    }

#ifdef WITH_PERF_TRACES_PER_INDIV
    apply_mutation[indiv_id] = -1;
#endif
  }
}

void ExpManager_7::apply_mutations(int indiv_id) {
#ifdef WITH_PERF_TRACES_PER_INDIV
    allocate_individual_start_[indiv_id] = std::chrono::steady_clock::now().time_since_epoch().count();
#endif

    current_individuals[indiv_id] = Individual_7::make_clone(
        previous_individuals[exp_m_->next_generation_reproducer_[indiv_id]], fuzzy_factory_);

    current_individuals[indiv_id]->last_id = indiv_id;
    current_individuals[indiv_id]->parent_id = exp_m_->next_generation_reproducer_[indiv_id];

    if (exp_m_->record_tree()) {
      int x = indiv_id / exp_m_->world()->height();
      int y = indiv_id % exp_m_->world()->height();

      NewIndivEvent *eindiv = new NewIndivEvent(
            current_individuals[indiv_id],
            previous_individuals[exp_m_->next_generation_reproducer_[indiv_id]],
            x, y,indiv_id,
            exp_m_->next_generation_reproducer_[indiv_id]);

      exp_m_->tree()->update_new_indiv(eindiv);
      delete eindiv;
    }

#ifdef WITH_PERF_TRACES_PER_INDIV
    auto t_start = std::chrono::steady_clock::now();
    allocate_individual_stop_[indiv_id] = t_start.time_since_epoch().count();
    apply_mutation_start_[indiv_id] = t_start.time_since_epoch().count();
#endif

  auto undergone_mutations =
      current_individuals[indiv_id]->annotated_chromosome_->dna_->apply_mutations(dna_mutator_array_[indiv_id]->mutation_list_);

  for (const auto& mut : undergone_mutations) {
    if (exp_manager->record_tree()) {
      exp_manager->tree()->report_by_index(AeTime::time(), indiv_id)->dna_replic_report().add_mut(mut);
    }
    delete mut;
  }

#ifdef WITH_PERF_TRACES_PER_INDIV
    auto t_end = std::chrono::steady_clock::now();
    apply_mutation_stop_[indiv_id] = t_end.time_since_epoch().count();
    apply_mutation[indiv_id] = t_end.time_since_epoch().count() - t_start.time_since_epoch().count();
#endif

}

ExpManager_7::~ExpManager_7() {
  delete stats_best;
  delete stats_mean;

  /* No need to delete current_individuals, at the end of a generation all the
   * element of the table are nullptr
   */
  #pragma omp single
  {
  for (int indiv_id = 0; indiv_id < (int) exp_m_->nb_indivs(); indiv_id++) {
      if (previous_individuals[indiv_id]->usage_count_ > 1)
        previous_individuals[indiv_id]->usage_count_--;
      else {
          delete previous_individuals[indiv_id];
          previous_individuals[indiv_id] = nullptr;
      }

    delete dna_mutator_array_[indiv_id];
  }
  }

  delete[] previous_individuals;
  delete[] current_individuals;


#ifdef __REGUL
  delete phenotypic_target_handler_;
#else
  fuzzy_factory_->give_back(target);
#endif

  delete fuzzy_factory_;
}

void ExpManager_7::start_stop_RNA(int indiv_id) {
  current_individuals[indiv_id]->start_stop_RNA();
}

void ExpManager_7::prom_compute_RNA(int indiv_id) {
  current_individuals[indiv_id]->prom_compute_RNA(*exp_m_->exp_s());
}

void ExpManager_7::start_protein(int indiv_id) {
  current_individuals[indiv_id]->start_protein();
}

void ExpManager_7::compute_protein(int indiv_id) {
  current_individuals[indiv_id]->compute_protein();
}

void ExpManager_7::translate_protein(int indiv_id, double w_max) {
  current_individuals[indiv_id]->translate_protein(*exp_m_->exp_s(), w_max);
}

void ExpManager_7::compute_phenotype(int indiv_id) {
  current_individuals[indiv_id]->compute_phenotype();
}

#ifdef __REGUL
void ExpManager_7::compute_fitness(int indiv_id,
                                   double selection_pressure,
                                   int env_id) {
  current_individuals[indiv_id]->compute_fitness(selection_pressure, env_id, phenotypic_target_handler_);
}
#else
void ExpManager_7::compute_fitness(int indiv_id,
                                   double selection_pressure) {
  current_individuals[indiv_id]->compute_fitness(selection_pressure, target);
}
#endif

#ifdef __REGUL
void ExpManager_7::compute_network(int indiv_id) {
  current_individuals[indiv_id]->compute_network(phenotypic_target_handler_);
}

void ExpManager_7::update_network(int indiv_id) {
  current_individuals[indiv_id]->update_network(*exp_m_->exp_s());
}

void ExpManager_7::evaluate_network(int indiv_id, double selection_pressure, int env_id) {
  current_individuals[indiv_id]->evaluate_network(selection_pressure, env_id, phenotypic_target_handler_);
}

void ExpManager_7::finalize_network(int indiv_id, double selection_pressure) {
  current_individuals[indiv_id]->finalize_network(selection_pressure, phenotypic_target_handler_);
}

void ExpManager_7::solve_network(int indiv_id, double selection_pressure) {
  current_individuals[indiv_id]->solve_network(*exp_m_->exp_s(), selection_pressure, phenotypic_target_handler_);
}

void ExpManager_7::update_phenotype( int indiv_id ) {
  current_individuals[indiv_id]->update_phenotype();
}
#endif

int32_t ExpManager_7::find_best_indiv_idx(Individual_7** indivs, int32_t nb_indivs) const {
  double best_fitness = indivs[0]->fitness_;
  int idx_best = 0;
  for (int indiv_idx = 1 ; indiv_idx < (int) nb_indivs ; ++indiv_idx) {
    if (indivs[indiv_idx]->fitness_ > best_fitness) {
      idx_best = indiv_idx;
      best_fitness = indivs[indiv_idx]->fitness_;
    }
  }
  return idx_best;
}

void ExpManager_7::setup_individuals(double w_max, double selection_pressure) {
  if (exp_m_->sel()->fitness_func() == FITNESS_GLOBAL_SUM) {
#ifdef __REGUL
    fitness_sum_tab_ = new double[phenotypic_target_handler_->nb_env_];
    for (int env_id = 0; env_id < phenotypic_target_handler_->nb_env_; env_id++) {
      fitness_sum_tab_[env_id] = 0;
      for (int indiv_id = 0; indiv_id < exp_m_->nb_indivs(); indiv_id++)
        fitness_sum_tab_[env_id] += previous_individuals[indiv_id]->fitness_by_env_id_[env_id];
    }
#else
    printf("Fitness local sum is not supported for Aevol (only R-Aevol)\n");
      exit(-1);
#endif
  }

  for (int indiv_id = 0; indiv_id < nb_indivs_; ++indiv_id) {
    start_stop_RNA(indiv_id);
    prom_compute_RNA(indiv_id);
    start_protein(indiv_id);
    compute_protein(indiv_id);
    translate_protein(indiv_id, w_max);

#ifdef __REGUL
    solve_network(indiv_id, selection_pressure);
#else
    compute_phenotype(indiv_id);
    compute_fitness(indiv_id, selection_pressure);
#endif
  }

  for (int indiv_id = 0; indiv_id < (int) exp_m_->nb_indivs(); indiv_id++) {
    previous_individuals[indiv_id] = current_individuals[indiv_id];
  }

  stats_best = new Stats_7(this, AeTime::time(), true);
  write_stat();
}

void ExpManager_7::run_a_step(double w_max, double selection_pressure) {
  #ifdef WITH_PERF_TRACES
      auto runtime_start = std::chrono::steady_clock::now();
      auto runtime_end = std::chrono::steady_clock::now();
      auto runtime_counter = runtime_end.time_since_epoch().count() - runtime_start.time_since_epoch().count();
      runtime_counter = -1;
      auto stats_tree_backup_start = std::chrono::steady_clock::now();
      auto stats_tree_backup_end = std::chrono::steady_clock::now();
      auto stats_tree_backup_counter = stats_tree_backup_end.time_since_epoch().count() - stats_tree_backup_start.time_since_epoch().count();

      auto selection_start = std::chrono::steady_clock::now();
      auto selection_end = std::chrono::steady_clock::now();
      auto selection_counter = selection_end.time_since_epoch().count() - selection_start.time_since_epoch().count();

      auto evaluate_start = std::chrono::steady_clock::now();
      auto evaluate_end = std::chrono::steady_clock::now();
      auto evalute_counter = evaluate_end.time_since_epoch().count() - evaluate_start.time_since_epoch().count();

      auto sort_start = std::chrono::steady_clock::now();
      auto sort_end = std::chrono::steady_clock::now();
      auto sort_counter = sort_end.time_since_epoch().count() - sort_start.time_since_epoch().count();

      auto tree_start = std::chrono::steady_clock::now();
      auto tree_end = std::chrono::steady_clock::now();
      auto tree_counter = tree_end.time_since_epoch().count() - tree_start.time_since_epoch().count();

      auto clean_start = std::chrono::steady_clock::now();
      auto clean_end = std::chrono::steady_clock::now();
      auto clean_counter = clean_end.time_since_epoch().count() - clean_start.time_since_epoch().count();

      auto stats_start = std::chrono::steady_clock::now();
      auto stats_end = std::chrono::steady_clock::now();
      auto stats_counter = clean_end.time_since_epoch().count() - clean_start.time_since_epoch().count();


      auto tree_backup_start = std::chrono::steady_clock::now();
      auto tree_backup_end = std::chrono::steady_clock::now();
      auto tree_backup_counter = clean_end.time_since_epoch().count() - clean_start.time_since_epoch().count();
  #endif

  #ifdef WITH_PERF_TRACES_PER_INDIV
    for (int i = 0; i < nb_indivs_; i++) {
        apply_mutation[i] = -1;
        compute_rna_[i] = -1;
        start_protein_[i] = -1;
        compute_protein_[i] = -1;
        translate_protein_[i] = -1;
        compute_phenotype_[i] = -1;
        compute_fitness_[i] = -1;
        total_[i] = -1;


        allocate_individual_start_[i] = -1;
        apply_mutation_start_[i] = -1;
        compute_rna_start_[i] = -1;
        start_protein_start_[i] = -1;
        compute_protein_start_[i] = -1;
        translate_protein_start_[i] = -1;
        compute_phenotype_start_[i] = -1;
        compute_fitness_start_[i] = -1;
        allocate_individual_stop_[i] = -1;
        apply_mutation_stop_[i] = -1;
        compute_rna_stop_[i] = -1;
        start_protein_stop_[i] = -1;
        compute_protein_stop_[i] = -1;
        translate_protein_stop_[i] = -1;
        compute_phenotype_stop_[i] = -1;
        compute_fitness_stop_[i] = -1;
        total_start_[i] = -1;
        total_stop_[i] = -1;


        omp_tid_[i] = -1;
    }
  #endif


  #ifdef WITH_PERF_TRACES
  #pragma omp single copyprivate(runtime_start,selection_start)
  #else
  #pragma omp single
  #endif
  {
    nb_clones_ = 0;
    mutant_list_.clear();
    mutant_list_.reserve(nb_indivs_);

    if (exp_m_->sel()->fitness_func() == FITNESS_GLOBAL_SUM) {
#ifdef __REGUL
      fitness_sum_tab_ = new double[phenotypic_target_handler_->nb_env_];
        for (int env_id = 0; env_id < phenotypic_target_handler_->nb_env_; env_id++) {
          fitness_sum_tab_[env_id] = 0;
          for (int indiv_id = 0; indiv_id < exp_m_->nb_indivs(); indiv_id++)
            fitness_sum_tab_[env_id] += previous_individuals[indiv_id]->fitness_by_env_id_[env_id];
        }
#else
      printf("Fitness global sum is not supported for Aevol (only R-Aevol)\n");
      exit(-1);
#endif
    }
    #ifdef WITH_PERF_TRACES
      runtime_start = std::chrono::steady_clock::now();
      selection_start = std::chrono::steady_clock::now();
    #endif
  }

#pragma omp single
{
  if (exp_m_->sel()->selection_scope() == SCOPE_GLOBAL) {
    double *  local_fit_array   = new double[nb_indivs_];
    double *  probs             = new double[nb_indivs_];
    double    sum_local_fit     = 0.0;

    for (int loc_indiv_id = 0; loc_indiv_id < nb_indivs_; loc_indiv_id++) {
      local_fit_array[loc_indiv_id] = previous_individuals[loc_indiv_id]->fitness_;
      sum_local_fit += local_fit_array[loc_indiv_id];
    }

    for (int loc_indiv_id = 0; loc_indiv_id < nb_indivs_; loc_indiv_id++) {
      probs[loc_indiv_id] = local_fit_array[loc_indiv_id]/sum_local_fit;
    }

    int32_t* nb_offsprings = new int32_t[nb_indivs_];
    exp_m_->world()->grid(0,0)->reprod_prng_simd_->multinomial_drawing(nb_offsprings, probs, nb_indivs_, nb_indivs_);

    int index = 0;

    int32_t x = 0;
    int32_t y = 0;

    for (int32_t loc_indiv = 0; loc_indiv < nb_indivs_; loc_indiv++) {
      for (int32_t j = 0; j < nb_offsprings[loc_indiv]; j++) {
        x = index / grid_height_;
        y = index % grid_height_;
        exp_m_->next_generation_reproducer_[x*grid_height_+y] = loc_indiv;
        index++;
      }
    }

    delete [] nb_offsprings;
    delete [] local_fit_array;
    delete [] probs;
  }
}

#pragma omp for schedule(dynamic)
  for (int indiv_id = 0; indiv_id < nb_indivs_; indiv_id++) {
    selection(indiv_id);
    do_mutation(indiv_id);
  }

#ifdef WITH_PERF_TRACES
#pragma omp single copyprivate(selection_end,sort_start)
{
  selection_end = std::chrono::steady_clock::now();
  sort_start = std::chrono::steady_clock::now();
}
#endif

#ifdef __OMP_LIST_SORT
#pragma omp single
{
  #if 0 == __OMP_LIST_SORT
  sort(mutant_list_.begin(), mutant_list_.end(), [this](int a, int b) {
        return dna_mutator_array_[a]->length_ > dna_mutator_array_[b]->length_;
    });
  #elif 1 == __OMP_LIST_SORT
  unsigned seed = std::chrono::system_clock::now().time_since_epoch().count();

  std::shuffle (mutant_list_.begin(), mutant_list_.end(), std::default_random_engine(seed));
  #endif
}
#endif


#ifdef WITH_PERF_TRACES
#pragma omp single copyprivate(sort_end,evaluate_start)
{
  sort_end = std::chrono::steady_clock::now();
  evaluate_start = std::chrono::steady_clock::now();
}
#endif

#ifdef __GNUC__
#pragma omp for schedule(dynamic,1)
#else
#pragma omp for schedule(monotonic:dynamic,1)
#endif
for (size_t index = 0; index < mutant_list_.size(); index++) {
  int32_t indiv_id = mutant_list_[index];

    #ifdef WITH_PERF_TRACES_PER_INDIV
      auto indiv_runtime_start = std::chrono::steady_clock::now();
      omp_tid_[indiv_id] = omp_get_thread_num();
    #endif

    if (dna_mutator_array_[indiv_id]->hasMutate())
      apply_mutations(indiv_id);
}

#ifdef __GNUC__
#pragma omp for schedule(dynamic,1)
#else
#pragma omp for schedule(monotonic:dynamic,1)
#endif
for (size_t index = 0; index < mutant_list_.size(); index++) {
  int32_t indiv_id = mutant_list_[index];

    #ifdef WITH_PERF_TRACES_PER_INDIV
      auto indiv_runtime_start = std::chrono::steady_clock::now();
      omp_tid_[indiv_id] = omp_get_thread_num();
    #endif

    if (dna_mutator_array_[indiv_id]->hasMutate()
    #ifdef __REGUL
    || phenotypic_target_handler_->hasChanged_
    #endif
    ) {

      #ifdef __REGUL
      if (dna_mutator_array_[indiv_id]->hasMutate()) {
      #endif

      prom_compute_RNA(indiv_id);
      start_protein(indiv_id);
      compute_protein(indiv_id);
      translate_protein(indiv_id, w_max);
      #ifdef __REGUL
        if (exp_m_->exp_s()->get_with_heredity()) {
          current_individuals[indiv_id]->annotated_chromosome_->proteins_.add_inherited_proteins();
        }
      } else {
        if (exp_m_->exp_s()->get_with_heredity()) {
          current_individuals[indiv_id]->annotated_chromosome_->proteins_.reinit_proteins_inherited();
        }

        // Reinit rnas affinity list
        for (Rna_7* rna : current_individuals[indiv_id]->annotated_chromosome_->rnas()) {
          if (rna != nullptr) {
            if (rna->is_coding_) {
              rna->affinity_list.clear();
            }
          }
        }
      }
      #endif
#ifdef __REGUL
      solve_network(indiv_id, selection_pressure);
#else
      compute_phenotype(indiv_id);
      compute_fitness(indiv_id, selection_pressure);
#endif
        }

    #ifdef WITH_PERF_TRACES_PER_INDIV
      auto indiv_runtime_end = std::chrono::steady_clock::now();
      total_[indiv_id] = indiv_runtime_end.time_since_epoch().count() - indiv_runtime_start.time_since_epoch().count();
      total_start_[indiv_id] = indiv_runtime_start.time_since_epoch().count();
      total_stop_[indiv_id] = indiv_runtime_end.time_since_epoch().count();
    #endif
}



// Traces
#ifdef WITH_PERF_TRACES_PER_INDIV
#pragma omp single
{
    std::ofstream perf_traces_file_;
            perf_traces_file_.open("simd_perf_traces.csv", std::ofstream::app);

                            std::ofstream gantt_traces_file_;
            gantt_traces_file_.open("gantt_perf_traces.csv", std::ofstream::app);

            for (int indiv_id = 0; indiv_id < (int) exp_m_->nb_indivs(); indiv_id++) {
                perf_traces_file_ << AeTime::time() << "," << indiv_id << "," << apply_mutation[indiv_id]
                      << "," << start_protein_[indiv_id] << "," << compute_protein_[indiv_id]
                      << "," << translate_protein_[indiv_id] << "," << compute_phenotype_[indiv_id]
                      << "," << compute_fitness_[indiv_id] << "," << total_[indiv_id]  << std::endl;


                gantt_traces_file_ << AeTime::time() << "," << indiv_id << "," << omp_tid_[indiv_id] << ","
                      << ",TOTAL," << total_start_[indiv_id]
                      << "," << total_stop_[indiv_id] << "," << dna_mutator_array_[indiv_id]->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->dna_->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->rnas().size()
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->proteins().size() << std::endl;

                gantt_traces_file_ << AeTime::time() << "," << indiv_id << "," << omp_tid_[indiv_id] << ","
                      << ",ALLOCATE," << allocate_individual_start_[indiv_id]
                      << "," << allocate_individual_stop_[indiv_id] << "," << dna_mutator_array_[indiv_id]->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->dna_->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->rnas().size()
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->proteins().size() << std::endl;

                gantt_traces_file_ << AeTime::time() << "," << indiv_id << "," << omp_tid_[indiv_id] << ","
                      << ",APPLY_MUTATION," << apply_mutation_start_[indiv_id]
                      << "," << apply_mutation_stop_[indiv_id] << "," << dna_mutator_array_[indiv_id]->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->dna_->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->rnas().size()
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->proteins().size() << std::endl;

                gantt_traces_file_ << AeTime::time() << "," << indiv_id << "," << omp_tid_[indiv_id] << ","
                      << ",COMPUTE_RNA," << compute_rna_start_[indiv_id]
                      << "," << compute_rna_stop_[indiv_id] << "," << dna_mutator_array_[indiv_id]->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->dna_->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->rnas().size()
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->proteins().size() << std::endl;

                gantt_traces_file_ << AeTime::time() << "," << indiv_id << "," << omp_tid_[indiv_id] << ","
                      << ",START_PROTEIN," << start_protein_start_[indiv_id]
                      << "," << start_protein_stop_[indiv_id] << "," << dna_mutator_array_[indiv_id]->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->dna_->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->rnas().size()
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->proteins().size() << std::endl;

                gantt_traces_file_ << AeTime::time() << "," << indiv_id << "," << omp_tid_[indiv_id] << ","
                      << ",COMPUTE_PROTEIN," << compute_protein_start_[indiv_id]
                      << "," << compute_protein_stop_[indiv_id] << "," << dna_mutator_array_[indiv_id]->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->dna_->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->rnas().size()
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->proteins().size() << std::endl;

                gantt_traces_file_ << AeTime::time() << "," << indiv_id << "," << omp_tid_[indiv_id] << ","
                      << ",TRANSLATE_PROTEIN," << translate_protein_start_[indiv_id]
                      << "," << translate_protein_stop_[indiv_id] << "," << dna_mutator_array_[indiv_id]->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->dna_->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->rnas().size()
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->proteins().size() << std::endl;

                gantt_traces_file_ << AeTime::time() << "," << indiv_id << "," << omp_tid_[indiv_id] << ","
                      << ",COMPUTE_PHENOTYPE," << compute_phenotype_start_[indiv_id]
                      << "," << compute_phenotype_stop_[indiv_id] << "," << dna_mutator_array_[indiv_id]->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->dna_->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->rnas().size()
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->proteins().size() << std::endl;

                gantt_traces_file_ << AeTime::time() << "," << indiv_id << "," << omp_tid_[indiv_id] << ","
                      << ",COMPUTE_FITNESS," << compute_fitness_start_[indiv_id]
                      << "," << compute_fitness_stop_[indiv_id] << "," << dna_mutator_array_[indiv_id]->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->dna_->length_
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->rnas().size()
                      << "," << current_individuals[indiv_id]->annotated_chromosome_->proteins().size() << std::endl;

                apply_mutation_start_[indiv_id] = -1;
                compute_rna_start_[indiv_id] = -1;
                start_protein_start_[indiv_id] = -1;
                compute_protein_start_[indiv_id] = -1;
                translate_protein_start_[indiv_id] = -1;
                compute_phenotype_start_[indiv_id] = -1;
                compute_fitness_start_[indiv_id] = -1;

                apply_mutation_stop_[indiv_id] = -1;
                compute_rna_stop_[indiv_id] = -1;
                start_protein_stop_[indiv_id] = -1;
                compute_protein_stop_[indiv_id] = -1;
                translate_protein_stop_[indiv_id] = -1;
                compute_phenotype_stop_[indiv_id] = -1;
                compute_fitness_stop_[indiv_id] = -1;

                apply_mutation[indiv_id] = -1;
                compute_rna_[indiv_id] = -1;
                start_protein_[indiv_id] = -1;
                compute_protein_[indiv_id] = -1;
                translate_protein_[indiv_id] = -1;
                compute_phenotype_[indiv_id] = -1;
                compute_fitness_[indiv_id] = -1;
                total_[indiv_id] = -1;
                total_start_[indiv_id] = -1;
                total_stop_[indiv_id] = -1;
                omp_tid_[indiv_id] = -1;
            }
            perf_traces_file_.close();
            gantt_traces_file_.close();
}
#endif

#ifdef WITH_PERF_TRACES
#pragma omp single copyprivate(evaluate_end,tree_start)
{
  evaluate_end = std::chrono::steady_clock::now();
  tree_start = std::chrono::steady_clock::now();
}
#endif

#pragma omp for schedule(dynamic)
  for (int indiv_id = 0; indiv_id < nb_indivs_; indiv_id++) {
    if (exp_m_->record_tree()) {
      int x = indiv_id / exp_m_->world()->height();
      int y = indiv_id % exp_m_->world()->height();

      auto* eindiv =
          new EndReplicationEvent(current_individuals[indiv_id], x, y);
      // Tell observers the replication is finished
      exp_m_->tree()->update_end_replication(eindiv);
      delete eindiv;
    }

    current_individuals[indiv_id]->first_to_add = true;
  }

#ifdef WITH_PERF_TRACES
#pragma omp single copyprivate(runtime_counter,tree_end,clean_start)
#else
#pragma omp single
#endif
  {
    #ifdef WITH_PERF_TRACES
      runtime_end = std::chrono::steady_clock::now();
      runtime_counter = runtime_end.time_since_epoch().count() - runtime_start.time_since_epoch().count();

      stats_tree_backup_start = std::chrono::steady_clock::now();
    #endif

    if (exp_m_->record_tree())
      exp_m_->tree()->update_end_generation();

    #ifdef WITH_PERF_TRACES
      tree_end =  std::chrono::steady_clock::now();
      clean_start = std::chrono::steady_clock::now();
    #endif


  }

#pragma omp for schedule(static)
    for (int indiv_id = 0; indiv_id < (int) exp_m_->nb_indivs(); indiv_id++) {

      bool toDelete = false;

#pragma omp critical(indiv_list)
      {
        previous_individuals[indiv_id]->usage_count_--;
        if (previous_individuals[indiv_id]->usage_count_ == 0) {
          toDelete = true;
        }
      }


      if (toDelete) {
        delete previous_individuals[indiv_id];
      }


      previous_individuals[indiv_id] = current_individuals[indiv_id];
      current_individuals[indiv_id]  = nullptr;


    }


    #pragma omp single
    {
      if (AeTime::time() % DnaFactory::cleanup_step == 0) Dna_7::reduce_dna_factory_memory_usage(this);
    }
#ifdef WITH_PERF_TRACES
#pragma omp single copyprivate(clean_end,stats_start)
  {
      clean_end = std::chrono::steady_clock::now();
      stats_start = std::chrono::steady_clock::now();
  }
#endif

#ifndef AEVOL_NO_STATS
#pragma omp single
  {
      double best_fitness = previous_individuals[0]->fitness_;

  for (int indiv_id = 1; indiv_id < (int) exp_m_->nb_indivs(); indiv_id++) {
    if (previous_individuals[indiv_id]->fitness_ > best_fitness) {
      best_fitness = previous_individuals[indiv_id]->fitness_;

    }

  }
    write_stat();
  }
#endif

  #pragma omp single
  write_ui_output();


    #ifdef WITH_PERF_TRACES
#pragma omp single copyprivate(stats_end,tree_backup_start)
  {
      stats_end = std::chrono::steady_clock::now();
      tree_backup_start = std::chrono::steady_clock::now();
  }
#endif

#pragma omp single
    {
      if (exp_m_->record_tree() && AeTime::time() %  exp_m_->tree_step() == 0) {
        int status;
        status = mkdir(TREE_DIR, 0755);
        if ((status == -1) && (errno != EEXIST))
        {
          err(EXIT_FAILURE, "Impossible to create the directory %s", TREE_DIR);
        }

        std::ostringstream tree_file_name;

        tree_file_name << "tree/tree_" << std::setfill('0') << std::setw(9) << AeTime::time() << ".ae";
        std::cout << "  Writing tree for generation " << AeTime::time() << "...";
        std::cout.flush();
        exp_m_->output_m()->tree()->write_to_tree_file(tree_file_name.str().c_str());
        std::cout << " OK" << std::endl;
      }
    }


  if (AeTime::time() % exp_m_->backup_step() == 0) {
#pragma omp single
    {
      std::cout << "  Writing checkpoint for generation " << AeTime::time() << "...";
      std::cout.flush();

      for (int indiv_id = 0; indiv_id < (int) exp_m_->nb_indivs(); indiv_id++) {
        int x = indiv_id / exp_m_->world()->height();
        int y = indiv_id % exp_m_->world()->height();


        exp_m_->world()->grid(x, y)->individual()->clear_everything_except_dna_and_promoters();
        exp_m_->world()->grid(x, y)->individual()->genetic_unit_list_nonconst().clear();
        delete exp_m_->world()->grid(x, y)->individual();

#ifdef __REGUL
        Individual_R *indiv = new Individual_R(exp_m_,
                                                         exp_m_->world()->grid(x, y)->mut_prng(),
                                                         exp_m_->world()->grid(x, y)->stoch_prng(),
                                                         exp_m_->exp_s()->mut_params(),
                                                         w_max,
                                                         exp_m_->exp_s()->min_genome_length(),
                                                         exp_m_->exp_s()->max_genome_length(),
                                                         indiv_id,
                                                         "",
                                                         0);
#else
        Individual *indiv = new Individual(exp_m_,
                                           exp_m_->world()->grid(x, y)->mut_prng(),
                                           exp_m_->world()->grid(x, y)->stoch_prng(),
                                           exp_m_->exp_s()->mut_params(),
                                           w_max,
                                           exp_m_->exp_s()->min_genome_length(),
                                           exp_m_->exp_s()->max_genome_length(),
                                           indiv_id,
                                           "",
                                           0);
#endif

        int32_t nb_blocks_ =
            previous_individuals[indiv_id]->annotated_chromosome_->dna_->nb_block();
        char *dna_string = new char[nb_blocks_ * BLOCK_SIZE];
        memset(dna_string, 0,
               (previous_individuals[indiv_id]->annotated_chromosome_->dna_->length() + 1) * sizeof(char));


        char *to_copy = previous_individuals[indiv_id]->annotated_chromosome_->dna_->to_char();


        memcpy(dna_string, to_copy,
               (previous_individuals[indiv_id]->annotated_chromosome_->dna_->length() + 1) * sizeof(char));

        dna_string[previous_individuals[indiv_id]->annotated_chromosome_->dna_->length()] = '\0';

        indiv->add_GU(dna_string,
                      previous_individuals[indiv_id]->annotated_chromosome_->dna_->length());
        indiv->genetic_unit_nonconst(0).set_min_gu_length(exp_m_->exp_s()->min_genome_length());
        indiv->genetic_unit_nonconst(0).set_max_gu_length(exp_m_->exp_s()->max_genome_length());
        indiv->compute_statistical_data();
        indiv->EvaluateInContext(exp_m_->world()->grid(x, y)->habitat());


        exp_m_->world()->grid(x, y)->set_individual(indiv);
      }

      auto idx_best = find_best_indiv_idx(previous_individuals, exp_m_->nb_indivs());
      best_indiv = previous_individuals[idx_best];
      exp_m_->world()->x_best = idx_best / exp_m_->world()->height();
      exp_m_->world()->y_best = idx_best % exp_m_->world()->height();

      exp_m_->WriteDynamicFiles();

      std::ofstream last_gener_file(LAST_GENER_FNAME,
                                    std::ofstream::out);

      last_gener_file << AeTime::time() << std::endl;
      last_gener_file.close();
      std::cout << " OK" << std::endl;
    }
  }

#ifdef WITH_PERF_TRACES
#pragma omp single copyprivate(tree_backup_end)
  {
      tree_backup_end = std::chrono::steady_clock::now();
  }
#endif

  #pragma omp single
  {
#ifdef WITH_PERF_TRACES
    clean_end = std::chrono::steady_clock::now();

    stats_tree_backup_end = std::chrono::steady_clock::now();
    stats_tree_backup_counter = stats_tree_backup_end.time_since_epoch().count() - stats_tree_backup_start.time_since_epoch().count();

      selection_counter = selection_end.time_since_epoch().count() - selection_start.time_since_epoch().count();

      evalute_counter = evaluate_end.time_since_epoch().count() - evaluate_start.time_since_epoch().count();
      sort_counter = sort_end.time_since_epoch().count() - sort_start.time_since_epoch().count();
      tree_counter = tree_end.time_since_epoch().count() - tree_start.time_since_epoch().count();

      clean_counter = clean_end.time_since_epoch().count() - clean_start.time_since_epoch().count();

      stats_counter = stats_end.time_since_epoch().count() - stats_start.time_since_epoch().count();

      clean_counter = tree_backup_end.time_since_epoch().count() - tree_backup_start.time_since_epoch().count();
    std::ofstream perf_traces_file_;

    perf_traces_file_.open("sm_perf_traces.csv", std::ofstream::app);
    perf_traces_file_<< AeTime::time() << "," << runtime_counter << "," << stats_tree_backup_counter << "," <<
                       selection_counter << "," << sort_counter <<"," << evalute_counter << "," <<
                       tree_counter << "," <<clean_counter<<","<<stats_counter<<","<<tree_backup_counter<<"," <<
                       nb_clones_ <<std::endl;
    perf_traces_file_.close();
#endif
  }
}

void ExpManager_7::setup_ui_output(const std::filesystem::path& outdir, int32_t frequency) {
  user_interface_output_ = std::make_unique<UserInterfaceOutput>(outdir, frequency);
}

void ExpManager_7::write_ui_output() {
  if (not user_interface_output_) return;
  if (user_interface_output_->nothing_to_output()) return;

  auto idx_best = find_best_indiv_idx(previous_individuals, exp_m_->nb_indivs());
  user_interface_output_->write_indiv_output(*previous_individuals[idx_best]);
  user_interface_output_->write_grid_output(grid_width_, grid_height_, previous_individuals);
  #ifndef __REGUL
    if (AeTime::time() == 0) {
      user_interface_output_->write_environment_output(target);
    }
  #endif
}

void ExpManager_7::write_stat(bool non_coding) {
  #ifdef PROGENY_STATS
  std::map<int32_t,int32_t> progeny_per_indiv;

  for (int indiv_id = 1; indiv_id < (int) exp_m_->nb_indivs(); indiv_id++) {
    int32_t parent_id = exp_m_->next_generation_reproducer_[indiv_id];

    if ( progeny_per_indiv.find(parent_id) == progeny_per_indiv.end() ) {
      progeny_per_indiv[parent_id]=1;
    } else {
      progeny_per_indiv[parent_id]++;
    }
  }

  std::string file = "stats/progeny.csv";
  std::string file_rep = "stats/reproducer_progeny.csv";

  std::ofstream progeny;
  progeny.open(file,std::ofstream::app);
  std::ofstream progeny_rep;
  progeny_rep.open(file_rep,std::ofstream::app);
  progeny_rep<<AeTime::time()<<","<<progeny_per_indiv.size()<<std::endl;
  progeny_rep.close();

  for (auto prog : progeny_per_indiv) {
    progeny<<AeTime::time()<<","<<prog.first<<","<<prog.second<<std::endl;
  }

  progeny.close();
  #endif


  stats_best->reinit(AeTime::time());

  // Search for the best
  auto idx_best = find_best_indiv_idx(previous_individuals, exp_m_->nb_indivs());
  best_indiv = previous_individuals[idx_best];
  best_indiv->annotated_chromosome_->reset_stats();

  // ANNOTATE_SITE_BEGIN(mainloop);
  for (Rna_7* rna : best_indiv->annotated_chromosome_->rnas()) {
    if (rna != nullptr) {
      if (rna->is_coding_)
        best_indiv->annotated_chromosome_->nb_coding_RNAs++;
      else
        best_indiv->annotated_chromosome_->nb_non_coding_RNAs++;
    }
  }


  for (const auto& prot : best_indiv->annotated_chromosome_->proteins()) {
    if (prot != nullptr) {
      if (prot->is_functional) {
        best_indiv->annotated_chromosome_->nb_func_genes++;
      } else {
        best_indiv->annotated_chromosome_->nb_non_func_genes++;
      }
      if (prot->h > 0) {
        best_indiv->annotated_chromosome_->nb_genes_activ++;
      } else {
        best_indiv->annotated_chromosome_->nb_genes_inhib++;
      }
    }
  }

  stats_best->write(best_indiv,non_coding);
}



void ExpManager_7::write_stat(Stats_7* stats, Individual_7* indiv, int32_t generation, bool non_coding) {
  if (generation == -1) stats->reinit(AeTime::time());
  else stats->reinit(generation);

  indiv->annotated_chromosome_->reset_stats();
  for (Rna_7* rna : indiv->annotated_chromosome_->rnas()) {
    if (rna != nullptr) {
      if (rna->is_coding_)
        indiv->annotated_chromosome_->nb_coding_RNAs++;
      else
        indiv->annotated_chromosome_->nb_non_coding_RNAs++;
    }
  }

  for (const auto& prot : indiv->annotated_chromosome_->proteins()) {
    if (prot != nullptr) {
      if (prot->is_functional) {
        indiv->annotated_chromosome_->nb_func_genes++;
      } else {
        indiv->annotated_chromosome_->nb_non_func_genes++;
      }
      if (prot->h > 0) {
        indiv->annotated_chromosome_->nb_genes_activ++;
      } else {
        indiv->annotated_chromosome_->nb_genes_inhib++;
      }
    }
  }

  stats->write(indiv,non_coding);
}

}
