// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************


#ifndef AEVOL_EXPMANAGER_7_H
#define AEVOL_EXPMANAGER_7_H

#include "DnaFactory.h"
#include "DnaMutator.h"
#include "Observable.h"
#include "PhenotypicTargetHandler.h"
#include "raevol/SIMD_PhenotypicTargetHandler_R.h"
#include "ExpSetup.h"
#include "7/fuzzy/FuzzyFactory_7.h"
#include "World.h"
#include "io/UserInterfaceOutput.h"

namespace aevol {

#ifdef BASE_2
constexpr const char* PROM_SEQ_LEAD = "0101011001110010010110";
constexpr const char* PROM_SEQ_LAG  = "1010100110001101101001";

#if defined(__INTEL_COMPILER) || defined(__INTEL_LLVM_COMPILER)
__declspec(align(64)) constexpr const char* SHINE_DAL_SEQ_LEAD_7 = "0110111111000";
__declspec(align(64)) constexpr const char* SHINE_DAL_SEQ_LAG_7  = "1001001111111";
#else
constexpr const char* SHINE_DAL_SEQ_LEAD_7 = "0110111111000";
constexpr const char* SHINE_DAL_SEQ_LAG_7  = "1001001111111";
#endif

// constexpr const char* SHINE_DAL_SEQ_LEAD = "011011000";
// constexpr const char* SHINE_DAL_SEQ_LAG  = "100100111";

constexpr const char* PROTEIN_END_LEAD = "001";
constexpr const char* PROTEIN_END_LAG  = "110";
#endif

class ToFetchIndividual {
  public:
    ToFetchIndividual(int32_t id, int32_t x, int32_t y, int32_t rank) 
    {
      id_ = id; x_ = x; y_ = y; rank_ = rank; 
      // printf("---> Add FETCHIndividual %d (%d %d)\n",id_,x_,y_);
    
    }

    int32_t x_;
    int32_t y_;
    int32_t id_;
    int32_t rank_;


    int32_t local_x_;
    int32_t local_y_;

    char* dna_ = nullptr;
    int32_t dna_length_ = -1;

    int32_t lead_prom_size;

    int32_t* lead_prom_pos = nullptr;
    int8_t* lead_prom_error = nullptr;
            
    int32_t lag_prom_size;
            
    int32_t* lag_prom_pos = nullptr;
    int8_t* lag_prom_error = nullptr;


    bool fetched_ = false;
};

class ExpManager;
class Stats_7;
// class FuzzyFactory_7;

class ExpManager_7 : public Observable{
 public:
  ExpManager_7() =  default;
  ~ExpManager_7();

  void load_legacy(ExpManager* exp_m, const ExpSetup* exp_s, const World* world);
  void setup_individuals(double w_max, double selection_pressure);
  void run_a_step(double w_max, double selection_pressure);

// Function with ID
  void apply_mutations(int indiv_id);

  void do_mutation(int indiv_id);
  void start_stop_RNA(int indiv_id);
  void prom_compute_RNA(int indiv_id);
  void start_protein(int indiv_id);
  void compute_protein(int indiv_id);
  void translate_protein(int indiv_id, double w_max);
  void compute_phenotype(int indiv_id);
  #ifdef __REGUL
  void compute_fitness(int indiv_id, double selection_pressure, int env_id = -1);
  #else
  void compute_fitness(int indiv_id, double selection_pressure);
  #endif
#ifdef __REGUL
  void compute_network(int indiv_id);
  void update_network(int indiv_id);
  void evaluate_network(int indiv_id, double selection_pressure, int env_id);
  void finalize_network(int indiv_id, double selection_pressure);
  void solve_network(int indiv_id, double selection_pressure);
  void update_phenotype( int indiv_id );
#endif

  void setup_ui_output(const std::filesystem::path& outdir, int32_t frequency);
  void write_ui_output();

  void write_stat(bool non_coding = false);
  void write_stat(Stats_7* stats, Individual_7* indiv, int32_t generation = -1, bool non_coding = false);

  int32_t find_best_indiv_idx(Individual_7** indivs, int32_t nb_indivs) const;

  Individual_7** current_individuals;
  Individual_7** previous_individuals;
  Individual_7* best_indiv;

  std::vector<int> mutant_list_;

  int32_t nb_indivs_;
  int32_t nb_clones_;

  int rna_grain_size = 32;
  int protein_grain_size = 32;

  // static bool compute_diff_rnas;

  FuzzyFactory_7* fuzzy_factory_;
  // DNA Mutator tables
  DnaMutator** dna_mutator_array_ = nullptr;

  double* fitness_sum_tab_;
#ifdef __REGUL
  SIMD_PhenotypicTargetHandler_R* phenotypic_target_handler_;
#else
  AbstractFuzzy* target;
#endif

  int32_t grid_width_;
  int32_t grid_height_;

 private:
  ExpManager* exp_m_;

  Stats_7* stats_best = nullptr;
  Stats_7* stats_mean = nullptr;

  void selection(int indiv_id);

  std::unique_ptr<UserInterfaceOutput> user_interface_output_;

#ifdef WITH_PERF_TRACES_PER_INDIV
  long* apply_mutation;
  long* compute_rna_;
  long* start_protein_;
  long* compute_protein_;
  long* translate_protein_;
  long* compute_phenotype_;
  long* compute_fitness_;
  long* total_;

  long* allocate_individual_start_;
  long* apply_mutation_start_;
  long* compute_rna_start_;
  long* start_protein_start_;
  long* compute_protein_start_;
  long* translate_protein_start_;
  long* compute_phenotype_start_;
  long* compute_fitness_start_;
  
  long* allocate_individual_stop_;
  long* apply_mutation_stop_;
  long* compute_rna_stop_;
  long* start_protein_stop_;
  long* compute_protein_stop_;
  long* translate_protein_stop_;
  long* compute_phenotype_stop_;
  long* compute_fitness_stop_;

  long* total_start_;
  long* total_stop_;
  long* omp_tid_;
#endif


};
}

#endif //AEVOL_EXPMANAGER_7_H
