// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************


#include "Individual_7.h"
#include "DnaMutator.h"
#include "Stats_7.h"
#include "7/List_Metadata.h"
#include "7/Protein_7.h"

#include "legacy/ExpManager.h"
#include "legacy/phenotype/fuzzy/Fuzzy.h"

#include <algorithm>

namespace aevol {

Individual_7::Individual_7(double w_max, FuzzyFactory_7* fuzzy_factory) :
    w_max_(w_max),
    fuzzy_factory_(fuzzy_factory) {
}

Individual_7* Individual_7::make_empty(double w_max, FuzzyFactory_7* fuzzy_factory) {
  Individual_7* indiv = new Individual_7(w_max, fuzzy_factory);

  indiv->annotated_chromosome_= AnnotatedChromosome::make_empty();

  return indiv;
}

Individual_7* Individual_7::make_from_sequence(double w_max,
                                               FuzzyFactory_7* fuzzy_factory,
                                               const char* seq,
                                               Dna_7::size_type len) {
  Individual_7* indiv = new Individual_7(w_max, fuzzy_factory);

  indiv->annotated_chromosome_ = AnnotatedChromosome::make_from_sequence(seq, len);
  indiv->annotated_chromosome_->dna_->set_indiv(indiv);

  return indiv;
}

#ifdef __REGUL
Individual_7* Individual_7::make_clone(Individual_7* orig,
                                       FuzzyFactory_7* fuzzy_factory,
                                       bool no_metadata) {
#else
Individual_7* Individual_7::make_clone(Individual_7* orig,
                                       FuzzyFactory_7* fuzzy_factory) {
#endif
  Individual_7* indiv = new Individual_7(orig->w_max_, fuzzy_factory);

  indiv->annotated_chromosome_= AnnotatedChromosome::make_clone(orig->annotated_chromosome_, indiv);

  #ifdef __REGUL
    if (not no_metadata && exp_manager->exp_s()->get_with_heredity()) {
      for (auto prot : orig->annotated_chromosome_->proteins()) {
        if (prot != nullptr && prot->is_init_ && (!prot->signal_ && !prot->inherited_)) {
          if (prot->e > exp_manager->exp_s()->get_protein_presence_limit() && !(prot->signal_)) {
            indiv->annotated_chromosome_->proteins_.add_inherited_protein(Protein_7::make_inherited_protein(prot));
          }
        }
      }
    }
  #endif

  indiv->fitness_  = orig->fitness_;
  indiv->metaerror = orig->metaerror;

  return indiv;
}

Individual_7::~Individual_7() {
  if (phenotype!=nullptr) {
    fuzzy_factory_->give_back(phenotype);
    phenotype = nullptr;
  }

  delete annotated_chromosome_;

  clearAllObserver();
}

#ifdef __REGUL
void Individual_7::evaluate(const ExpSetup& exp_s,
                            double w_max,
                            double selection_pressure,
                            SIMD_PhenotypicTargetHandler_R* pth) {
#else
void Individual_7::evaluate(const ExpSetup& exp_s,
                            double w_max,
                            double selection_pressure,
                            const AbstractFuzzy* target) {
#endif
  start_stop_RNA();
  prom_compute_RNA(exp_s);
  start_protein();
  compute_protein();
  translate_protein(exp_s, w_max);

  #ifdef __REGUL
    solve_network(exp_s, selection_pressure, pth);
  #else
    compute_phenotype();
    compute_fitness(selection_pressure, target);
  #endif
}

void Individual_7::start_stop_RNA() {
  for (Dna_7::size_type dna_pos = 0 ; dna_pos < annotated_chromosome_->dna_->length() ; ++dna_pos) {
    auto len = annotated_chromosome_->dna_->length();
    if (len < PROM_SIZE) continue;

    #ifdef BASE_2
      int prom_dist_leading[26];
      int prom_dist_lagging[26];

      for (int motif_id = 0; motif_id < 52; motif_id++) {
        if (motif_id >= 26 && motif_id < 48) {
          // LAGGING
          int t_motif_id = motif_id - 26;
          prom_dist_lagging[t_motif_id] =
              PROM_SEQ_LAG[t_motif_id] ==
                      annotated_chromosome_->dna_->data_[
                          dna_pos - t_motif_id < 0 ? len +
                                                         dna_pos -
                                                         t_motif_id :
                                                   dna_pos - t_motif_id]
                  ? 0 : 1;
        } else if (motif_id < 22) {
          // LEADING
          prom_dist_leading[motif_id] =
              PROM_SEQ_LEAD[motif_id] ==
                      annotated_chromosome_->dna_->data_[
                          dna_pos + motif_id >= len ? dna_pos +
                                                          motif_id -
                                                          len
                                                    : dna_pos +
                                                          motif_id]
                  ? 0
                  : 1;
        } else {
          // removed terminator search
        }
      }

      int i_prom_dist_leading = prom_dist_leading[0] +
                                prom_dist_leading[1] +
                                prom_dist_leading[2] +
                                prom_dist_leading[3] +
                                prom_dist_leading[4] +
                                prom_dist_leading[5] +
                                prom_dist_leading[6] +
                                prom_dist_leading[7] +
                                prom_dist_leading[8] +
                                prom_dist_leading[9] +
                                prom_dist_leading[10] +
                                prom_dist_leading[11] +
                                prom_dist_leading[12] +
                                prom_dist_leading[13] +
                                prom_dist_leading[14] +
                                prom_dist_leading[15] +
                                prom_dist_leading[16] +
                                prom_dist_leading[17] +
                                prom_dist_leading[18] +
                                prom_dist_leading[19] +
                                prom_dist_leading[20] +
                                prom_dist_leading[21];

      int i_prom_dist_lagging = prom_dist_lagging[0] +
                                prom_dist_lagging[1] +
                                prom_dist_lagging[2] +
                                prom_dist_lagging[3] +
                                prom_dist_lagging[4] +
                                prom_dist_lagging[5] +
                                prom_dist_lagging[6] +
                                prom_dist_lagging[7] +
                                prom_dist_lagging[8] +
                                prom_dist_lagging[9] +
                                prom_dist_lagging[10] +
                                prom_dist_lagging[11] +
                                prom_dist_lagging[12] +
                                prom_dist_lagging[13] +
                                prom_dist_lagging[14] +
                                prom_dist_lagging[15] +
                                prom_dist_lagging[16] +
                                prom_dist_lagging[17] +
                                prom_dist_lagging[18] +
                                prom_dist_lagging[19] +
                                prom_dist_lagging[20] +
                                prom_dist_lagging[21];

    #elif BASE_4
      int i_prom_dist_leading = List_Metadata::is_promoter_leading(*annotated_chromosome_->dna_, dna_pos);
      int i_prom_dist_lagging = List_Metadata::is_promoter_lagging(*annotated_chromosome_->dna_, dna_pos);
    #endif

    if (i_prom_dist_leading <= PROM_MAX_DIFF) {
      annotated_chromosome_->promoter_list().promoter_add(Promoter(dna_pos, i_prom_dist_leading, Strand::LEADING));
    }

    if (i_prom_dist_lagging <= PROM_MAX_DIFF) {
      annotated_chromosome_->promoter_list().promoter_add(Promoter(dna_pos, i_prom_dist_lagging, Strand::LAGGING));
    }
  }
}

#ifdef BASE_4
void Individual_7::prom_compute_RNA(const ExpSetup& exp_s) {
#else
void Individual_7::prom_compute_RNA(const ExpSetup&) {
#endif
  #ifdef WITH_PERF_TRACES_PER_INDIV
    auto runtime_start = std::chrono::steady_clock::now();
    compute_rna_start_[indiv_id] = runtime_start.time_since_epoch().count();
  #endif

  annotated_chromosome_->proteins_.clear();
  annotated_chromosome_->rna_list_.clear();

  #ifdef BASE_4
    int8_t term_size = TERM_SIZE + exp_s.terminator_polya_sequence_length();
  #endif

  // ANNOTATE_SITE_BEGIN(mainloop);
  for (int prom_idx = 0 ; prom_idx < annotated_chromosome_->promoter_list().promoter_count() ; prom_idx++) {
    Promoter* prom = annotated_chromosome_->promoter_list().promoter_at(prom_idx);

    if (prom != nullptr) {
      Dna_7*dna = annotated_chromosome_->dna_;
      int32_t dna_length = dna->length();
      int prom_pos;
      Strand strand;
      double prom_error;

      prom_pos = prom->pos;
      strand = prom->strand_;
      prom_error = fabs(static_cast<float>(prom->error));

      #ifdef BASE_2
        int32_t length = dna->length_;
        char* dna_data = dna->data_;
        #if defined(__INTEL_COMPILER)
          __declspec(align(dna_data));
        #elif defined(__INTEL_LLVM_COMPILER)
          void* vec_r = __builtin_assume_aligned(dna_data,64);
        #endif
      #endif

      if (strand == Strand::LEADING) {
        /* Search for terminators */
        int cur_pos =
            Utils::mod(prom_pos + PROM_SIZE,dna_length);
        int start_pos = cur_pos;

        bool terminator_found = false;
        bool no_terminator = false;
        #ifdef BASE_2
          int8_t term_dist_leading = 0;
        #endif

        int loop_size = 0;

        while (!terminator_found) {
          #ifdef BASE_2
            // ANNOTATE_SITE_BEGIN(leading_motif_loop);
            if (cur_pos + 10 < length && cur_pos + 10 >= 0) {
              #pragma omp simd
              for (int32_t t_motif_id = 0; t_motif_id < 4; t_motif_id++) {
                int32_t pos_begin = cur_pos + t_motif_id;
                int32_t pos_end   = cur_pos - t_motif_id + 10;

                // ANNOTATE_ITERATION_TASK(searchLeading);
                term_dist_leading += dna_data[pos_begin] != dna_data[pos_end] ? 1 : 0;
              }
            } else {
              for (int32_t t_motif_id = 0; t_motif_id < 4; t_motif_id++) {
                int32_t pos_begin = cur_pos + t_motif_id;
                int32_t pos_end   = cur_pos - t_motif_id + 10;

                while (pos_begin < 0)       pos_begin += length;
                while (pos_begin >= length) pos_begin -= length;

                while (pos_end < 0)       pos_end += length;
                while (pos_end >= length) pos_end -= length;

                // ANNOTATE_ITERATION_TASK(searchLeading);
                term_dist_leading += dna_data[pos_begin] != dna_data[pos_end] ? 1 : 0;
              }
            }
            // ANNOTATE_SITE_END(leading_motif_loop);

            if (term_dist_leading == 4) {
              terminator_found = true;
            }
            else {
              cur_pos = Utils::mod(cur_pos + 1,dna_length);

              term_dist_leading = 0;
              if (cur_pos == start_pos) {
                no_terminator = true;
                terminator_found = true;
              }
            }


          #elif BASE_4
            terminator_found = annotated_chromosome_->metadata_->is_terminator_leading(cur_pos);

            if(!terminator_found) {
              cur_pos = cur_pos + 1 >= dna_length ? cur_pos + 1 - dna_length : cur_pos + 1;

              if (cur_pos == start_pos) {
                no_terminator = true;
                terminator_found = true;
              }
            }
          #endif

          loop_size++;
        }

        if (!no_terminator) {
          #ifdef BASE_2
            int32_t rna_end = Utils::mod(cur_pos + TERM_SIZE - 1, dna_length);
            int32_t rna_length = 0;

            rna_length = loop_size + TERM_SIZE -1;

            if (rna_length > 0) {
              annotated_chromosome_->rna_list_.rna_add(
                  prom_pos, rna_end, strand, 1.0 - prom_error / 5.0, rna_length, prom);
            }
          #elif BASE_4
            int32_t rna_end    = Utils::mod(cur_pos + term_size - 1, dna_length);
            int32_t rna_length = loop_size - 1 + term_size;

            if (rna_length >= term_size) {  // check that terminator is not overlapping with promoter
              annotated_chromosome_->rna_list_.rna_add(
                  prom_pos, rna_end, strand, BASAL_LEVELS[prom_error], rna_length, prom);
            }
        #endif
      }
    } else {
        /* Search for terminator */
        int cur_pos =
            prom_pos - 22;
        cur_pos =
            cur_pos < 0 ? dna_length + (cur_pos) : cur_pos;
        int start_pos = cur_pos;
        bool terminator_found = false;
        bool no_terminator = false;
        #ifdef BASE_2
          int term_dist_lagging = 0;
        #endif

        int loop_size = 0;

        while (!terminator_found) {
          #ifdef BASE_2
            // ANNOTATE_SITE_BEGIN(lagging_motif_loop);
            if (cur_pos - 10 < length && cur_pos - 10 >= 0) {
              #pragma omp simd
              for (int32_t t_motif_id = 0; t_motif_id < 4; t_motif_id++) {
                int32_t pos_begin = cur_pos - t_motif_id;
                int32_t pos_end   = cur_pos + t_motif_id - 10;

                // ANNOTATE_ITERATION_TASK(searchLeading);
                term_dist_lagging += dna_data[pos_begin] != dna_data[pos_end] ? 1 : 0;
              }
            } else {
              for (int32_t t_motif_id = 0; t_motif_id < 4; ++t_motif_id) {
                int32_t pos_begin = cur_pos - t_motif_id;
                int32_t pos_end   = cur_pos + t_motif_id - 10;

                while (pos_begin < 0)       pos_begin += length;
                while (pos_begin >= length) pos_begin -= length;

                while (pos_end < 0)       pos_end += length;
                while (pos_end >= length) pos_end -= length;

                // ANNOTATE_ITERATION_TASK(searchLeading);
                term_dist_lagging += dna_data[pos_begin] != dna_data[pos_end] ? 1 : 0;
              }
            }
            // ANNOTATE_SITE_END(lagging_motif_loop);

            if (term_dist_lagging == 4) {
              terminator_found = true;
            } else {
              cur_pos = cur_pos - 1 < 0 ? dna_length + (cur_pos - 1) : cur_pos - 1;
              term_dist_lagging = 0;
              if (cur_pos == start_pos) {
                no_terminator = true;
                terminator_found = true;
              }
            }
          #elif BASE_4
            terminator_found = annotated_chromosome_->metadata_->is_terminator_lagging(cur_pos);

            if(!terminator_found) {
              cur_pos = cur_pos - 1 < 0 ? dna_length + (cur_pos - 1) : cur_pos - 1;

              if (cur_pos == start_pos) {
                no_terminator = true;
                terminator_found = true;
              }
            }
          #endif
          loop_size++;
        }

        if (!no_terminator) {
          #ifdef BASE_2
            int32_t rna_end = Utils::mod(cur_pos - 10,dna_length);

            int32_t rna_length = 0;
            rna_length = loop_size + TERM_SIZE -1;

            if (rna_length >= 0) {
              annotated_chromosome_->rna_list_.rna_add(
                  prom_pos, rna_end, strand, 1.0 - prom_error / 5.0, rna_length, prom);
            }

          #elif BASE_4
            int32_t rna_end    = Utils::mod(cur_pos - term_size + 1, dna_length);
            int32_t rna_length = loop_size - 1 + term_size;

            if (rna_length >= term_size) {
              annotated_chromosome_->rna_list_.rna_add(
                  prom_pos, rna_end, strand, BASAL_LEVELS[prom_error], rna_length, prom);
            }
          #endif
        }
      }
    }
  }
  // ANNOTATE_SITE_END(mainloop);

  #ifdef WITH_PERF_TRACES_PER_INDIV
    auto runtime_end = std::chrono::steady_clock::now();
    compute_rna_stop_[indiv_id] = runtime_end.time_since_epoch().count();
    compute_rna_[indiv_id] = runtime_end.time_since_epoch().count() - runtime_start.time_since_epoch().count();
  #endif
}

void Individual_7::start_protein() {
  #ifdef WITH_PERF_TRACES_PER_INDIV
    auto runtime_start             = std::chrono::steady_clock::now();
    start_protein_start_[indiv_id] = runtime_start.time_since_epoch().count();
  #endif

  for (Rna_7* rna: annotated_chromosome_->rnas()) {
    int32_t dna_length = annotated_chromosome_->dna_->length_;

    #ifdef BASE_2
      if (rna->to_compute_start_pos_) {
        rna->start_prot.clear();
        rna->start_prot_count_ = 0;

        char* dna_data = annotated_chromosome_->dna_->data_;
        #if defined(__INTEL_COMPILER)
          __declspec(align(dna_data));
        #elif defined(__INTEL_LLVM_COMPILER)
          void* vec_r = __builtin_assume_aligned(dna_data, 64);
        #endif

        int32_t rna_length = rna->length;
        auto strand        = rna->strand_;
        if (rna->is_init_) {
          int32_t s_pos = rna->begin;
          if (rna_length >= 21) {
            if (strand == Strand::LEADING) {
              s_pos += 22;
              s_pos = s_pos >= dna_length ? s_pos - dna_length : s_pos;
            } else {
              s_pos -= 22;
              s_pos = s_pos < 0 ? static_cast<int>(dna_length) + s_pos : s_pos;
            }

            #ifdef __VECTORIZE_STRCMP
              char buffer_str[13];
            #endif

            for (int32_t loop_size = 0; loop_size < rna_length - DO_TRANSLATION_LOOP; loop_size++) {
              #ifdef __VECTORIZE_STRCMP
                bool b_start   = false;
                bool start[14] = {
                    false, false, false, false, false, false, false, false, false, false, false, false, false, false};
              #else
                #if defined(__INTEL_COMPILER) || defined(__INTEL_LLVM_COMPILER)
                  __declspec(align(64)) bool start[14] = {
                      false, false, false, false, false, false, false, false, false, false, false, false, false, false};
                #else
                  bool start[14] = {
                      false, false, false, false, false, false, false, false, false, false, false, false, false, false};
                #endif
              #endif

              int32_t c_pos = s_pos;
              if (strand == Strand::LEADING) {
                c_pos += loop_size;
                c_pos = c_pos >= dna_length ? c_pos - dna_length : c_pos;
              } else {
                c_pos -= loop_size;
                c_pos = c_pos < 0 ? dna_length + c_pos : c_pos;
              }

              if (strand == Strand::LEADING) {
                // Search for Shine Dalgarro + START codon on LEADING
                if (c_pos + 15 < dna_length && c_pos + 15 >= 0) {
                  #ifdef __VECTORIZE_STRCMP
                    bool compare_1 = strncmp(&(dna_data[c_pos]), SHINE_DAL_SEQ_LEAD_7, 6) == 0 ? true : false;
                    bool compare_2 = strncmp(&(dna_data[c_pos + 10]), &(SHINE_DAL_SEQ_LEAD_7[10]), 3) == 0 ? true : false;
                    b_start        = compare_1 && compare_2;
                  #else
                    #if defined(__INTEL_COMPILER) || defined(__INTEL_LLVM_COMPILER)
                      #pragma omp simd aligned(dna_data : 64) aligned(SHINE_DAL_SEQ_LEAD_7 : 64) aligned(start : 64)
                    #else
                      #pragma omp simd
                    #endif
                    for (int32_t k = 0; k < 12; k++) {
                      start[k] = dna_data[c_pos + k] == SHINE_DAL_SEQ_LEAD_7[k];
                    }
                    start[12] = dna_data[c_pos + 12] == SHINE_DAL_SEQ_LEAD_7[12];
                  #endif
                } else {
                    for (int32_t k = 0; k < 9; k++) {
                      int32_t k_t   = k >= 6 ? k + 4 : k;
                      int32_t pos_m = c_pos + k_t;

                      while (pos_m < 0)
                        pos_m += dna_length;
                      while (pos_m >= dna_length)
                        pos_m -= dna_length;

                      start[k_t] = (dna_data[pos_m] == SHINE_DAL_SEQ_LEAD_7[k_t]) ? true : false;
                    }
                    #ifdef __VECTORIZE_STRCMP
                      b_start = start[0] && start[1] && start[2] && start[3] && start[4] && start[5] && start[10] &&
                                start[11] && start[12];
                    #endif
                }
              } else {
                // Search for Shine Dalgarro + START codon on LAGGING
                if (c_pos - 15 < dna_length && c_pos - 15 >= 0) {
                #ifdef __VECTORIZE_STRCMP
                  for (int i = 0; i < 13; i++)
                    buffer_str[i] = dna_data[c_pos - i];
                  bool compare_1 = strncmp(buffer_str, SHINE_DAL_SEQ_LAG_7, 6) == 0 ? true : false;
                  bool compare_2 = strncmp(&(buffer_str[10]), &(SHINE_DAL_SEQ_LAG_7[10]), 3) == 0 ? true : false;

                  b_start = compare_1 && compare_2;
                #else
                  #pragma omp simd
                  for (int32_t k = 0; k < 12; k++) {
                    start[k] = dna_data[c_pos - k] == SHINE_DAL_SEQ_LAG_7[k];
                  }
                  start[12] = dna_data[c_pos - 12] == SHINE_DAL_SEQ_LAG_7[12];
                #endif
                } else {
                  for (int32_t k = 0; k < 9; k++) {
                    int32_t k_t   = k >= 6 ? k + 4 : k;
                    int32_t pos_m = c_pos - k_t;

                    while (pos_m < 0)
                        pos_m += dna_length;
                    while (pos_m >= dna_length)
                        pos_m -= dna_length;

                    start[k_t] = (dna_data[pos_m] == SHINE_DAL_SEQ_LAG_7[k_t]) ? true : false;
                  }
                  #ifdef __VECTORIZE_STRCMP
                    b_start = start[0] && start[1] && start[2] && start[3] && start[4] && start[5] && start[10] &&
                          start[11] && start[12];
                  #endif
                }
              }

              #ifdef __VECTORIZE_STRCMP
              if (b_start) {
              #else
              if (start[0] && start[1] && start[2] && start[3] && start[4] && start[5] && start[10] && start[11] &&
                    start[12]) {
              #endif
                rna->start_prot_count_++;
                rna->start_prot.push_back(c_pos);
              }
            }
          }
        }
      }
    #elif BASE_4
      if (rna->is_init_ && rna->length >= DO_TRANSLATION_LOOP) {
        int c_pos = rna->begin;

        if (rna->strand_ == Strand::LEADING) {
          c_pos += PROM_SIZE;
          c_pos = c_pos >= dna_length ? c_pos - dna_length : c_pos;
        } else {
          c_pos -= PROM_SIZE;
          c_pos = c_pos < 0 ? dna_length + c_pos : c_pos;
        }

        int loop_size = 0;
        while (loop_size + DO_TRANSLATION_LOOP < rna->length) {
          bool start = false;

          if (rna->strand_ == Strand::LEADING) {
            // Search for Shine Dalgarro + START codon on LEADING
            start = annotated_chromosome_->metadata_->is_shine_dal_start_prot_leading(c_pos);
          } else {
            // Search for Shine Dalgarro + START codon on LAGGING
            start = annotated_chromosome_->metadata_->is_shine_dal_start_prot_lagging(c_pos);
          }

          if (start) {
            rna->start_prot_count_++;
            rna->start_prot.push_back(c_pos);
          }

          if (rna->strand_ == Strand::LEADING) {
            c_pos++;
            c_pos = c_pos >= dna_length ? c_pos - dna_length : c_pos;
          } else {
            c_pos--;
            c_pos = c_pos < 0 ? dna_length + c_pos : c_pos;
          }
          loop_size++;
        }
      }
    #endif
  }

  #ifdef WITH_PERF_TRACES_PER_INDIV
    auto runtime_end              = std::chrono::steady_clock::now();
    start_protein_stop_[indiv_id] = runtime_end.time_since_epoch().count();
    start_protein_[indiv_id]      = runtime_end.time_since_epoch().count() - runtime_start.time_since_epoch().count();
  #endif
}

void Individual_7::compute_protein() {
  #ifdef WITH_PERF_TRACES_PER_INDIV
    auto runtime_start               = std::chrono::steady_clock::now();
    compute_protein_start_[indiv_id] = runtime_start.time_since_epoch().count();
  #endif

  Dna_7* dna         = annotated_chromosome_->dna_;
  int32_t dna_length = dna->length();

  for (Rna_7* rna: annotated_chromosome_->rnas()) {
    if (rna->is_init_) {
      for (auto it_start_pos = rna->start_prot.begin(); it_start_pos != rna->start_prot.end(); it_start_pos++) {
        int start_prot = *it_start_pos;
        int start_protein_pos =
            rna->strand_ == Strand::LEADING ? start_prot + PROT_START_SIZE : start_prot - PROT_START_SIZE;
        int length;

        if (rna->strand_ == Strand::LEADING) {
          start_protein_pos = start_protein_pos >= dna_length ? start_protein_pos - dna_length : start_protein_pos;

          if (start_prot < rna->end) {
            length = rna->end - start_prot;
          } else {
            length = dna_length - start_prot + rna->end;
          }

          length -= PROT_START_SIZE;
        } else {
          start_protein_pos = start_protein_pos < 0 ? dna_length + start_protein_pos : start_protein_pos;

          if (start_prot > rna->end) {
            length = (*it_start_pos) - rna->end;
          } else {
            length = *it_start_pos + dna_length - rna->end;
          }

          length -= PROT_START_SIZE;
        }

        bool is_protein = false;

        length += 1;
        length = length - (length % CODON_SIZE);

        int j                     = 0;
        int32_t transcribed_start = 0;

        if (rna->strand_ == Strand::LEADING) {
          transcribed_start = rna->begin + PROM_SIZE;
          transcribed_start = transcribed_start >= dna_length ? transcribed_start - dna_length : transcribed_start;

          if (transcribed_start <= start_prot) {
            j = start_prot - transcribed_start;
          } else {
            j = dna_length - transcribed_start + start_prot;
          }
        } else {
          transcribed_start = rna->begin - PROM_SIZE;
          transcribed_start = transcribed_start < 0 ? dna_length + transcribed_start : transcribed_start;

          if (transcribed_start >= start_prot) {
            j = transcribed_start - start_prot;
          } else {
            j = transcribed_start + dna_length - start_prot;
          }
        }

        j += PROT_START_SIZE;

        while (rna->length - j >= CODON_SIZE) {
          int t_k;

          if (rna->strand_ == Strand::LEADING) {
            start_protein_pos = start_protein_pos >= dna_length ? start_protein_pos - dna_length : start_protein_pos;

            #ifdef BASE_2
              is_protein = false;

              for (int k = 0; k < 3; k++) {
                t_k = start_protein_pos + k;

                if (dna->get_lead(t_k) == PROTEIN_END_LEAD[k]) {
                  is_protein = true;
                } else {
                  is_protein = false;
                  break;
                }
              }
            #elif BASE_4
              is_protein = codon_value_to_aminoacid(bases_to_codon_value(dna->get_lead(start_protein_pos),
                                                                         dna->get_lead(start_protein_pos + 1),
                                                                         dna->get_lead(start_protein_pos + 2))) == STOP;

              t_k = start_protein_pos + CODON_SIZE;
            #endif

            if (is_protein) {
              int prot_length = -1;

              if (start_prot + PROT_START_SIZE < t_k) {
                prot_length = t_k - (start_prot + PROT_START_SIZE);
              } else {
                prot_length = dna_length - (start_prot + PROT_START_SIZE) + t_k;
              }

              if (prot_length >= CODON_SIZE) {
                #ifdef BASE_4
                int32_t prot_len = prot_length / CODON_SIZE;
                prot_len--;
                if (prot_len > 0) {
                #endif

                auto new_prot = new Protein_7(Utils::mod(start_prot + PROT_START_SIZE, dna_length),
                                              Utils::mod(t_k, dna_length),
                                              #ifdef BASE_4
                                                  prot_len,
                                              #else
                                                 prot_length / 3,
                                              #endif
                                              rna->strand_,
                                              rna->e,
                                              rna
                                              #ifdef __REGUL
                                                ,
                                                true
                                              #endif
                );
                #ifdef __REGUL
                  annotated_chromosome_->proteins_.add_translated_protein(new_prot);
                #else
                  annotated_chromosome_->proteins_.add(new_prot);
                #endif

                rna->is_coding_ = true;
                #ifdef BASE_4
                }
                #endif
              }

              break;
            }

            start_protein_pos += CODON_SIZE;
            start_protein_pos = start_protein_pos >= dna_length ? start_protein_pos - dna_length : start_protein_pos;
          } else {
            start_protein_pos = start_protein_pos < 0 ? dna_length + start_protein_pos : start_protein_pos;

            #ifdef BASE_2
              is_protein = false;
              for (int k = 0; k < 3; k++) {
                t_k = start_protein_pos - k;

                if (dna->get_lag(t_k) == PROTEIN_END_LAG[k]) {
                  is_protein = true;
                } else {
                  is_protein = false;
                  break;
                }
              }
            #elif BASE_4
              is_protein = codon_value_to_aminoacid(bases_to_codon_value(dna->get_lag(start_protein_pos),
                                                                         dna->get_lag(start_protein_pos - 1),
                                                                         dna->get_lag(start_protein_pos - 2))) == STOP;

              t_k = start_protein_pos - CODON_SIZE;
            #endif

            if (is_protein) {
              int prot_length = -1;
              if (start_prot - PROT_START_SIZE > t_k) {
                prot_length = (start_prot - PROT_START_SIZE) - t_k;
              } else {
                prot_length = (start_prot - PROT_START_SIZE) + dna_length - t_k;
              }
              if (prot_length >= CODON_SIZE) {
                #ifdef BASE_4
                int32_t prot_len = prot_length / CODON_SIZE;
                prot_len--;

                if (prot_len > 0) {
                #endif

                auto new_prot = new Protein_7(Utils::mod(start_prot - PROT_START_SIZE, dna_length),
                                              Utils::mod(t_k, dna_length),
                                              #ifdef BASE_4
                                                prot_len,
                                              #else
                                               prot_length / 3,
                                              #endif
                                                rna->strand_,
                                                rna->e,
                                                rna
                                              #ifdef __REGUL
                                                ,
                                                true
                                              #endif
                );
                #ifdef __REGUL
                  annotated_chromosome_->proteins_.add_translated_protein(new_prot);
                #else
                  annotated_chromosome_->proteins_.add(new_prot);
                #endif
                rna->is_coding_ = true;
                #ifdef BASE_4
                }
                #endif
              }
              break;
            }
            start_protein_pos = start_protein_pos - CODON_SIZE;
            start_protein_pos = start_protein_pos < 0 ? dna_length + start_protein_pos : start_protein_pos;
          }
          j += CODON_SIZE;
        }
      }
    }
  }

  #ifdef WITH_PERF_TRACES_PER_INDIV
    auto runtime_end                = std::chrono::steady_clock::now();
    compute_protein_stop_[indiv_id] = runtime_end.time_since_epoch().count();
    compute_protein_[indiv_id]      = runtime_end.time_since_epoch().count() - runtime_start.time_since_epoch().count();
  #endif
}

#ifdef BASE_4
void Individual_7::translate_protein(const ExpSetup& exp_s, double w_max) {
#else
void Individual_7::translate_protein(const ExpSetup&, double w_max) {
#endif
  #ifdef WITH_PERF_TRACES_PER_INDIV
    auto runtime_start = std::chrono::steady_clock::now();
    translate_protein_start_[indiv_id] = runtime_start.time_since_epoch().count();
  #endif

  int32_t dna_length = annotated_chromosome_->dna_->length();

  for (const auto& prot: annotated_chromosome_->proteins()) {
  if (prot->is_init_) {
    int c_pos = prot->protein_start;
    #ifdef BASE_2
      int t_pos;
    #endif
    int end_pos = prot->protein_end;
    if (prot->strand_ == Strand::LEADING) {
      end_pos -= 3;

      c_pos   = c_pos >= dna_length ? c_pos - dna_length : c_pos;
      end_pos = end_pos < 0 ? dna_length + end_pos : end_pos;
    } else {
      end_pos += 3;

      end_pos = end_pos >= dna_length ? end_pos - dna_length : end_pos;
      c_pos   = c_pos < 0 ? dna_length + c_pos : c_pos;
    }

    int8_t value       = 0;
    int16_t codon_idx  = 0;
    int32_t count_loop = 0;

    #ifdef BASE_4
      char base_1 = -1;
      char base_2 = -1;
      char base_3 = -1;
    #endif

    if (prot->strand_ == Strand::LEADING) {
      #ifdef BASE_2
      while (count_loop < prot->protein_length && codon_idx < 64 * 3) {
        value = 0;
        for (int8_t i = 0; i < 3; i++) {
        t_pos = c_pos + i;
        if (annotated_chromosome_->dna_->get_lead(t_pos) == '1')
          value += 1 << (CODON_SIZE - i - 1);
        }

        prot->codon_list[codon_idx] = value;
      #elif BASE_4
      while (count_loop < prot->protein_length) {
        base_1 = annotated_chromosome_->dna_->get_lead(c_pos);
        base_2 = annotated_chromosome_->dna_->get_lead(c_pos + 1);
        base_3 = annotated_chromosome_->dna_->get_lead(c_pos + 2);

        value = bases_to_codon_value(base_1, base_2, base_3);
        prot->codon_list.push_back(value);
      #endif

        codon_idx++;

        count_loop++;
        c_pos += CODON_SIZE;
        c_pos = c_pos >= dna_length ? c_pos - dna_length : c_pos;
      }
    } else { // LAGGING
      #ifdef BASE_2
      while (count_loop < prot->protein_length && codon_idx < 64 * 3) {
        value = 0;
        for (int8_t i = 0; i < 3; i++) {
          t_pos = c_pos - i;
          if (annotated_chromosome_->dna_->get_lag(t_pos) != '1')
            value += 1 << (CODON_SIZE - i - 1);
        }

        prot->codon_list[codon_idx] = value;

      #elif BASE_4
      while (count_loop < prot->protein_length) {

        // get complementary base on lagging strain
        base_1 = get_complementary_base(annotated_chromosome_->dna_->get_lead(c_pos));
        base_2 = get_complementary_base(annotated_chromosome_->dna_->get_lead(c_pos - 1));
        base_3 = get_complementary_base(annotated_chromosome_->dna_->get_lead(c_pos - 2));

        value = bases_to_codon_value(base_1, base_2, base_3);
        prot->codon_list.push_back(value);
      #endif

      codon_idx++;

      count_loop++;

      c_pos -= CODON_SIZE;
      c_pos = c_pos < 0 ? c_pos + dna_length : c_pos;
    }
  }

  #ifdef BASE_2
    if (codon_idx >= 64 * 3) {
      std::ofstream last_gener_file("aevol_run.log", std::ofstream::out);
      last_gener_file << "Stop translating protein before end (length is greater than 196" << std::endl;
      std::cout << "Stop translating protein before end (length is greater than 196" << std::endl;
      last_gener_file.close();
    }

    double M = 0.0;
    double W = 0.0;
    double H = 0.0;
    #elif BASE_4
      //  --------------------------------
      //  1) Compute values for M, W and H
      //  --------------------------------
      auto base_m = exp_s.aa_base_m();
      auto base_w = exp_s.aa_base_w();
      auto base_h = exp_s.aa_base_h();

      int8_t base_m_size = exp_s.aa_base_m_size();
      int8_t base_w_size = exp_s.aa_base_w_size();
      int8_t base_h_size = exp_s.aa_base_h_size();

      long double M = 0.0;
      long double W = 0.0;
      long double H = 0.0;

      prot->nb_codons_ = codon_idx;
    #endif

    int32_t nb_m = 0;
    int32_t nb_w = 0;
    int32_t nb_h = 0;

    #ifdef BASE_2
      bool bin_m = false;  // Initializing to false will yield a conservation of the high weight bit
      bool bin_w = false;  // when applying the XOR operator for the Gray to standard conversion
      bool bin_h = false;

      prot->nb_codons_ = codon_idx - 1;

      for (int i = 0; i < codon_idx; i++) {
        switch (prot->codon_list[i]) {
          case CODON_M0: {
            // M codon found
            nb_m++;

            // Convert Gray code to "standard" binary code
            bin_m ^= false;  // as bin_m was initialized to false, the XOR will have no effect on the high weight bit

            // A lower-than-the-previous-lowest weight bit was found, make a left bitwise shift
            //~ M <<= 1;
            M *= 2;

            // Add this nucleotide's contribution to M
            if (bin_m)
              M += 1;

            break;
          }
          case CODON_M1: {
            // M codon found
            nb_m++;

            // Convert Gray code to "standard" binary code
            bin_m ^= true;  // as bin_m was initialized to false, the XOR will have no effect on the high weight bit

            // A lower-than-the-previous-lowest bit was found, make a left bitwise shift
            //~ M <<= 1;
            M *= 2;

            // Add this nucleotide's contribution to M
            if (bin_m)
              M += 1;

            break;
          }
          case CODON_W0: {
            // W codon found
            nb_w++;

            // Convert Gray code to "standard" binary code
            bin_w ^= false;  // as bin_m was initialized to false, the XOR will have no effect on the high weight bit

            // A lower-than-the-previous-lowest weight bit was found, make a left bitwise shift
            //~ W <<= 1;
            W *= 2;

            // Add this nucleotide's contribution to W
            if (bin_w)
              W += 1;

            break;
          }
          case CODON_W1: {
            // W codon found
            nb_w++;

            // Convert Gray code to "standard" binary code
            bin_w ^= true;  // as bin_m was initialized to false, the XOR will have no effect on the high weight bit

            // A lower-than-the-previous-lowest weight bit was found, make a left bitwise shift
            //~ W <<= 1;
            W *= 2;

            // Add this nucleotide's contribution to W
            if (bin_w)
              W += 1;

            break;
          }
          case CODON_H0:
          case CODON_START:  // Start codon codes for the same amino-acid as H0 codon
          {
            // H codon found
            nb_h++;

            // Convert Gray code to "standard" binary code
            bin_h ^= false;  // as bin_m was initialized to false, the XOR will have no effect on the high weight bit

            // A lower-than-the-previous-lowest weight bit was found, make a left bitwise shift
            //~ H <<= 1;
            H *= 2;

            // Add this nucleotide's contribution to H
            if (bin_h)
              H += 1;

            break;
          }
          case CODON_H1: {
              // H codon found
              nb_h++;

              // Convert Gray code to "standard" binary code
              bin_h ^= true;  // as bin_m was initialized to false, the XOR will have no effect on the high weight bit

              // A lower-than-the-previous-lowest weight bit was found, make a left bitwise shift
              //~ H <<= 1;
              H *= 2;

              // Add this nucleotide's contribution to H
              if (bin_h)
                H += 1;

              break;
          }
        }
      }
      #elif BASE_4
        double m_base_exponent = 1.25;
        double w_base_exponent = 1.25;
        double h_base_exponent = 1.25;

        double m_digit_factor = 1.0;
        double w_digit_factor = 1.0;
        double h_digit_factor = 1.0;

        for (int i = 0; i < prot->nb_codons_; i++) {
          AminoAcid amino_acid = codon_value_to_aminoacid(prot->codon_list.at(prot->nb_codons_ - i - 1));

          if (base_m[amino_acid] != -1) {
            M += base_m[amino_acid] * m_digit_factor;
            m_digit_factor *= m_base_exponent;  //base_m_size;
            nb_m++;
          }

          if (base_w[amino_acid] != -1) {
            W += base_w[amino_acid] * w_digit_factor;
            w_digit_factor *= w_base_exponent;  //base_w_size;
            nb_w++;
          }

          if (base_h[amino_acid] != -1) {
            H += base_h[amino_acid] * h_digit_factor;
            h_digit_factor *= h_base_exponent;  //base_h_size;
            nb_h++;
          }
        }
      #endif

      //  ----------------------------------------------------------------------------------
      //  2) Normalize M, W and H values in [0;1] according to number of codons of each kind
      //  ----------------------------------------------------------------------------------
      #ifdef BASE_2
        prot->m = nb_m != 0 ? M / (pow(2, nb_m) - 1) : 0.5;
        prot->w = nb_w != 0 ? W / (pow(2, nb_w) - 1) : 0.0;
        prot->h = nb_h != 0 ? H / (pow(2, nb_h) - 1) : 0.5;
      #elif BASE_4
        /*
        if (nb_m != 0) {
          prot->m = M / (pow(base_m_size, nb_m) - 1);
        } else {
          prot->m = 0.5;
        }

        if (nb_w != 0) {
          prot->w = W / (pow(base_w_size, nb_w) - 1);
        }
        else {
          prot->w = 0.0;
        }

        if (nb_h != 0) {
          prot->h = H / (pow(base_h_size, nb_h) - 1);
        } else {
          prot->h = 0.5;
          }*/

        if (nb_m != 0) {
          prot->m = M / ((base_m_size) * (pow(m_base_exponent, nb_m) - 1) / (m_base_exponent - 1));
        } else {
          prot->m = 0.5;
        }

        if (nb_w != 0) {
          prot->w = W / ((base_w_size) * (pow(w_base_exponent, nb_w) - 1) / (w_base_exponent - 1));
        } else {
          prot->w = 0.0;
        }

        if (nb_h != 0) {
          prot->h = H / ((base_h_size) * (pow(h_base_exponent, nb_h) - 1) / (h_base_exponent - 1));
        } else {
          prot->h = 0.5;
        }
      #endif

      //  ------------------------------------------------------------------------------------
      //  3) Normalize M, W and H values according to the allowed ranges (defined in macros.h)
      //  ------------------------------------------------------------------------------------
      // x_min <= M <= x_max
      // w_min <= W <= w_max
      // h_min <= H <= h_max
      prot->m = (X_MAX - X_MIN) * prot->m + X_MIN;
      prot->w = (w_max - W_MIN) * prot->w + W_MIN;
      prot->h = (H_MAX - H_MIN) * prot->h + H_MIN;

      if (nb_m == 0 || nb_w == 0 || nb_h == 0 || prot->w == 0.0 || prot->h == 0.0) {
        prot->is_functional = false;
      } else {
        prot->is_functional = true;
      }
    }
  }

  std::map<int32_t, Protein_7*> lookup;

  for (const auto& prot: annotated_chromosome_->proteins()) {
    if (prot->is_init_ && prot->strand_ == Strand::LEADING) {
      if (lookup.find(prot->protein_start) == lookup.end()) {
        lookup[prot->protein_start] = prot;
      } else {
        #ifdef __MULTI_PROMOTERS
          #if 0 == __MULTI_PROMOTERS
            if (prot->e > lookup[prot->protein_start]->e)
              lookup[prot->protein_start]->e = prot->e;
            if (prot->initial_e_ > lookup[prot->protein_start]->initial_e_)
              lookup[prot->protein_start]->initial_e_ = prot->initial_e_;
          #endif
        #else
          lookup[prot->protein_start]->e += prot->e;
          lookup[prot->protein_start]->initial_e_ += prot->initial_e_;
        #endif
        lookup[prot->protein_start]->rna_list_.insert(
            lookup[prot->protein_start]->rna_list_.end(), prot->rna_list_.begin(), prot->rna_list_.end());
        prot->is_init_ = false;
      }
    }
  }

  lookup.clear();

  for (const auto& prot: annotated_chromosome_->proteins()) {
    if (prot->is_init_ && prot->strand_ == Strand::LAGGING) {
      if (lookup.find(prot->protein_start) == lookup.end()) {
        lookup[prot->protein_start] = prot;
      } else {
        #ifdef __MULTI_PROMOTERS
          #if 0 == __MULTI_PROMOTERS
            if (prot->e > lookup[prot->protein_start]->e)
              lookup[prot->protein_start]->e = prot->e;
            if (prot->initial_e_ > lookup[prot->protein_start]->initial_e_)
              lookup[prot->protein_start]->initial_e_ = prot->initial_e_;
          #endif
        #else
          lookup[prot->protein_start]->e += prot->e;
          lookup[prot->protein_start]->initial_e_ += prot->initial_e_;
        #endif
        lookup[prot->protein_start]->rna_list_.insert(
            lookup[prot->protein_start]->rna_list_.end(), prot->rna_list_.begin(), prot->rna_list_.end());
        prot->is_init_ = false;
      }
    }
  }

  #ifdef __MULTI_PROMOTERS
    #if 1 == __MULTI_PROMOTERS
      for (const auto& prot: annotated_chromosome_->proteins()) {
        if (prot->is_init_) {
          Rna_7* first_rna   = *(prot->rna_list_.begin());
          int32_t max_length = -1;

          if (prot->strand_ == Strand::LAGGING) {
            max_length = prot->protein_start > first_rna->begin
                             ? prot->protein_start - first_rna->begin
                             : prot->protein_start + (annotated_chromosome_->dna_->length_ - first_rna->begin);
          } else {
            max_length = prot->protein_start < first_rna->begin
                             ? first_rna->begin - prot->protein_start
                             : first_rna->begin + (annotated_chromosome_->dna_->length_ - prot->protein_start);
          }

          for (auto rna: prot->rna_list_) {
            int32_t rna_length = -1;

            if (prot->strand_ == Strand::LAGGING) {
              max_length = prot->protein_start > rna->begin
                               ? prot->protein_start - rna->begin
                               : prot->protein_start + (annotated_chromosome_->dna_->length_ - rna->begin);
            } else {
              max_length = prot->protein_start < rna->begin
                               ? rna->begin - prot->protein_start
                               : rna->begin + (annotated_chromosome_->dna_->length_ - prot->protein_start);
            }

            if (rna_length > max_length) {
              first_rna  = rna;
              max_length = rna_length;
            }
          }

          prot->e          = first_rna->e;
          prot->initial_e_ = first_rna->e;
        }
      }
    #endif
  #endif

  #ifdef WITH_PERF_TRACES_PER_INDIV
    auto runtime_end                  = std::chrono::steady_clock::now();
    translate_protein_stop_[indiv_id] = runtime_end.time_since_epoch().count();
    translate_protein_[indiv_id] = runtime_end.time_since_epoch().count() - runtime_start.time_since_epoch().count();
  #endif
}

void Individual_7::compute_phenotype() {
  #ifdef WITH_PERF_TRACES_PER_INDIV
    auto runtime_start = std::chrono::steady_clock::now();
    compute_protein_start_[indiv_id]= runtime_start.time_since_epoch().count();
  #endif

    AbstractFuzzy* activ_phenotype = fuzzy_factory_->get_fuzzy();
    AbstractFuzzy* inhib_phenotype = fuzzy_factory_->get_fuzzy();


    std::vector<Protein_7*> protein_vector;
    for (const auto& prot : annotated_chromosome_->proteins()) {
      if (prot->is_init_) {
        protein_vector.push_back(prot);
      }
    }

    std::sort(protein_vector.begin(), protein_vector.end(), [](Protein_7* a, Protein_7* b) { return *a < *b; });

    for (auto prot : protein_vector) {
    if (fabs(prot->w) >= 1e-15 && fabs(prot->h) >= 1e-15) {
      if (prot->is_functional) {
        bool verbose = false;
        if (prot->h > 0)
          activ_phenotype->add_triangle(prot->m, prot->w, prot->h * prot->e, verbose);
        else
          inhib_phenotype->add_triangle(prot->m, prot->w, prot->h * prot->e, verbose);
      }
    }
  }

  activ_phenotype->clip(AbstractFuzzy::max,  Y_MAX);
  inhib_phenotype->clip(AbstractFuzzy::min, -Y_MAX);

  activ_phenotype->simplify();
  inhib_phenotype->simplify();

  if (phenotype != nullptr) {
    fuzzy_factory_->give_back(phenotype);
    phenotype = nullptr;
  }

  phenotype = fuzzy_factory_->get_fuzzy();
  phenotype->copy(activ_phenotype);
  phenotype->add(inhib_phenotype);
  phenotype->clip(AbstractFuzzy::min, Y_MIN);
  phenotype->simplify();
  fuzzy_factory_->give_back(activ_phenotype);
  fuzzy_factory_->give_back(inhib_phenotype);

  #ifdef WITH_PERF_TRACES_PER_INDIV
    auto runtime_end                  = std::chrono::steady_clock::now();
    compute_phenotype_stop_[indiv_id] = runtime_end.time_since_epoch().count();
    compute_phenotype_[indiv_id] =
        runtime_end.time_since_epoch().count() - runtime_start.time_since_epoch().count();
  #endif
}

#ifdef __REGUL
void Individual_7::compute_fitness(double selection_pressure,
                                   int env_id,
                                   SIMD_PhenotypicTargetHandler_R* pth) {
#else
void Individual_7::compute_fitness(double selection_pressure,
                                   const AbstractFuzzy* target) {
#endif
  #ifdef WITH_PERF_TRACES_PER_INDIV
    auto runtime_start = std::chrono::steady_clock::now();
    compute_fitness_start_[indiv_id] = runtime_start.time_since_epoch().count();
  #endif
  #ifdef __REGUL
    AbstractFuzzy* delta = fuzzy_factory_->get_fuzzy();
    delta->copy(phenotype);

    if (pth->var_method_ == SWITCH_IN_A_LIST)
    delta->sub(pth->targets_fuzzy_by_id_[env_id]);
    else if (pth->var_method_ == ONE_AFTER_ANOTHER)
    delta->sub(pth->targets_fuzzy_by_id_[env_id]);

    if (pth->var_method_ == SWITCH_IN_A_LIST)
    metaerror_by_env_id_[env_id] = delta->get_geometric_area();
    else if (pth->var_method_ == ONE_AFTER_ANOTHER)
    metaerror_by_env_id_[env_id] += delta->get_geometric_area();

    delete delta;

    if (pth->var_method_ == SWITCH_IN_A_LIST)
    fitness_by_env_id_[env_id] = exp(-selection_pressure * static_cast<double>(metaerror_by_env_id_[env_id]));
  #else
    AbstractFuzzy* delta = fuzzy_factory_->get_fuzzy();

    delta->copy(phenotype);
    bool verbose = false;
    delta->sub(target, verbose);

    metaerror = delta->get_geometric_area();

    fuzzy_factory_->give_back(delta);

    fitness_ = exp(-selection_pressure * static_cast<double>(metaerror));
  #endif

  #ifdef WITH_PERF_TRACES_PER_INDIV
    auto runtime_end = std::chrono::steady_clock::now();
    compute_fitness_stop_[indiv_id] = runtime_end.time_since_epoch().count();
    compute_fitness_[indiv_id] = runtime_end.time_since_epoch().count() - runtime_start.time_since_epoch().count();
  #endif
}

#ifdef __REGUL
void Individual_7::compute_network(SIMD_PhenotypicTargetHandler_R* pth) {
  // Add signals
  annotated_chromosome_->proteins_.add_signal_proteins(pth->signals_models_);

  for (Rna_7* rna : annotated_chromosome_->rnas()) {
    if (rna != nullptr) {
      if (rna->is_coding_) {
        rna->affinity_list.clear();
      }
    }
  }

  for (const auto& prot : annotated_chromosome_->proteins()) {
    if (prot != nullptr) {
      for (auto&& rna: prot->rna_list_) {
        rna->affinity_list.clear();
      }
    }
  }

  // Set influences
  for (Rna_7* rna : annotated_chromosome_->rnas()) {
    if (rna != nullptr) {
      if (rna->is_coding_) {
        rna->nb_influences_ = 0;

        int32_t enhancer_position = rna->enhancer_position(annotated_chromosome_->dna_->length());
        int32_t operator_position = rna->operator_position(annotated_chromosome_->dna_->length());

        double enhance,operate;
        for (const auto& prot : annotated_chromosome_->proteins()) {
          if (prot != nullptr) {
            if (prot->is_init_ || prot->signal_) {
              enhance = rna->affinity_with_protein(enhancer_position, prot, this);
              operate = rna->affinity_with_protein(operator_position, prot, this);

              if (enhance != 0.0 || operate != 0.0) {
                rna->affinity_list.push_back(AffinityFactor(prot, enhance, operate));
                prot->is_TF_ = true;

                rna->nb_influences_++;
              }
            }
          }
        }
      }
    }
  }
}

void Individual_7::update_network(const ExpSetup& exp_s) {
  for (const auto& prot : annotated_chromosome_->proteins()) {
    if (!prot->signal_) {
      if (prot->is_init_) {
        prot->delta_concentration_ = 0;

        for (auto rna: prot->rna_list_) {
          double synthesis_rate = rna->compute_synthesis_rate();
          prot->delta_concentration_ += synthesis_rate;
        }

        prot->delta_concentration_ -= exp_s.get_degradation_rate() * prot->e;
        prot->delta_concentration_ *= 1.0 / exp_s.get_nb_degradation_step();
      }
    }
  }


  // Apply the changes in concentrations we have just computed
  for (const auto& prot : annotated_chromosome_->proteins()) {
    if (!prot->signal_) {
      if (prot->is_init_) {
        prot->e += prot->delta_concentration_;
      }
    }
  }
}

void Individual_7::evaluate_network(double selection_pressure, int env_id, SIMD_PhenotypicTargetHandler_R* pth) {
  update_phenotype();
  compute_fitness(selection_pressure, env_id, pth);

  if (pth->var_method_ == SWITCH_IN_A_LIST)
    metaerror += metaerror_by_env_id_[env_id];
}

void Individual_7::finalize_network(double selection_pressure, SIMD_PhenotypicTargetHandler_R* pth) {
  if (pth->var_method_ == SWITCH_IN_A_LIST) {
    metaerror = metaerror / static_cast<double>(pth->nb_eval_);
  } else if (pth->var_method_ == ONE_AFTER_ANOTHER) {
    for (int env_id = 0; env_id < pth->nb_eval_; ++env_id) {
      metaerror_by_env_id_[env_id] = metaerror_by_env_id_[env_id] / 10.0;
      metaerror += metaerror_by_env_id_[env_id];

      fitness_by_env_id_[env_id] = exp(-selection_pressure * static_cast<double>(metaerror_by_env_id_[env_id]));
    }
    metaerror = metaerror / static_cast<double>(pth->nb_env_);
  }

  fitness_ = exp(-selection_pressure * static_cast<double>(metaerror));
}

void Individual_7::solve_network(
    const ExpSetup& exp_s,
    double selection_pressure,
    SIMD_PhenotypicTargetHandler_R* pth) {
  // Allocate fitness and metarror array
  if (pth->var_method_ == SWITCH_IN_A_LIST) {
    fitness_by_env_id_ = new double[pth->nb_indiv_age_];
    metaerror_by_env_id_ = new double[pth->nb_indiv_age_];
  } else if (pth->var_method_ == ONE_AFTER_ANOTHER) {
    fitness_by_env_id_ = new double[pth->nb_env_];
    metaerror_by_env_id_ = new double[pth->nb_env_];

    for (int env_id = 0; env_id < pth->nb_env_; ++env_id) {
      fitness_by_env_id_[env_id] = 0;
      metaerror_by_env_id_[env_id] = 0;
    }
  }

  for (const auto& prot : annotated_chromosome_->proteins()) {
    if (!prot->signal_) {
      if (prot->is_init_) {
        prot->e = prot->initial_e_;
      }
    }
  }

  metaerror = 0;
  compute_network(pth);
  metaerror = 0;

  if (pth->var_method_ == ONE_AFTER_ANOTHER) {
    for (int16_t env_i = 0; env_i < pth->nb_env_; env_i++) {
      // Set the concentration of signals for this age
      annotated_chromosome_->proteins_.reset_signal_proteins_concentration();
      annotated_chromosome_->proteins_.set_signal_proteins_concentration(pth->env_signals_list_[env_i], 0.9);

      // TODO REINIT PROT ?

      for (int16_t i = 0; i < 10; i++) {
        for (int j = 0; j < 10; j++) {
          update_network(exp_s);
        }

        // If we have to evaluate the individual at this age
        evaluate_network(selection_pressure, env_i, pth);
      }
    }

    finalize_network(selection_pressure, pth);
  } else {
    std::set<int>* eval = exp_s.get_list_eval_step();
    // i is thus the age of the individual
    int32_t prev_env_id = pth->list_env_id_[0];
    for (int16_t i = 0; i < pth->nb_indiv_age_; i++) {
      // Set the concentration of signals for this age
      annotated_chromosome_->proteins_.reset_signal_proteins_concentration();
      annotated_chromosome_->proteins_.set_signal_proteins_concentration(pth->env_signals_list_[pth->list_env_id_[i]],
                                                                         0.9);

      if (pth->env_switch_probability_ < 0) {
        if (prev_env_id != pth->list_env_id_[i]) {
          for (const auto& prot : annotated_chromosome_->proteins()) {
            if (!prot->signal_) {
              if (prot->is_init_) {
                prot->e = prot->initial_e_;
              }
            }
          }
        }

        prev_env_id = pth->list_env_id_[i];
      }

      for (int j = 0; j < exp_s.get_nb_degradation_step(); j++) {
        update_network(exp_s);
      }

      // If we have to evaluate the individual at this age
      if (eval->find(i+1) != eval->end()) {
        evaluate_network(selection_pressure, pth->list_env_id_[i], pth);
      }
    }

    finalize_network(selection_pressure, pth);
  }

  if (pth->var_method_ == SWITCH_IN_A_LIST) {
    delete [] fitness_by_env_id_;
    fitness_by_env_id_ = nullptr;
    delete [] metaerror_by_env_id_;
    metaerror_by_env_id_ = nullptr;
  } else if (pth->var_method_ == ONE_AFTER_ANOTHER) {
    delete [] fitness_by_env_id_;
    delete [] metaerror_by_env_id_;
  }
}

void Individual_7::recompute_network(
    const ExpSetup& exp_s,
    double selection_pressure,
    SIMD_PhenotypicTargetHandler_R* pth) {
  // Allocate fitness and metarror array
  if (pth->var_method_ == SWITCH_IN_A_LIST) {
    fitness_by_env_id_ = new double[pth->nb_indiv_age_];
    metaerror_by_env_id_ = new double[pth->nb_indiv_age_];
  } else if (pth->var_method_ == ONE_AFTER_ANOTHER) {
    fitness_by_env_id_ = new double[pth->nb_env_];
    metaerror_by_env_id_ = new double[pth->nb_env_];

    for (int env_id = 0; env_id < pth->nb_env_; env_id++) {
      fitness_by_env_id_[env_id] = 0;
      metaerror_by_env_id_[env_id] = 0;
    }
  }

  for (const auto& prot : annotated_chromosome_->proteins()) {
    if (!prot->signal_) {
      if (prot->is_init_) {
        prot->e = prot->initial_e_;
      }
    }
  }

  metaerror = 0;

  if (pth->var_method_ == ONE_AFTER_ANOTHER) {
    for (int16_t env_i = 0; env_i < pth->nb_env_; env_i++) {
      // Set the concentration of signals for this age
      annotated_chromosome_->proteins_.reset_signal_proteins_concentration();
      annotated_chromosome_->proteins_.set_signal_proteins_concentration(pth->env_signals_list_[env_i], 0.9);

      for (int16_t i = 0; i < 10; i++) {
        for (int j = 0; j < 10; j++) {
          update_network(exp_s);
        }

        // If we have to evaluate the individual at this age
        evaluate_network(selection_pressure, env_i, pth);
      }
    }

    finalize_network(selection_pressure, pth);
  } else {
    std::set<int>* eval = exp_s.get_list_eval_step();
    // i is thus the age of the individual
    for (int16_t i = 0; i < pth->nb_indiv_age_; i++) {
      // Set the concentration of signals for this age
      annotated_chromosome_->proteins_.reset_signal_proteins_concentration();
      annotated_chromosome_->proteins_.set_signal_proteins_concentration(pth->env_signals_list_[pth->list_env_id_[i]],
                                                                         0.9);

      for (int j = 0; j < exp_s.get_nb_degradation_step(); j++) {
        update_network(exp_s);
      }

      if (eval->find(i+1) != eval->end()) {
        evaluate_network(selection_pressure, pth->list_env_id_[i], pth);
      }
    }

    finalize_network(selection_pressure, pth);
  }

  if (pth->var_method_ == SWITCH_IN_A_LIST) {
    delete [] fitness_by_env_id_;
    delete [] metaerror_by_env_id_;
  } else if (pth->var_method_ == ONE_AFTER_ANOTHER) {
    delete [] fitness_by_env_id_;
    delete [] metaerror_by_env_id_;
  }
}

void Individual_7::update_phenotype() {
  if (phenotype != nullptr) {
    fuzzy_factory_->give_back(phenotype);
    phenotype = nullptr;
  }

  compute_phenotype();
}
#endif

void Individual_7::compute_non_coding() {
  annotated_chromosome_->compute_non_coding();
}


}
