// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************


#ifndef AEVOL_INDIVIDUAL_7_H_
#define AEVOL_INDIVIDUAL_7_H_

#include <vector>
#include <map>
#include <set>

#include "DnaFactory.h"
#include "Observable.h"
#include "PhenotypicTargetHandler.h"
#include "7/fuzzy/Vector_Fuzzy.h"
#include "7/fuzzy/FuzzyFactory_7.h"
#include "ae_enums.h"
#include "AnnotatedChromosome.h"

#ifdef __REGUL
  #include "raevol/SIMD_PhenotypicTargetHandler_R.h"
#endif

namespace aevol {

class List_Metadata;
class Promoter;
class Rna_7;

class Individual_7 : public Observable {
 public:
  Individual_7() = delete;
  Individual_7(const Individual_7&) = delete;
  Individual_7(Individual_7&&) = delete;
  Individual_7& operator=(const Individual_7&) = delete;
  Individual_7& operator=(Individual_7&&) = delete;

 private:
  Individual_7(double w_max, FuzzyFactory_7* fuzzy_factory);

 public:
  static Individual_7* make_empty(double w_max, FuzzyFactory_7* fuzzy_factory);
  static Individual_7* make_from_sequence(double w_max,
                                          FuzzyFactory_7* fuzzy_factory,
                                          const char* seq,
                                          Dna_7::size_type len);

  #ifdef __REGUL
    static Individual_7* make_clone(Individual_7* orig,
                                    FuzzyFactory_7* fuzzy_factory,
                                    bool no_metadata = false);
  #else
    static Individual_7* make_clone(Individual_7* orig,
                                    FuzzyFactory_7* fuzzy_factory);
  #endif

  ~Individual_7();

  double fitness() const { return fitness_; };

  #ifdef __REGUL
    void evaluate(const ExpSetup& exp_s,
                  double w_max,
                  double selection_pressure,
                  SIMD_PhenotypicTargetHandler_R* pth = nullptr);
  #else
    void evaluate(const ExpSetup& exp_s,
                  double w_max,
                  double selection_pressure,
                  const AbstractFuzzy* target);
  #endif
  void start_stop_RNA();
  void prom_compute_RNA(const ExpSetup& exp_s);
  void start_protein();
  void compute_protein();
  void translate_protein(const ExpSetup& exp_s, double w_max);
  void compute_phenotype();
  #ifdef __REGUL
    void compute_fitness(double selection_pressure,
                         int env_id,
                         SIMD_PhenotypicTargetHandler_R* pth);
  #else
    void compute_fitness(double selection_pressure,
                         const AbstractFuzzy* target);
  #endif
  #ifdef __REGUL
    void compute_network(SIMD_PhenotypicTargetHandler_R* pth);
    void update_network(const ExpSetup& exp_s);
    void evaluate_network(double selection_pressure, int env_id, SIMD_PhenotypicTargetHandler_R* pth);
    void finalize_network(double selection_pressure, SIMD_PhenotypicTargetHandler_R* pth);
    void solve_network(const ExpSetup& exp_s, double selection_pressure, SIMD_PhenotypicTargetHandler_R* pth);
    void recompute_network(const ExpSetup& exp_s, double selection_pressure, SIMD_PhenotypicTargetHandler_R* pth);
    void update_phenotype();
  #endif
  void compute_non_coding();

#ifdef PHENOTYPE_VECTOR
  double phenotype[PHENOTYPE_VECTOR_SIZE];
  double delta[PHENOTYPE_VECTOR_SIZE];
#else
  AbstractFuzzy* phenotype = nullptr;
  AbstractFuzzy* delta = nullptr;
#endif
  double fitness_;
  double metaerror;


  double* fitness_by_env_id_;
  double* metaerror_by_env_id_;

  AnnotatedChromosome* annotated_chromosome_ = nullptr;

  int32_t parent_id;
  int32_t last_id;
  bool first_to_add = true;
  int32_t added_id = -1;

  int32_t usage_count_ = 1;

  double w_max_;

  FuzzyFactory_7* fuzzy_factory_;
};

} // namespace aevol

#endif // AEVOL_INDIVIDUAL_7_H_
