// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************

#include "ProteinList.h"

#include <cassert>

#include "legacy/ExpManager.h"

namespace aevol {

ProteinList::~ProteinList() {
  #ifdef __REGUL
    for (auto& protein : inherited_proteins_) {
      delete protein;
    }
    signal_proteins_.clear();
  #endif
  clear();
}

void ProteinList::add(Protein_7* prot) {
  proteins_.push_front(prot);
}

void ProteinList::clear() {
  for (auto& protein : proteins_) {
    delete protein;
  }
  proteins_.clear();
}

#ifdef __REGUL
void ProteinList::add_inherited_proteins() {
  if (! inherited_added_) {
    for (auto prot : inherited_proteins_) {
      if (! prot->signal_) {
        add(Protein_7::make_inherited_protein(prot));
      }
    }
    inherited_added_ = true;
  }
}

// Add signal proteins to both the specific list of signal proteins and the main list of all the proteins
// Warning: as of now, this function is meant to *replace* (not *add*) the existing signals with the provided ones
void ProteinList::add_signal_proteins(std::vector<Protein_7*> signal_models) {
  signal_proteins_.resize(signal_models.size());
  size_t i = 0;
  for (const auto& signal : signal_models) {
    Protein_7* protein = Protein_7::make_signal_protein(signal);
    add(protein);
    signal_proteins_[i] = protein;
    ++i;
  }
}

/// This function is intended to be used when an existing object should be recycled i.e. reused to store its own
/// offspring instead of recreating a new object.
/// This is done by removing all the proteins that do not originate from the transcription-translation process
/// (as of now these would be signal proteins and inherited proteins) of the original individual (the parent).
/// The remaining proteins (translated from the parent's genome) are transferred into the inherited_proteins_ list
void ProteinList::reinit_proteins_inherited() {
  // Clear inherited_proteins_
  for (auto protein : inherited_proteins_) {
    delete protein;
  }
  inherited_proteins_.clear();

  // Populate inherited_proteins_ with all the translated proteins
  for (auto prot : translated_proteins_) {
    if (prot!=nullptr && prot->is_init_)
      if (prot->e > exp_manager->exp_s()->get_protein_presence_limit()) {
        inherited_proteins_.push_back(Protein_7::make_inherited_protein(prot));
      }
  }

  // Delete from the main protein list all the proteins that were not translated
  // (signal proteins and inherited proteins)
  for (Protein_7* protein : proteins_) {
    if (not protein->translated_)
      delete protein;
  }

  // Clear the main protein list
  proteins_.clear();

  // Clear affinity of all the translated proteins and add them to the main protein list
  for (auto& prot : translated_proteins_) {
    if (prot!=nullptr && prot->is_init_) {
      for (auto&& rna: prot->rna_list_) {
        rna->affinity_list.clear();
      }
      proteins_.push_back(prot);
    }
  }
}
#endif

} // namespace aevol
