// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************




// =================================================================
//                              Includes
// =================================================================
#include "OutputManager.h"

#include <err.h>
#include <errno.h>
#include <sys/stat.h>
#include <fstream>
#include <string>

#include "legacy/ExpManager.h"
#include "7/ExpManager_7.h"
#include "AeTime.h"

using std::string;
using std::endl;

namespace aevol {



//##############################################################################
//                                                                             #
//                           Class OutputManager                           #
//                                                                             #
//##############################################################################

// =================================================================
//                    Definition of static attributes
// =================================================================

// =================================================================
//                             Constructors
// =================================================================
OutputManager::OutputManager(ExpManager * exp_m)
{
  exp_m_  = exp_m;
  stats_  = nullptr;
  tree_   = nullptr;
  record_tree_ = false;
  logs_  = new Logging();

  backup_step_ = -1;
}

// =================================================================
//                             Destructors
// =================================================================
OutputManager::~OutputManager()
{
  delete stats_;
  delete tree_;
  delete logs_;
}

// =================================================================
//                            Public Methods
// =================================================================
void OutputManager::InitStats() {
  stats_ = new Stats(exp_m_);
}

void OutputManager::WriteSetupFile(gzFile setup_file) const
{
  // Write the backup steps
  gzwrite(setup_file, &backup_step_,      sizeof(backup_step_));
  gzwrite(setup_file, &big_backup_step_,  sizeof(big_backup_step_));

  // Dummy write (backward compatibility, was compute_phen_contrib_by_GU_)
  bool dummy = false;
  gzwrite(setup_file, &dummy, sizeof(dummy));

  // Tree
  int8_t record_tree = record_tree_;
  gzwrite(setup_file, &record_tree, sizeof(record_tree));

  // Dummy write (backward compatibility, was record_light_tree)
  int8_t dummy_light_tree = false;
  gzwrite(setup_file, &dummy_light_tree, sizeof(dummy_light_tree));

  // Tree step
  if (record_tree_)
  {
      int64_t tmp_tree_step = tree_->tree_step();
    gzwrite(setup_file, &tmp_tree_step, sizeof(tmp_tree_step));
  }

  // Dummy writes (backward compatibility, was make_dumps + dump_step)
  int8_t dummy_make_dumps = 0;
  gzwrite(setup_file, &dummy_make_dumps,  sizeof(dummy_make_dumps));
  int64_t dummy_dump_step = 0;
  gzwrite(setup_file, &dummy_dump_step,  sizeof(dummy_dump_step));

  // Logs
  int8_t logs = logs_->logs();
  gzwrite(setup_file, &logs,  sizeof(logs));
}

void OutputManager::CopyStats(const std::string& outdir, int64_t time) const {
  stats_->CreateTmpFiles(time);
  stats_->MoveTmpFiles(outdir);
}

void OutputManager::load(gzFile setup_file, bool to_be_run)
{
  // Write the backup steps
  gzread(setup_file, &backup_step_,      sizeof(backup_step_));
  gzread(setup_file, &big_backup_step_,  sizeof(big_backup_step_));
  // printf("Backup step %d\n",backup_step_);
  // Stats
  if (to_be_run)
  {
    delete stats_;
    stats_ = new Stats(exp_m_, AeTime::time());
  }

  // Dummy read (backward compatibility, was compute_phen_contrib_by_GU_)
  bool dummy;
  gzread(setup_file, &dummy, sizeof(dummy));

  // Tree
  int8_t record_tree;
  gzread(setup_file, &record_tree, sizeof(record_tree));
  record_tree_ = record_tree;

  // Dummy read (backward compatibility, was record_light_tree)
  int8_t dummy_light_tree;
  gzread(setup_file, &dummy_light_tree, sizeof(dummy_light_tree));

  // Tree step
  if (record_tree_)
  {
    int64_t tmp_tree_step;
    gzread(setup_file, &tmp_tree_step, sizeof(tmp_tree_step));

    tree_ = new Tree(exp_m_, tmp_tree_step);
  }

  // Dummy reads (backward compatibility, was make_dumps + dump_step)
  int8_t dummy_make_dumps;
  int64_t dummy_dump_step;
  gzread(setup_file, &dummy_make_dumps, sizeof(dummy_make_dumps));
  gzread(setup_file, &dummy_dump_step,  sizeof(dummy_dump_step));

  // Logs
  int8_t logs;
  gzread(setup_file, &logs, sizeof(logs));
  if (to_be_run)
  {
    logs_->load(logs, AeTime::time());
  }
}

void OutputManager::write_current_generation_outputs(bool create) const
{
  // we use the variable t because of the parallelisation.
  int64_t t = AeTime::time();
  std::list<Individual*> indivs = exp_m_->indivs();

  stats_->write_current_generation_statistics(t);

  if (record_tree_ &&
      t > 0 &&
      (t % tree_->tree_step() == 0)) {
    std::cout << "writing tree for gen : " << t << '\n';
    write_tree();
  }

  // Write backup
  if (t % backup_step_ == 0) {
    std::cout << "writing backup for gen : " << t << '\n';
    stats_->flush();
    exp_m_->WriteDynamicFiles(create);

    WriteLastGenerFile(".", t);
  }
}

// TODO <david.parsons@inria.fr> we need an output_dir attribute in this class !
    void OutputManager::WriteLastGenerFile(const string& output_dir /* = "." */, int64_t gen /* = -1 */) const {
      if(gen < 0) {
        gen = AeTime::time();
      }
      std::ofstream last_gener_file(output_dir + "/" + LAST_GENER_FNAME,
                                    std::ofstream::out);
      if (last_gener_file.fail()) {
        Utils::ExitWithUsrMsg(string("could not open file ") + LAST_GENER_FNAME);
      }
      else {
        last_gener_file << gen << endl;
        last_gener_file.close();
      }
    }

// TODO <david.parsons@inria.fr> we need an input_dir attribute in this class !
int64_t OutputManager::last_gener() {
  int64_t time = 0;
  FILE* lg_file = fopen(LAST_GENER_FNAME, "r");
  if (lg_file != NULL) {
    if (fscanf(lg_file, "%" PRId64 "\n", &time) == EOF) {
      Utils::ExitWithDevMsg("failed to read last generation", __FILE__,
                            __LINE__);
    }
    fclose(lg_file);
  }
  return time;
}
    void OutputManager::flush() {
      stats_->flush();

      // Write backup
      if (AeTime::time() % backup_step_ != 0) {
        stats_->flush();
      }
    }

// =================================================================
//                           Protected Methods
// =================================================================
    void OutputManager::write_tree() const
    {
      // Create the tree directory if it doesn't exist
      int status;
      status = mkdir(TREE_DIR, 0755);
      if ((status == -1) && (errno != EEXIST))
      {
        err(EXIT_FAILURE, "Impossible to create the directory %s", TREE_DIR);
      }

      char tree_file_name[50];
      snprintf(tree_file_name, 50, "tree/tree_" TIMESTEP_FORMAT ".ae", AeTime::time());
        
//      gzFile tree_file = gzopen( tree_file_name, "w" );
      // Write phylogenetic data (tree)
      tree_->write_to_tree_file(tree_file_name);

//      gzclose(tree_file);

      // debug std::cout << "OK" << '\n';
    }

// =================================================================
//                          Non inline accessors
// =================================================================
} // namespace aevol
