// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************


#ifndef AEVOL_OUPUT_MANAGER_H_
#define AEVOL_OUPUT_MANAGER_H_


// =================================================================
//                              Libraries
// =================================================================
#include <inttypes.h>



// =================================================================
//                            Project Files
// =================================================================
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "ae_enums.h"
#include "legacy/io/stats/Stats.h"
#include "Tree.h"
#include "Logging.h"

namespace aevol {


// =================================================================
//                          Class declarations
// =================================================================
class ExpManager;

#ifdef __OPENMP_TASK
    enum dep_omp {
  d_LT_LT  = 0,
  d_B_B    = 1,
  d_S_S    = 2,
  nb_dep   = 3
};
#endif




class OutputManager {
 public :
  // =================================================================
  //                             Constructors
  // =================================================================
  OutputManager() = delete;
  OutputManager(const OutputManager&) = delete;
  OutputManager(ExpManager* exp_m);

  // =================================================================
  //                             Destructors
  // =================================================================
  virtual ~OutputManager();

  // =================================================================
  //                        Accessors: getters
  // =================================================================

  // Backup
  inline int64_t	backup_step() const;
  inline int64_t	big_backup_step() const;

  // Tree
  inline bool record_tree() const;
  inline int64_t tree_step() const;
  inline Tree* tree() const;

  // Logs
  inline FILE* log(LogType log_type) const;
  inline bool is_logged(LogType log_type) const;

  // Stats
  inline Stats* stats() const;

  // =================================================================
  //                        Accessors: setters
  // =================================================================
  inline void set_backup_step(int64_t backup_step);
  inline void set_big_backup_step(int64_t big_backup_step);
  inline void init_tree(ExpManager* exp_m, int64_t tree_step_);
  inline void set_logs (int8_t logs);

  // =================================================================
  //                              Operators
  // =================================================================

  // =================================================================
  //                            Public Methods
  // =================================================================
  void InitStats();
  void WriteSetupFile(gzFile setup_file) const;
  void WriteLastGenerFile(const std::string& input_dir = ".", int64_t gen = -1) const;
  void CopyStats(const std::string& outdir, int64_t time) const;
  void load(gzFile file, bool to_be_run);
  void write_current_generation_outputs(bool create = false) const;
  void flush();

  static int64_t last_gener();

    void write_tree() const;

 protected :
  // =================================================================
  //                           Protected Methods
  // =================================================================


  // =================================================================
  //                          Protected Attributes
  // =================================================================
  ExpManager* exp_m_;

  // Backups
  int64_t backup_step_;
  int64_t big_backup_step_;

  // Stats
  Stats* stats_;

  // Tree
  bool record_tree_;
  Tree* tree_;

  // Logs
  Logging* logs_;

#ifdef __OPENMP_TASK
        //omp sychronization
  static bool dep[nb_dep];
#endif
};


// =====================================================================
//                           Getters' definitions
// =====================================================================

// Backup
inline int64_t OutputManager::backup_step() const {
  return backup_step_;
}

inline int64_t OutputManager::big_backup_step() const {
  return big_backup_step_;
}

// Tree
inline bool OutputManager::record_tree() const {
  return record_tree_;
}

inline int64_t OutputManager::tree_step() const {
  return tree_->tree_step();
}

inline Tree *OutputManager::tree() const {
  return tree_;
}

// Logs
inline FILE* OutputManager::log(LogType log_type) const {
  return logs_->log(log_type);
}

inline bool  OutputManager::is_logged(LogType log_type) const {
  return logs_->is_logged(log_type);
}

// Stats
inline Stats* OutputManager::stats() const {
  return stats_;
}

// =====================================================================
//                           Setters' definitions
// =====================================================================
void OutputManager::set_backup_step(int64_t backup_step) {
  backup_step_ = backup_step;
}

void OutputManager::set_big_backup_step(int64_t big_backup_step) {
  big_backup_step_ = big_backup_step;
}

void OutputManager::init_tree(ExpManager* exp_m, int64_t tree_step_) {
  record_tree_ = true;
  tree_ = new Tree(exp_m, tree_step_);
}

void OutputManager::set_logs(int8_t logs) {
  logs_->set_logs(logs);
}

// =====================================================================
//                          Operators' definitions
// =====================================================================

// =====================================================================
//                       Inline functions' definition
// =====================================================================

} // namespace aevol

#endif // AEVOL_OUPUT_MANAGER_H_
