// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************

#include "UserInterfaceOutput.h"

#include <fstream>
#include <iostream>
#include <sstream>

#include <nlohmann/json.hpp>

// Shorthand for nlohmann's json library
// We use the ordered_json specialization that maintains the insertion order
// of object keys instead of the default that uses alphabetical order for
// serialization
using json = nlohmann::ordered_json;

#include "AeTime.h"
#include "Individual_7.h"
#include "Protein_7.h"
#include "Rna_7.h"

namespace aevol {

UserInterfaceOutput::UserInterfaceOutput(const std::filesystem::path& outdir, int32_t frequency)
    : outdir_(outdir), frequency_(frequency) {
  std::filesystem::create_directories(outdir_);
}

bool UserInterfaceOutput::nothing_to_output() const {
  return AeTime::time() % frequency_ != 0;
}

void UserInterfaceOutput::write_indiv_output(const Individual_7& indiv) {
  // Open output file
  std::ostringstream ofnstream;
  ofnstream << std::string("indiv_") << std::setfill('0') << std::setw(9) << std::to_string(AeTime::time()) << ".json";
  std::ofstream indiv_out(outdir_ / ofnstream.str(), std::ios::out);

  // Fill json with indiv's data
  json indiv_json;
  indiv_json["indiv"]["genome_len"] = indiv.annotated_chromosome_->dna_->length();
  indiv_json["indiv"]["fitness"] = indiv.fitness();

  // Rnas
  indiv_json["indiv"]["rnas"] = json::array();
  for (const auto& rna : indiv.annotated_chromosome_->rnas()) {
    indiv_json["indiv"]["rnas"].push_back(json::object({
        {"begin", rna->begin},
        {"end", rna->end},
        {"strand", to_string(rna->strand_)},
        {"is_coding", rna->is_coding_},
        {"expression_level", rna->e}}));
  }

  // Proteins
  indiv_json["indiv"]["proteins"] = json::array();
  for (const auto& protein : indiv.annotated_chromosome_->proteins()) {
    indiv_json["indiv"]["proteins"].push_back(json::object({
        {"begin", protein->protein_start},
        {"end", protein->protein_end},
        {"strand", to_string(protein->strand_)},
        {"m", protein->m},
        {"w", protein->w},
        {"h", protein->h},
        {"expression_level", protein->e}}));
  }

  // Phenotype
  indiv_json["indiv"]["phenotype"]["points"] = json::array();
  auto nb_points = indiv.phenotype->nb_points();
  for (decltype(nb_points) i = 0 ; i < nb_points ; ++i) {
    indiv_json["indiv"]["phenotype"]["points"].push_back(json::object({
      {"x", indiv.phenotype->point(i).x},
      {"y", indiv.phenotype->point(i).y}}));
  }

  // Write json to output file
  indiv_out << indiv_json.dump(2) << std::endl;
}

void UserInterfaceOutput::write_grid_output(size_t grid_width, size_t grid_height, const Individual_7* const* indivs) {
  // Open output file
  std::ostringstream ofnstream;
  ofnstream << std::string("grid_") << std::setfill('0') << std::setw(9) << std::to_string(AeTime::time()) << ".json";
  std::ofstream grid_out(outdir_ / ofnstream.str(), std::ios::out);

  // Fill json with grid data
  json grid_json;
  grid_json["grid_data"]["grid_width"] = grid_width;
  grid_json["grid_data"]["grid_height"] = grid_height;

  // Individuals' fitness
  std::vector<double> fitnesses;
  auto grid_size = grid_width * grid_height;
  fitnesses.reserve(grid_size);
  for (decltype(grid_size) i = 0 ; i < grid_size ; ++i) {
    fitnesses.push_back(indivs[i]->fitness());
  }
  grid_json["grid_data"]["fitness_grid"] = fitnesses;

  // Write json to output file
  grid_out << grid_json.dump(2) << std::endl;
}

void UserInterfaceOutput::write_environment_output(AbstractFuzzy* target) {
  // Open output file
  std::ostringstream ofnstream;
  ofnstream << std::string("environment") << ".json";
  std::ofstream env_out(outdir_ / ofnstream.str(), std::ios::out);

  // Fill json with environmental target data
  json env_json;
  env_json["environment"]["target"]["points"] = json::array();
  auto nb_points = target->nb_points();
  for (decltype(nb_points) i = 0 ; i < nb_points ; ++i) {
    env_json["environment"]["target"]["points"].push_back(json::object({
        {"x", target->point(i).x},
        {"y", target->point(i).y}}));
  }

  // Write json to output file
  env_out << env_json.dump(2) << std::endl;
}

}
