// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************




// =================================================================
//                              Includes
// =================================================================
#include "ExpManager.h"

#include "7/Dna_7.h"
#include "7/Individual_7.h"
#include "7/ExpManager_7.h"
#include "7/Stats_7.h"
#include "legacy/population/Individual.h"

#include <cerrno>
#include <err.h>
#include <iostream>
#include <sstream>
#include <list>
#include <sys/stat.h>
#include <zlib.h>

#ifdef __REGUL
#include "raevol/Individual_R.h"
#endif
#include<chrono>

#include <iostream>
#include <unordered_map>

#ifdef __TRACING__
#include "ae_logger.h"


unordered_map<int,unordered_multiset<string>> ae_logger::logMap;
string ae_logger::logFile = "logger_csv.log";
mutex ae_logger::loggerMtx;
#endif

using std::cout;
using std::endl;


namespace aevol {

ExpManager* exp_manager;

//##############################################################################
//                                                                             #
//                             Class ExpManager                            #
//                                                                             #
//##############################################################################

// ===========================================================================
//                        Definition of static attributes
// ===========================================================================

// ===========================================================================
//                                 Constructors
// ===========================================================================
ExpManager::ExpManager()
{
  // ------------------------------------------------------ Experimental setup
  exp_s_ = new ExpSetup(this);

  // ------------------------------------------------------------------- World
  world_ = nullptr;

  // ---------------------------------------------------------- Output manager
  output_m_ = new OutputManager(this);


  // -------------------------------- Timestep up to which we want to simulate
  t_end_ = 0;


  // ------------------------------------------------------------- Quit signal
  quit_signal_received_ = false;


  // ------------------------ Which Individuals will reproduce in each GridCell


#ifdef __PROXY_POW_APPROX
#ifdef __REGUL
  Rna_R::load_lookup_table();
#endif
#endif
}

// ===========================================================================
//                                  Destructor
// ===========================================================================
ExpManager::~ExpManager() noexcept
{
  delete exp_m_7_;
  delete output_m_;
  delete world_;


  delete exp_s_;

  delete [] next_generation_reproducer_;
}

// ===========================================================================
//                                   Algorithms
// ===========================================================================

// ===========================================================================
//                                 Public Methods
// ===========================================================================
void ExpManager::InitializeWorld(int16_t grid_width,
                                     int16_t grid_height,
                                     std::shared_ptr<JumpingMT> prng,
                                     std::shared_ptr<JumpingMT> mut_prng,
                                     std::shared_ptr<JumpingMT> stoch_prng,
                                     Habitat& habitat,
                                     bool share_phenotypic_target)
{
  world_ = new World();
  world_->set_prng(prng);
  world_->set_mut_prng(mut_prng);
  world_->set_stoch_prng(stoch_prng);
  world_->InitGrid(grid_width, grid_height, habitat, share_phenotypic_target);
  next_generation_reproducer_ = new int32_t[nb_indivs()];
}

/*!
  \brief Save the experiment
*/
void ExpManager::Save(bool create) const
{
  WriteSetupFiles(create);
  output_m_->write_current_generation_outputs(create);
}

/*!
  \brief Save all the static data (the data that is constant throughout
  a simulation)

  Save
    * The experimental setup
    * The output profile

  These are written in the current directory in gzipped binary files
  as well as plain text files
  This data should be saved once and for all when the experiment is created.
  There is no need to save this data in the middle of a simulation since it
  is constant throughout the experiment.

  \see load(int64_t t0,
            char* exp_setup_file_name,
            char* out_prof_file_name,
            char* pop_file_name,
            char* sel_file_name,
            char* world_file_name,
            bool verbose)
  \see WriteDynamicFiles()
  \see save_copy(char* dir, int64_t t)
*/
void ExpManager::WriteSetupFiles(bool first) const
{
  // 1) Create missing directories
  create_missing_directories();

  // 2) Open setup files (experimental setup and output profile)
  gzFile exp_s_file, out_p_file;

  if (first) {
    open_setup_files(exp_s_file, out_p_file, "w");
    exp_s_->write_setup_file(exp_s_file);
    output_m_->WriteSetupFile(out_p_file);
    close_setup_files(exp_s_file, out_p_file);
  } else {
    open_setup_files(exp_s_file, out_p_file, "w");

    // 4) Write setup data
    exp_s_->write_setup_file(exp_s_file);
    output_m_->WriteSetupFile(out_p_file);

    // 5) Close setup files
    close_setup_files(exp_s_file, out_p_file);
  }
  


}

/*!
  \brief Save all the dynamic data (the data that changes over the simulation)

  Save the state of
    * The population
    * The world
    * The PRNG (random generator) used for the selection

  These are written in the current directory in gzipped binary files
  whose names are appended with the current generation number.

  WARNING: The experimental setup and output profile are *not* saved. These
  should be saved once and for all when the experiment is created.

  \see load(int64_t t0,
            char* exp_setup_file_name,
            char* out_prof_file_name,
            char* pop_file_name,
            char* sel_file_name,
            char* world_file_name,
            bool verbose)
  \see WriteSetupFiles()
  \see save_copy(char* dir, int64_t t)
*/
void ExpManager::WriteDynamicFiles(bool first) const
{
  // Create missing directories
  create_missing_directories();

  
  // Save experiment
  if (first) {
    // Open backup files
    gzFile sel_file, world_file;
    open_backup_files(sel_file, world_file, AeTime::time(), "w");

    sel()->save(sel_file);
    
    world_->save(world_file);
    
    // Close backup files
    close_backup_files(sel_file, world_file);
  } else {
    // Open backup files
    gzFile sel_file, world_file;
    open_backup_files(sel_file, world_file, AeTime::time(), "w");

    sel()->save(sel_file);
    
    world_->save(world_file);
    
    // Close backup files
    close_backup_files(sel_file, world_file);
  }


}

#ifdef __REGUL
void ExpManager::WriteDynamicFiles(int64_t gen, SaveWorld* world) const {
#else
void ExpManager::WriteDynamicFiles(int64_t gen,
                                   SaveWorld* world,
                                   bool create) const {
#endif
      // Create missing directories
      create_missing_directories();

      // Open backup files
      gzFile sel_file, world_file;
      open_backup_files(sel_file, world_file, gen, "w");

      // Save experiment
      #ifdef __REGUL
      world->save(world_file);
      #else
      world->save(world_file, create);
      #endif

      // Close backup files
      close_backup_files(sel_file, world_file);
    }

/*!
  \brief Saves a complete copy of the experiment at the provided location.

  Save both the setup (constant) and the backup (dynamic) files to the <dir>
  directory. Dynamic file names will be appended with the <num_gener>

  Static files (saved both as gzipped binary and plain text)
    * The experimental setup
    * The output profile

  Dynamic files (saved as gzipped binary)
    * The world
    * The PRNG (random generator) used for the selection

  \see load(int32_t first_gener,
             char* exp_setup_file_name,
             char* out_prof_file_name,
             char* pop_file_name,
             char* sel_file_name,
             char* world_file_name,
             bool verbose)
  \see WriteSetupFiles()
  \see WriteDynamicFiles()
*/
void ExpManager::save_copy(char* dir, int64_t time) const
{
  // Set time to time
  AeTime::set_time(time);

  // Create missing directories
  create_missing_directories(dir);

  // Open setup files and backup files
  gzFile exp_s_file, out_p_file, sel_file, world_file;
  open_setup_files(exp_s_file, out_p_file, "w", dir);
  open_backup_files(sel_file, world_file, time, "w", dir);

  // Write setup data
  exp_s_->write_setup_file(exp_s_file);
  output_m_->WriteSetupFile(out_p_file);

  // Write the state of selection and world into the backups
  sel()->save(sel_file);
  world_->save(world_file);

  // Close setup and backup files
  close_setup_files(exp_s_file, out_p_file);
  close_backup_files(sel_file, world_file);

  // Copy stats
  output_m_->CopyStats(dir, time);

  // Write last gener file
  output_m_->WriteLastGenerFile(dir);
}



void ExpManager::step_to_next_generation() {
  // TODO <david.parsons@inria.fr> Apply phenotypic target variation and noise
#ifdef __PERF_LOG__
    auto t1 = high_resolution_clock::now();
    auto t2 = high_resolution_clock::now();
    auto duration_2 = std::chrono::duration_cast<std::chrono::microseconds>(t2 - t1).count();

    auto ta = high_resolution_clock::now();
#endif

#pragma omp single
    {
      #ifdef __REGUL
        exp_m_7_->phenotypic_target_handler_->ApplyVariation();
      #else
        world_->ApplyHabitatVariation();
      #endif

        // Take a step in time
        AeTime::plusplus();

        // Create the corresponding new generation
#ifdef __PERF_LOG__
        auto t1 = high_resolution_clock::now();

        //auto     s_t1 = high_resolution_clock::now();
#endif

/*  bool simd_first = false;
  if (exp_m_7_ == nullptr) {
    simd_first = true;
  }*/

        //}
#ifdef __PERF_LOG__
        t1 = high_resolution_clock::now();
#endif
  }

#pragma omp single
        {
  #ifdef __PERF_LOG__
          auto t2 = high_resolution_clock::now();
  #endif

#ifdef __PERF_LOG__
          auto duration_2 =
              std::chrono::duration_cast<std::chrono::microseconds>(t2 - t1)
                  .count();

          auto ta = high_resolution_clock::now();
#endif
        }

    exp_m_7_->run_a_step(w_max_, selection_pressure());

#pragma omp single
    {
#ifdef __PERF_LOG__
        auto tb = high_resolution_clock::now();
        auto duration_simd = std::chrono::duration_cast<std::chrono::microseconds>(tb - ta).count();
#endif

#ifdef __PERF_LOG__
        std::cout << "PERFLOG," << AeTime::time() << "," << duration_simd << ","<< exp_m_7_->cumulate_size << "," << exp_m_7_->cumulate_diff <<std::endl;
#endif
    }
}

/*!
  \brief Load an experiment with the provided files
 */
void ExpManager::load(gzFile& exp_s_file,
                      gzFile& exp_backup_file,
                      gzFile& world_file,
                      gzFile& out_p_file,
                      bool to_be_run /*  = true */) {
  // ---------------------------------------- Retrieve experimental setup data
  printf("  Loading experimental setup...");
  fflush(stdout);
  exp_s_->load(exp_s_file, exp_backup_file);
  printf(" OK\n");

  // ---------------------------------------------------------- Retrieve world
  printf("  Loading world...");
  fflush(stdout);
  world_ = new World();
  world_->load(world_file, this);
  printf(" OK\n");

  //printf("-------------- Nb Gaussians %d\n",world()->phenotypic_target_handler()->gaussians().size());

  // --------------------------------------------- Retrieve output profile data
  printf("  Loading output profile...");
  fflush(stdout);
  output_m_->load(out_p_file, to_be_run);
  printf(" OK\n");

  exp_m_7_ = new ExpManager_7;
  exp_m_7_->load_legacy(this, this->exp_s_, this->world_);
  exp_m_7_->protein_grain_size = grain_size;
  exp_m_7_->rna_grain_size     = grain_size;


  // -------------------------------------------- Link world and output profile
  if (record_tree()) {
    if (to_be_run) {
      exp_m_7_->addObserver(tree(), NEW_INDIV);
        for (int16_t x = 0; x < grid_width(); x++) {
            for (int16_t y = 0; y < grid_height(); y++) {
              exp_m_7_->current_individuals[x * grid_height() + y]->addObserver(
                        tree(),
                        END_REPLICATION);
            }
        }
        exp_m_7_->addObserver(tree(), END_GENERATION);
    }
  }

  // --------------------------------------------------- Recompute unsaved data
  world_->evaluate_individuals();
  //update_best();
}


/*!
  \brief Load an experiment with default files from a given directory
 */
// TODO <david.parsons@inria.fr> check verbose (what does it do ?, is it consistent ?)
void ExpManager::load(const char* dir,
    int64_t t0, bool to_be_run /*  = true */) {
  AeTime::set_time(t0);

  // -------------------------------------------------------------------------
  // Open setup files and backup files
  // -------------------------------------------------------------------------
  gzFile exp_s_file, out_p_file;
  gzFile exp_backup_file, world_file;
  open_setup_files(exp_s_file, out_p_file, "r", dir);
  open_backup_files(exp_backup_file, world_file, t0, "r", dir);

  // -------------------------------------------------------------------------
  // Load data from backup and parameter files
  // -------------------------------------------------------------------------
  load(exp_s_file, exp_backup_file, world_file, out_p_file, to_be_run);

  // -------------------------------------------------------------------------
  // Close setup and backup files
  // -------------------------------------------------------------------------
  close_setup_files(exp_s_file, out_p_file);
  close_backup_files(exp_backup_file, world_file);
}



//int mutator = 0;

/**
 * Run the simulation
 */
void ExpManager::run_evolution() {
  next_generation_reproducer_ = new int32_t[nb_indivs()];

  for (int indiv_id = 0; indiv_id < nb_indivs(); indiv_id++) {
    next_generation_reproducer_[indiv_id] = indiv_id;
  }

  // We are running a simulation.
  // Save the setup files to keep track of the setup history
  WriteSetupFiles();
#ifdef __TRACING__
  ae_logger::init("logger_csv.log");

    printf("Launching TRACING...\n");
#endif

#ifdef __TRACING__
  high_resolution_clock::time_point t_t1 = high_resolution_clock::now();
    high_resolution_clock::time_point t_t2,t1,t2;
#endif

        w_max_ = world()->indiv_at(0,0)->w_max();

  bool first_run = true;

      exp_m_7_->setup_individuals(w_max_, selection_pressure());
      exp_m_7_->write_ui_output();

      bool finished=false;
//      int64_t time = AeTime::time();
        // For each generation
#pragma omp parallel default(shared)
                while (!finished) {
#pragma omp single
                        {
                            if (AeTime::time() % 1 == 0) {
                                printf(
                                        "============================== %" PRId64 " ==============================\n",
                                        AeTime::time());
                                if (!first_run) {
                                      printf(
                                          "  Best individual's distance to target (metabolic) : %f (clones %d %ld)\n",
                                          //exp_m_7_->best_indiv->indiv_id,
                                          exp_m_7_->best_indiv->metaerror,
                                          exp_m_7_->nb_clones_,
                                          exp_m_7_->mutant_list_.size());
                                } else {
                                    printf("  Aevol_6 : Best individual's distance to target (metabolic) : %f\n",
                                            //best_indiv()->id(),
                                           best_indiv()->dist_to_target_by_feature(METABOLISM));
                                }
                            }

                            first_run = false;

#ifdef __TRACING__
                            t1 = high_resolution_clock::now();
#endif


                        }

#pragma omp single
                  {
                    if (with_mrca_) {
                      if (quit_signal_received())
                        finished = true;
                    } else if (AeTime::time() >= t_end_ or quit_signal_received())
                      finished = true;
                        }

        if (!finished)
                        // Take one step in the evolutionary loop
                        step_to_next_generation();



                }
#ifdef __TRACING__
  t_t2 = high_resolution_clock::now();
        auto duration = std::chrono::duration_cast<std::chrono::microseconds>( t_t2 - t_t1 ).count();
        ae_logger::addLog(TOTAL,duration);
        ae_logger::flush(AeTime::time());
#endif

  printf("================================================================\n");
  printf("  The run is finished. \n");
}

void ExpManager::update_best()
{
  world_->update_best();
}

// ===========================================================================
//                                Protected Methods
// ===========================================================================
void ExpManager::create_missing_directories(const char* dir /*= "."*/) const
{
  int status;

  // Base directory
  status = mkdir(dir, 0755);
  if ((status == -1) && (errno != EEXIST))
  {
    err(EXIT_FAILURE, dir, errno);
  }

  // Experimental setup
  std::ostringstream exp_s_dir_name;
  exp_s_dir_name << dir << '/' << EXP_S_DIR;
  status = mkdir(exp_s_dir_name.str().c_str(), 0755);
  if ((status == -1) && (errno != EEXIST))
  {
    err(EXIT_FAILURE, exp_s_dir_name.str().c_str(), errno);
  }

  // Output profile
  std::ostringstream out_p_dir_name;
  out_p_dir_name << dir << '/' << OUT_P_DIR;
  status = mkdir(out_p_dir_name.str().c_str(), 0755);
  if ((status == -1) && (errno != EEXIST))
  {
    err(EXIT_FAILURE, out_p_dir_name.str().c_str(), errno);
  }

  // World
  std::ostringstream world_dir_name;
  world_dir_name << dir << '/' << WORLD_DIR;
  status = mkdir(world_dir_name.str().c_str(), 0755);
  if (status == -1 && errno != EEXIST)
  {
    err(EXIT_FAILURE, world_dir_name.str().c_str(), errno);
  }

  // Stats
  std::ostringstream stats_dir_name;
  stats_dir_name << dir << '/' << STATS_DIR;
  status = mkdir(stats_dir_name.str().c_str(), 0755);
  if (status == -1 && errno != EEXIST)
  {
    err(EXIT_FAILURE, stats_dir_name.str().c_str(), errno);
  }
}

void ExpManager::open_backup_files(
    gzFile& exp_backup_file, gzFile& world_file, int64_t t, const char mode[3], const char* dir /*= "."*/) const {
  assert(strcmp(mode, "w") == 0 or strcmp(mode, "r") == 0);

  // -------------------------------------------------------------------------
  // Generate backup file names for mandatory files.
  // -------------------------------------------------------------------------
  char exp_backup_file_name[255];
  char world_file_name[255];

  snprintf(world_file_name, 255, "%s/" WORLD_FNAME_FORMAT, dir, t);
  snprintf(exp_backup_file_name, 255, "%s/" EXP_S_FNAME_FORMAT, dir, t);

  // -------------------------------------------------------------------------
  // Open backup files
  // -------------------------------------------------------------------------
  exp_backup_file = gzopen(exp_backup_file_name, mode);
  world_file = gzopen(world_file_name, mode);


  // -------------------------------------------------------------------------
  // Check that files were correctly opened
  // -------------------------------------------------------------------------
  if (exp_backup_file == Z_NULL)
  {
    printf("%s:%d: error: could not open backup file %s\n",
            __FILE__, __LINE__, exp_backup_file_name);
    exit(EXIT_FAILURE);
  }
  if (world_file == Z_NULL)
  {
    printf("%s:%d: error: could not open backup file %s\n",
            __FILE__, __LINE__, world_file_name);
    exit(EXIT_FAILURE);
  }
}

void ExpManager::close_backup_files(gzFile& exp_backup_file,
                                        gzFile& world_file) const
{
  gzclose(exp_backup_file);
  gzclose(world_file);
}

void ExpManager::open_setup_files(
    gzFile& exp_s_file, gzFile& out_p_file, const char mode[3], const char* dir /*= "."*/) const {
  // 1) Generate setup file names
  char exp_s_file_name[255];
  char out_p_file_name[255];

  snprintf(exp_s_file_name, 255, "%s/" EXP_S_CONST_FNAME_FORMAT, dir);
  snprintf(out_p_file_name, 255, "%s/" OUT_P_FNAME_FORMAT, dir);

  // 2) Open backup files
  exp_s_file = gzopen(exp_s_file_name, mode);
  out_p_file = gzopen(out_p_file_name, mode);

  // 3) Check that files were correctly opened
  if (exp_s_file == Z_NULL)
  {
    printf("%s:%d: error: could not open backup file %s\n",
            __FILE__, __LINE__, exp_s_file_name);
    exit(EXIT_FAILURE);
  }
  if (out_p_file == Z_NULL)
  {
    printf("%s:%d: error: could not open backup file %s\n",
            __FILE__, __LINE__, out_p_file_name);
    exit(EXIT_FAILURE);
  }
}

void ExpManager::close_setup_files(gzFile& exp_s_file,
                                       gzFile& out_p_file) const
{
  gzclose(exp_s_file);
  gzclose(out_p_file);
}

// ===========================================================================
//                               Non inline accessors
// ===========================================================================
Individual* ExpManager::indiv_by_id(int32_t id) const {
  return world_->indiv_by_id(id);
}

Individual* ExpManager::indiv_by_rank(int32_t rank) const {
  return world_->indiv_by_rank(rank);
}
Individual* ExpManager::indiv_by_position(int16_t x, int16_t y) const {
  return world_->indiv_at(x,y);
}

/**
 * Returns a list of all the individuals with their replication report
 */
std::list<std::pair<Individual*, ReplicationReport*>>
    ExpManager::indivs_annotated() const {
  std::list<std::pair<Individual*, ReplicationReport*>> annotated_list;
  for (const auto& indiv : indivs()) {
    annotated_list.emplace_back(indiv, tree() ?
        tree()->report_by_index(AeTime::time(), indiv->id()) : nullptr);
  }
  return annotated_list;
}


} // namespace aevol
