// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************




// =================================================================
//                              Includes
// =================================================================
#include "Selection.h"

#include "7/Dna_7.h"
#include "7/ExpManager_7.h"

#include "DnaMutator.h"
//#include <math.h>

#ifdef _OPENMP
#include <omp.h>
#endif

#if __cplusplus == 201103L
#include "make_unique.h"
#endif

#include<chrono>

#include <iostream>
#include <unordered_map>

using std::chrono::high_resolution_clock;

#include "ae_logger.h"
#include "legacy/ExpManager.h"
#include "VisAVis.h"

#ifndef __REGUL
  #include "legacy/population/Individual.h"
#else
  #include "legacy/raevol/Individual_R.h"
#endif

#include "7/Individual_7.h"

namespace aevol {


//##############################################################################
//
//                              Class Selection
//
//##############################################################################

// =================================================================
//                    Definition of static attributes
// =================================================================

// =================================================================
//                             Constructors
// =================================================================
Selection::Selection(ExpManager* exp_m) {
  exp_m_ = exp_m;

  // ----------------------------------------- Pseudo-random number generator
  //prng_ = NULL;

  // -------------------------------------------------------------- Selection
  selection_scheme_   = RANK_EXPONENTIAL;
  selection_pressure_ = 0.998;


        fitness_function_ = FITNESS_EXP;
        fitness_function_scope_x_ = 3;
        fitness_function_scope_y_ = 3;

  // --------------------------- Probability of reproduction of each organism
  prob_reprod_ = NULL;

#ifdef WITH_PERF_TRACES
        std::ofstream perf_traces_file_;
        perf_traces_file_.open("vanilla_perf_traces.csv",std::ofstream::trunc);
        perf_traces_file_<<"Generation,Indiv_ID,Runtime"<<std::endl;
        perf_traces_file_.close();
#endif

}

// =================================================================
//                             Destructors
// =================================================================
Selection::~Selection() {
  delete [] prob_reprod_;
}


int mutator = 0;

// =================================================================
//                            Public Methods
// =================================================================
/*!
*/
void Selection::write_setup_file(gzFile exp_setup_file) const {
  // ---------------------------------------------------- Selection Parameters
  int8_t tmp_sel_scheme = static_cast<int8_t>(selection_scheme_);
  gzwrite(exp_setup_file, &tmp_sel_scheme,      sizeof(tmp_sel_scheme));
  gzwrite(exp_setup_file, &selection_pressure_, sizeof(selection_pressure_));

  int8_t tmp_sel_scope = static_cast<int8_t>(selection_scope_);
  gzwrite(exp_setup_file, &tmp_sel_scope,      sizeof(tmp_sel_scope));
  gzwrite(exp_setup_file, &selection_scope_x_, sizeof(selection_scope_x_));
  gzwrite(exp_setup_file, &selection_scope_y_, sizeof(selection_scope_y_));


        int8_t tmp_fit_func= fitness_function_;
        gzwrite(exp_setup_file, &tmp_fit_func,      sizeof(tmp_fit_func));
        gzwrite(exp_setup_file, &fitness_function_scope_x_, sizeof(fitness_function_scope_x_));
        gzwrite(exp_setup_file, &fitness_function_scope_y_, sizeof(fitness_function_scope_y_));
}

/*!
*/
void Selection::save(gzFile&) const {
  /*if (prng_ == NULL) {
    printf("%s:%d: error: PRNG not initialized.\n", __FILE__, __LINE__);
    exit(EXIT_FAILURE);
  }*/

  // ----------------------------------------- Pseudo-random number generator
  //prng_->save(backup_file);
}

void Selection::load(gzFile& exp_setup_file,
                     gzFile&) {
  // ---------------------------------------------------- Selection parameters
  int8_t tmp_sel_scheme;
  gzread(exp_setup_file, &tmp_sel_scheme, sizeof(tmp_sel_scheme));
  selection_scheme_ = static_cast<SelectionScheme>(tmp_sel_scheme);
  gzread(exp_setup_file, &selection_pressure_, sizeof(selection_pressure_));

  int8_t tmp_sel_scope;
  gzread(exp_setup_file, &tmp_sel_scope, sizeof(tmp_sel_scope));
  selection_scope_ = static_cast<SelectionScope>(tmp_sel_scope);
  gzread(exp_setup_file, &selection_scope_x_, sizeof(selection_scope_x_));
  gzread(exp_setup_file, &selection_scope_y_, sizeof(selection_scope_y_));


    int8_t tmp_fit_func;
    gzread(exp_setup_file, &tmp_fit_func, sizeof(tmp_fit_func));
    fitness_function_ = (FitnessFunction) tmp_fit_func;
    gzread(exp_setup_file, &fitness_function_scope_x_, sizeof(fitness_function_scope_x_));
    gzread(exp_setup_file, &fitness_function_scope_y_, sizeof(fitness_function_scope_y_));
  // ----------------------------------------- Pseudo-random number generator
  /*
#if __cplusplus == 201103L
  prng_ = make_unique<JumpingMT>(backup_file);
#else
  prng_ = std::make_unique<JumpingMT>(backup_file);
#endif*/
}


// =================================================================
//                           Protected Methods
// =================================================================
#ifdef __REGUL
void Selection::run_life(Individual_R* new_indiv) {

    if (dynamic_cast<PhenotypicTargetHandler_R*>(&new_indiv->grid_cell()->habitat().
        phenotypic_target_handler_nonconst())->hasChanged()) {
      new_indiv->evaluated_ = false;
    }

    // Evaluate new individual
    new_indiv->Evaluate();

    // Compute statistics
    new_indiv->compute_statistical_data();

}
#else
void Selection::run_life(Individual* new_indiv) {
  // Evaluate new individual
  new_indiv->Evaluate();

  // Compute statistics
  new_indiv->compute_statistical_data();

}
#endif

// =================================================================
//                          Non inline accessors
// =================================================================
} // namespace aevol
