// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************

// =================================================================
//                              Libraries
// =================================================================
#include <zlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <set>

// =================================================================
//                            Project Files
// =================================================================
#ifdef __REGUL
  #include "raevol/BindingMatrix.h"
#endif
#include "ExpSetup.h"
#include "JumpingMT.h"


//#include "GzHelpers.h"

namespace aevol {
//##############################################################################
//                                                                             #
//                              Class ExpSetup                                 #
//                                                                             #
//##############################################################################

// ===========================================================================
//                         Definition of static attributes
// ===========================================================================

// ===========================================================================
//                                  Constructors
// ===========================================================================
ExpSetup::ExpSetup( ExpManager * exp_m )
{
  exp_m_  = exp_m;

  // -------------------------------------------------------------- Selection
  sel_ = new Selection( exp_m );

  // --------------------------------------------------------------- Transfer
  with_HT_                    = false;
  repl_HT_with_close_points_  = false;
  HT_ins_rate_                = 0.0;
  HT_repl_rate_               = 0.0;
  repl_HT_detach_rate_         = 0.0;

  // -------------------------------------------------------------- Secretion
  with_secretion_ = false;
  secretion_contrib_to_fitness_ = 0.0;
  secretion_cost_               = 0.0;

#ifdef BASE_4
  // -------------------------------------------------------------- Terminators
  terminator_polya_sequence_length_ = 0;

  // -------------------------------------------------------------- MWH bases configuration
  for(int8_t i = 0; i < NB_AMINO_ACIDS; i++) {
    aa_base_m_[i] = -1;
    aa_base_w_[i] = -1;
    aa_base_h_[i] = -1;
  }

   // randomly generated no epistasis code
  
  aa_base_m_[ARGININE] = 2;    
  aa_base_m_[ASPARAGINE] = 4;
  aa_base_m_[ASPARTIC_ACID] = 5;
  aa_base_m_[GLUTAMIC_ACID] = 6;
  aa_base_m_[GLUTAMINE] = 0; 
  aa_base_m_[HISTIDINE] = 1;   
  aa_base_m_[SERINE] = 3;       

  aa_base_w_[ALANINE] = 5;  
  aa_base_w_[CYSTEINE] = 0;   
  aa_base_w_[ISOLEUCINE] = 4;
  aa_base_w_[LEUCINE] = 3; 
  aa_base_w_[THREONINE] = 1;      
  aa_base_w_[PHENYLALANINE] = 2;   
  
  aa_base_h_[GLYCINE] = 3;    
  aa_base_h_[METHIONINE] = 2;
  aa_base_h_[LYSINE] = 5;          
  aa_base_h_[PROLINE] = 6;       
  aa_base_h_[TRYPTOPHAN] = 1;    
  aa_base_h_[TYROSINE] = 4;   
  aa_base_h_[VALINE] = 0; 

  /*          
  aa_base_x_[ALANINE] = ;  
  aa_base_x_[ARGININE] = ;    
  aa_base_x_[ASPARAGINE] = ;
  aa_base_x_[ASPARTIC_ACID] = ;
  aa_base_x_[CYSTEINE] = ;
  aa_base_x_[GLUTAMIC_ACID] = ;
  aa_base_x_[GLUTAMINE] = ; 
  aa_base_x_[GLYCINE] = ;
  aa_base_x_[HISTIDINE] = ;
  aa_base_x_[ISOLEUCINE] = ; 
  aa_base_x_[LEUCINE] = ; 
  aa_base_x_[LYSINE] = ;       
  aa_base_x_[METHIONINE] = ;
  aa_base_x_[PHENYLALANINE] = ;    
  aa_base_x_[PROLINE] = ;      
  aa_base_x_[SERINE] = ;     
  aa_base_x_[THREONINE] = ;   
  aa_base_x_[TRYPTOPHAN] = ;    
  aa_base_x_[TYROSINE] = ;   
  aa_base_x_[VALINE] = ;  
 */

  // GB 07/01/22 randomly generated full code

  // aa_base_m_[ALANINE] = 9;  
  // aa_base_m_[ARGININE] = 2;    
  // aa_base_m_[ASPARAGINE] = 4;
  // aa_base_m_[ASPARTIC_ACID] = 5;
  // aa_base_m_[CYSTEINE] = 14;
  // aa_base_m_[GLUTAMIC_ACID] = 6;
  // aa_base_m_[GLUTAMINE] = 0; 
  // aa_base_m_[GLYCINE] = 16;
  // aa_base_m_[HISTIDINE] = 1;
  // aa_base_m_[ISOLEUCINE] = 17; 
  // aa_base_m_[LEUCINE] = 19; 
  // aa_base_m_[LYSINE] = 11;       
  // aa_base_m_[METHIONINE] = 10;
  // aa_base_m_[PHENYLALANINE] = 12;    
  // aa_base_m_[PROLINE] = 15;      
  // aa_base_m_[SERINE] = 3;     
  // aa_base_m_[THREONINE] = 8;   
  // aa_base_m_[TRYPTOPHAN] = 7;    
  // aa_base_m_[TYROSINE] = 13;   
  // aa_base_m_[VALINE] = 18;

  // aa_base_w_[ALANINE] = 5;  
  // aa_base_w_[ARGININE] = 0;    
  // aa_base_w_[ASPARAGINE] = 7;
  // aa_base_w_[ASPARTIC_ACID] = 4;
  // aa_base_w_[CYSTEINE] = 13;
  // aa_base_w_[GLUTAMIC_ACID] = 15;
  // aa_base_w_[GLUTAMINE] = 10; 
  // aa_base_w_[GLYCINE] = 12;
  // aa_base_w_[HISTIDINE] = 19;
  // aa_base_w_[ISOLEUCINE] = 11; 
  // aa_base_w_[LEUCINE] = 3; 
  // aa_base_w_[LYSINE] = 1;       
  // aa_base_w_[METHIONINE] = 18;
  // aa_base_w_[PHENYLALANINE] = 14;    
  // aa_base_w_[PROLINE] = 8;      
  // aa_base_w_[SERINE] = 17;     
  // aa_base_w_[THREONINE] = 6;   
  // aa_base_w_[TRYPTOPHAN] = 2;    
  // aa_base_w_[TYROSINE] = 9;   
  // aa_base_w_[VALINE] = 16;  
  
  // aa_base_h_[ALANINE] = 13;  
  // aa_base_h_[ARGININE] = 3;    
  // aa_base_h_[ASPARAGINE] = 2;
  // aa_base_h_[ASPARTIC_ACID] = 15;
  // aa_base_h_[CYSTEINE] = 14;
  // aa_base_h_[GLUTAMIC_ACID] = 10;
  // aa_base_h_[GLUTAMINE] = 7; 
  // aa_base_h_[GLYCINE] = 9;
  // aa_base_h_[HISTIDINE] = 8;
  // aa_base_h_[ISOLEUCINE] = 17; 
  // aa_base_h_[LEUCINE] = 12; 
  // aa_base_h_[LYSINE] = 5;       
  // aa_base_h_[METHIONINE] = 19;
  // aa_base_h_[PHENYLALANINE] = 16;    
  // aa_base_h_[PROLINE] = 6;      
  // aa_base_h_[SERINE] = 18;     
  // aa_base_h_[THREONINE] = 11;   
  // aa_base_h_[TRYPTOPHAN] = 1;    
  // aa_base_h_[TYROSINE] = 4;   
  // aa_base_h_[VALINE] = 0;  
  
  
  /* Standard code
aa_base_m_[PHENYLALANINE] = 0;  // 2
  aa_base_m_[LEUCINE] = 1;        // 6
  aa_base_m_[ISOLEUCINE] = 2;     // 3+1
  aa_base_m_[METHIONINE] = 2;
  aa_base_m_[VALINE] = 3;         // 4
  aa_base_m_[SERINE] = 4;         // 6
                                // TOTAL = 2+6+4+4+6 = 22

  aa_base_w_[PROLINE] = 0;        // 4
  aa_base_w_[THREONINE] = 1;      // 4
  aa_base_w_[ALANINE] = 2;        // 4
  aa_base_w_[TYROSINE] = 3;       // 2+2
  aa_base_w_[HISTIDINE] = 3;
  aa_base_w_[GLUTAMINE] = 4;      // 2+2
  aa_base_w_[ASPARAGINE] = 4;
  aa_base_w_[LYSINE] = 5;         // 2
                                // TOTAL = 4+4+4+4+4+2 = 22

  aa_base_h_[ASPARAGINE] = 0;     // 2+2
  aa_base_h_[LYSINE] = 0;
  aa_base_h_[ASPARTIC_ACID] = 1;  // 2+2+2
  aa_base_h_[CYSTEINE] = 1;
  aa_base_h_[GLUTAMIC_ACID] = 1;
  aa_base_h_[TRYPTOPHAN] = 2;     // 1+4
  aa_base_h_[ARGININE] = 3;       // 4
  aa_base_h_[GLYCINE] = 4;        // 4
                                // TOTAL = 4+6+5+4+4 = 23
				*/
  // Compute bases sizes
  aa_base_m_size_ = 0;
  aa_base_w_size_ = 0;
  aa_base_h_size_ = 0;

  for(int8_t i = 0; i < NB_AMINO_ACIDS; i++) {
    if(aa_base_m_[i] > aa_base_m_size_)
      aa_base_m_size_ = aa_base_m_[i];

    if(aa_base_w_[i] > aa_base_w_size_)
      aa_base_w_size_ = aa_base_w_[i];

    if(aa_base_h_[i] > aa_base_h_size_)
      aa_base_h_size_ = aa_base_h_[i];
  }
  aa_base_m_size_++; // base_size = max_digit + 1
  aa_base_w_size_++;
  aa_base_h_size_++;
#endif

  simd_metadata_flavor_         = STD_MAP;

#ifdef __REGUL
  _protein_presence_limit = 1e-2;
  _degradation_rate  = 1;
  _nb_degradation_step  = 10;
  _with_heredity          = false;
  _nb_indiv_age      = 20;

  _hill_shape_n      = 4;
  _hill_shape_theta  = 0.5;
  _hill_shape        = std::pow( _hill_shape_theta, _hill_shape_n );

  _list_eval_step    = new std::set<int>();
#endif
}


// ===========================================================================
//                                 Destructor
// ===========================================================================
ExpSetup::~ExpSetup() noexcept {
  delete sel_;
#ifdef __REGUL
  delete _list_eval_step;
#endif
}

#ifdef BASE_4

// =====================================================================
//                          Accessors' definitions
// =====================================================================
void ExpSetup::set_aa_base_m(const int8_t *base_m) {
  aa_base_m_size_ = 0;
  for(int8_t i = 0; i < NB_AMINO_ACIDS; i++) {
    aa_base_m_[i] = base_m[i];

    if(base_m[i] > aa_base_m_size_)
      aa_base_m_size_ = base_m[i];
  }
  aa_base_m_size_++; // base_size = base_max + 1
}

void ExpSetup::set_aa_base_w(const int8_t *base_w) {
  aa_base_w_size_ = 0;
  for(int8_t i = 0; i < NB_AMINO_ACIDS; i++) {
    aa_base_w_[i] = base_w[i];

    if(base_w[i] > aa_base_w_size_)
      aa_base_w_size_ = base_w[i];
  }
  aa_base_w_size_++; // base_size = base_max + 1
}

void ExpSetup::set_aa_base_h(const int8_t *base_h) {
  aa_base_h_size_ = 0;
  for(int8_t i = 0; i < NB_AMINO_ACIDS; i++) {
    aa_base_h_[i] = base_h[i];

    if(base_h[i] > aa_base_h_size_)
      aa_base_h_size_ = base_h[i];
  }
  aa_base_h_size_++; // base_size = base_max + 1
}
#endif

/*!
*/
void ExpSetup::write_setup_file(gzFile exp_setup_file) const {
  // legacy compatibility (write 32 bits)
  // TODO<dpa> remove legacy compatibility
  int32_t tmp_fuzzy_flavor = to_underlying(fuzzy_flavor_);
  gzwrite(exp_setup_file, &tmp_fuzzy_flavor, sizeof(tmp_fuzzy_flavor));

  gzwrite(exp_setup_file, &simd_metadata_flavor_, sizeof(simd_metadata_flavor_));
  // --------------------------------------------------------------- Transfer
  int8_t tmp_with_HT = static_cast<int8_t>(with_HT_);
  int8_t tmp_repl_HT_with_close_points = static_cast<int8_t>(repl_HT_with_close_points_);

  gzwrite(exp_setup_file,
          &tmp_with_HT,
          sizeof(tmp_with_HT));

  gzwrite(exp_setup_file,&tmp_repl_HT_with_close_points,sizeof(tmp_repl_HT_with_close_points));

  if (with_HT_) {
    gzwrite(exp_setup_file,
            &HT_ins_rate_,sizeof(HT_ins_rate_));
    gzwrite(exp_setup_file,&HT_repl_rate_,sizeof(HT_repl_rate_));
  }


  if(repl_HT_with_close_points_)
    gzwrite(exp_setup_file,
            &repl_HT_detach_rate_,sizeof(repl_HT_detach_rate_));


  // Dummy write (backward compatibility, was allow_plasmids)
  int8_t dummy = 0;
  gzwrite(exp_setup_file, &dummy, sizeof(dummy));

  // -------------------------------------------------------------- Secretion
  int8_t tmp_with_secretion_ = static_cast<int8_t>(with_secretion_);
  gzwrite(exp_setup_file,
          &tmp_with_secretion_,sizeof(tmp_with_secretion_));
  gzwrite(exp_setup_file,&secretion_contrib_to_fitness_,sizeof(secretion_contrib_to_fitness_));
  gzwrite(exp_setup_file,&secretion_cost_,sizeof(secretion_cost_));

  sel()->write_setup_file(exp_setup_file);

  mut_params_->save(exp_setup_file);

  #ifdef BASE_4
  gzwrite(exp_setup_file, &terminator_polya_sequence_length_, sizeof(terminator_polya_sequence_length_));
  #endif

  // ---------------------------------------------- Retrieve SIMD parameters
  gzwrite(exp_setup_file, &min_genome_length_,sizeof(min_genome_length_));
  gzwrite(exp_setup_file, &max_genome_length_,sizeof(max_genome_length_));


#ifdef __REGUL
  //printf("Nb indiv age : %d\n",_nb_indiv_age);
  gzwrite( exp_setup_file,&_hill_shape,sizeof(_hill_shape));
  gzwrite( exp_setup_file, &_hill_shape_n,sizeof(_hill_shape_n));
  gzwrite( exp_setup_file, &_hill_shape_theta,sizeof(_hill_shape_theta));
  gzwrite( exp_setup_file, &_degradation_rate,sizeof(_degradation_rate));
  gzwrite( exp_setup_file,  &_nb_degradation_step,sizeof(_nb_degradation_step));
  gzwrite( exp_setup_file, &_nb_indiv_age,sizeof(_nb_indiv_age));
  gzwrite( exp_setup_file, &_with_heredity,sizeof(_with_heredity));
  gzwrite( exp_setup_file, &_protein_presence_limit,sizeof(_protein_presence_limit));

  BindingMatrix::save();

  unsigned int eval_step_size = _list_eval_step->size();
  gzwrite(exp_setup_file, &eval_step_size,sizeof(eval_step_size));

  for(auto eval_step : *_list_eval_step) {
    gzwrite(exp_setup_file, &eval_step,sizeof(eval_step));
  }
#endif
}

void ExpSetup::load(gzFile setup_file, gzFile backup_file) {
  // legacy compatibility (read 32 bits)
  // TODO<dpa> remove legacy compatibility
  int32_t tmp_fuzzy_flavor;
  gzread(setup_file, &tmp_fuzzy_flavor, sizeof(tmp_fuzzy_flavor));
  fuzzy_flavor_ = static_cast<FuzzyFlavor>(tmp_fuzzy_flavor);

  gzread(setup_file,&simd_metadata_flavor_,sizeof(simd_metadata_flavor_));

  // -------------------------------------------- Retrieve transfer parameters
  int8_t tmp_with_HT;
  int8_t tmp_repl_HT_with_close_points;


  gzread(setup_file,
         &tmp_with_HT,sizeof(tmp_with_HT)
  );

  gzread(setup_file,
         &tmp_repl_HT_with_close_points,sizeof(tmp_repl_HT_with_close_points)
  );

  with_HT_ = static_cast<bool>(tmp_with_HT);

  repl_HT_with_close_points_ = static_cast<bool>(tmp_repl_HT_with_close_points);



  if (with_HT_)
  {
    gzread(setup_file,
           &HT_ins_rate_,sizeof(HT_ins_rate_));
    gzread(setup_file,&HT_repl_rate_,sizeof(HT_repl_rate_));
  }
   if(repl_HT_with_close_points_)
    gzread(setup_file, &repl_HT_detach_rate_,sizeof(repl_HT_detach_rate_));


   // Dummy read (backward compatibility, was allow_plasmids)
   int8_t dummy;
   gzread(setup_file, &dummy, sizeof(dummy));

  // ------------------------------------------ Retrieve secretion parameters
  int8_t tmp_with_secretion;
  gzread(setup_file, &tmp_with_secretion,sizeof(tmp_with_secretion));
  gzread(setup_file, &secretion_contrib_to_fitness_,sizeof(secretion_contrib_to_fitness_));
  gzread(setup_file, &secretion_cost_,sizeof(secretion_cost_));
  with_secretion_ = static_cast<bool>(tmp_with_secretion);
  // ---------------------------------------------- Retrieve selection context

  sel()->load(setup_file, backup_file);

  mut_params_ = std::make_shared<MutationParams>(setup_file);

  #ifdef BASE_4
  // ---------------------------------------------- Retrieve terminators configuration
  gzread(setup_file, &terminator_polya_sequence_length_, sizeof(terminator_polya_sequence_length_));
  #endif

  // ---------------------------------------------- Retrieve SIMD parameters
  gzread(setup_file, &min_genome_length_,sizeof(min_genome_length_));
  gzread(setup_file, &max_genome_length_,sizeof(max_genome_length_));

#ifdef __REGUL
  gzread( setup_file, &_hill_shape,sizeof(_hill_shape));
  gzread( setup_file, &_hill_shape_n,sizeof(_hill_shape_n));
  gzread( setup_file, &_hill_shape_theta,sizeof(_hill_shape_theta));
  gzread( setup_file, &_degradation_rate,sizeof(_degradation_rate));
  gzread( setup_file, &_nb_degradation_step,sizeof(_nb_degradation_step));
  gzread( setup_file, &_nb_indiv_age,sizeof(_nb_indiv_age));
  gzread( setup_file, &_with_heredity,sizeof(_with_heredity));
  gzread( setup_file, &_protein_presence_limit,sizeof(_protein_presence_limit));


  //printf("Nb indiv age : %d\n",_nb_indiv_age);

  BindingMatrix::load();

  unsigned int eval_step_size;
  gzread(setup_file, &eval_step_size,sizeof(eval_step_size));

  int eval_val;
  for(unsigned int i = 0; i < eval_step_size; i++) {
    gzread(setup_file,&eval_val,sizeof(eval_val));
    _list_eval_step->insert(eval_val);
  }
#endif
}

// ===========================================================================
//                                Protected Methods
// ===========================================================================
} // namespace aevol
