// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************

#ifndef AEVOL_EXP_SETUP_H_
#define AEVOL_EXP_SETUP_H_

// =================================================================
//                              Includes
// =================================================================
#include <cinttypes>
#include <cstdlib>
#include <set>

#include "Selection.h"
#include "legacy/io/stats/Stats.h"
#include "Logging.h"
#include "Tree.h"
#include "JumpingMT.h"
#include "fuzzytypes.h"

namespace aevol {

// ===========================================================================
//                          Class declarations
// ===========================================================================
class ExpSetup {
  friend class ExpManager;

  public:
  // =======================================================================
  //                             Constructors
  // =======================================================================
  ExpSetup() = delete;
  ExpSetup(const ExpSetup&) = delete;
  ExpSetup(ExpManager* exp_m);

  // =======================================================================
  //                             Destructors
  // =======================================================================
  virtual ~ExpSetup() noexcept;

  // =======================================================================
  //                         Accessors: getters
  // =======================================================================
  inline FuzzyFlavor get_fuzzy_flavor() const;
  inline int get_simd_metadata_flavor( void ) const;

  // ----------------------------------------------------- Selection context
  Selection * sel() const { return sel_; }

  // --------------------------------------------------------------- Transfer
  bool with_HT() const { return with_HT_; };
  bool repl_HT_with_close_points( void ) const { return repl_HT_with_close_points_; };
  double HT_ins_rate() const { return HT_ins_rate_; };
  double HT_repl_rate() const { return HT_repl_rate_; };
  double repl_HT_detach_rate() const { return repl_HT_detach_rate_; };

  // -------------------------------------------------------------- Secretion
  bool   with_secretion() const { return with_secretion_; }
  double secretion_contrib_to_fitness() const { return secretion_contrib_to_fitness_; }
  double secretion_cost() const { return secretion_cost_; }

  #ifdef BASE_4
  // -------------------------------------------------------------- Terminators
  int8_t terminator_polya_sequence_length() const { return terminator_polya_sequence_length_; }

  // -------------------------------------------------------------- MWH bases configuration
  const int8_t * aa_base_m() const { return aa_base_m_; }
  const int8_t * aa_base_w() const { return aa_base_w_; }
  const int8_t * aa_base_h() const { return aa_base_h_; }

  int8_t aa_base_m_size() const { return aa_base_m_size_; }
  int8_t aa_base_w_size() const { return aa_base_w_size_; }
  int8_t aa_base_h_size() const { return aa_base_h_size_; }
  #endif
  // ------------------------------------------------------------ Parameter for SIMD
  std::shared_ptr<MutationParams> mut_params() { return mut_params_;}
  int32_t min_genome_length() const { return min_genome_length_; }
  int32_t max_genome_length() const { return max_genome_length_; }


#ifdef __REGUL
    inline bool   get_with_heredity( void ) const;
    inline double get_degradation_rate( void ) const;
    inline int    get_nb_degradation_step( void ) const;
    inline double get_protein_presence_limit( void ) const;

    inline double get_hill_shape( void ) const;
    inline double get_hill_shape_n( void ) const;
    inline double get_hill_shape_theta( void ) const;

    inline int get_nb_indiv_age( void ) const;
    std::set<int>* get_list_eval_step( void ) const {
      return _list_eval_step;
    }
#endif

  // =======================================================================
  //                         Accessors: setters
  // =======================================================================
  inline void set_fuzzy_flavor(FuzzyFlavor fuzzy_flavor);
    inline void set_simd_metadata_flavor(int metadata_flavor );
  // --------------------------------------------------------------- Transfer
  void set_with_HT(bool with_HT) { with_HT_ = with_HT; }
  void set_repl_HT_with_close_points(bool repl_HT_with_close_points) { repl_HT_with_close_points_ = repl_HT_with_close_points; }
  void set_HT_ins_rate(double HT_ins_rate) { HT_ins_rate_ = HT_ins_rate; }
  void set_HT_repl_rate(double HT_repl_rate) { HT_repl_rate_ = HT_repl_rate; }
  void set_repl_HT_detach_rate(double repl_HT_detach_rate) { repl_HT_detach_rate_ = repl_HT_detach_rate; }

  // -------------------------------------------------------------- Secretion
  void set_with_secretion(bool with_secretion) { with_secretion_ = with_secretion; }
  void set_secretion_contrib_to_fitness(double secretion_contrib) { secretion_contrib_to_fitness_ = secretion_contrib; }
  void set_secretion_cost(double secretion_cost) { secretion_cost_ = secretion_cost; }

  #ifdef BASE_4
  // -------------------------------------------------------------- Terminators
  void set_terminator_polya_sequence_length(int8_t sequence_length) { terminator_polya_sequence_length_ = sequence_length; }

  // -------------------------------------------------------------- MWH bases configuration
  void set_aa_base_m(const int8_t *base_m);
  void set_aa_base_w(const int8_t *base_w);
  void set_aa_base_h(const int8_t *base_h);

  void set_aa_base_m_size(int8_t size) { aa_base_m_size_ = size; }
  void set_aa_base_w_size(int8_t size) { aa_base_w_size_ = size; }
  void set_aa_base_h_size(int8_t size) { aa_base_h_size_ = size; }
  #endif

    // ------------------------------------------------------------ Parameter for SIMD
    void set_min_genome_length(int32_t min_genome_length) { min_genome_length_ = min_genome_length; }
    void set_max_genome_length(int32_t max_genome_length) { max_genome_length_ = max_genome_length; }

    void set_mutation_parameters(std::shared_ptr<MutationParams> mut_params) { mut_params_ = mut_params; }

#ifdef __REGUL
    inline void set_with_heredity( bool with_heredity );
    inline void set_degradation_rate( double degradation_rate );
    inline void set_nb_degradation_step( int nb_degradation_step );
    inline void set_protein_presence_limit( double protein_presence_limit );

    inline void set_hill_shape( double hill_shape );
    inline void set_hill_shape_n( double hill_shape_n );
    inline void set_hill_shape_theta( double hill_shape_theta );

    inline void set_nb_indiv_age( int nb_indiv_age );

    inline void set_list_eval_step(std::set<int> list_eval_step);
#endif

  // =======================================================================
  //                            Public Methods
  // =======================================================================
  void write_setup_file(gzFile exp_setup_file) const;
  void save(gzFile backup_file) const;
  void load(gzFile setup_file, gzFile backup_file);

  // =======================================================================
  //                           Public Attributes
  // =======================================================================

 protected :
  // =======================================================================
  //                           Protected Methods
  // =======================================================================
  virtual void display() {};

  // =======================================================================
  //                          Protected Attributes
  // =======================================================================
  ExpManager* exp_m_;

  FuzzyFlavor fuzzy_flavor_ = FuzzyFlavor::VECTOR;
  int simd_metadata_flavor_;
  // ----------------------------------------------------- Selection context
  Selection* sel_;

  // --------------------------------------------------- Transfer parameters
  bool   with_HT_ = false;
  bool   repl_HT_with_close_points_ = false;
  double HT_ins_rate_;
  double HT_repl_rate_;
  double repl_HT_detach_rate_;

  // -------------------------------------------------- Secretion parameters
  bool   with_secretion_;
  double secretion_contrib_to_fitness_;
  double secretion_cost_;

  #ifdef BASE_4
  // -------------------------------------------------- Terminator parameters
  int8_t terminator_polya_sequence_length_;

  // -------------------------------------------------- MWH bases configuration
  int8_t aa_base_m_[NB_AMINO_ACIDS];
  int8_t aa_base_w_[NB_AMINO_ACIDS];
  int8_t aa_base_h_[NB_AMINO_ACIDS];

  int8_t aa_base_m_size_;
  int8_t aa_base_w_size_;
  int8_t aa_base_h_size_;
  #endif

    // Mutation rates etc...
    std::shared_ptr<MutationParams> mut_params_;

    int32_t min_genome_length_;
    int32_t max_genome_length_;

#ifdef __REGUL
    bool    _with_heredity;
    double  _protein_presence_limit;

    double  _degradation_rate;
    int     _nb_degradation_step;

    double _hill_shape_n;
    double _hill_shape;
    double _hill_shape_theta;

    int    _nb_indiv_age;

    std::set<int>* _list_eval_step;
#endif
};


// =====================================================================
//                           Getters' definitions
// =====================================================================

inline FuzzyFlavor ExpSetup::get_fuzzy_flavor() const
{
  return fuzzy_flavor_;
}

inline int ExpSetup::get_simd_metadata_flavor( void ) const
{
   return simd_metadata_flavor_;
}

#ifdef __REGUL
inline bool ExpSetup::get_with_heredity( void ) const
{
  return _with_heredity;
}

inline double ExpSetup::get_degradation_rate( void ) const
{
  return _degradation_rate;
}


inline int ExpSetup::get_nb_degradation_step( void ) const
{
  return _nb_degradation_step;
}

inline double ExpSetup::get_protein_presence_limit( void ) const
{
  return _protein_presence_limit;
}

inline double ExpSetup::get_hill_shape( void ) const
{
  return _hill_shape;
}

inline double ExpSetup::get_hill_shape_n( void ) const
{
  return _hill_shape_n;
}

inline double ExpSetup::get_hill_shape_theta( void ) const
{
  return _hill_shape_theta;
}

inline int ExpSetup::get_nb_indiv_age( void ) const
{
  return _nb_indiv_age;
}

#endif

// =====================================================================
//                           Setters' definitions
// =====================================================================
// --------------------------------------------------------------- Transfer
inline void ExpSetup::set_fuzzy_flavor(FuzzyFlavor fuzzy_flavor)
{
  fuzzy_flavor_ = fuzzy_flavor;
}

inline void ExpSetup::set_simd_metadata_flavor(int metadata_flavor)
{
  simd_metadata_flavor_ = metadata_flavor;
}

#ifdef __REGUL
inline void ExpSetup::set_with_heredity( bool with_heredity )
{
  _with_heredity = with_heredity;
}

inline void ExpSetup::set_degradation_rate( double degradation_rate )
{
  _degradation_rate = degradation_rate;
}

inline void ExpSetup::set_nb_degradation_step( int degradation_step )
{
  _nb_degradation_step = degradation_step;
}

inline void ExpSetup::set_protein_presence_limit( double protein_presence_limit )
{
  _protein_presence_limit = protein_presence_limit;
}

inline void ExpSetup::set_hill_shape( double hill_shape )
{
  _hill_shape = hill_shape;
}

inline void ExpSetup::set_hill_shape_theta( double hill_shape_theta )
{
  _hill_shape_theta = hill_shape_theta;
}

inline void ExpSetup::set_hill_shape_n( double hill_shape_n )
{
  _hill_shape_n = hill_shape_n;
}


inline void ExpSetup::set_nb_indiv_age( int nb_indiv_age )
{
  _nb_indiv_age = nb_indiv_age;
}

inline void ExpSetup::set_list_eval_step( std::set<int> list_eval_step )
{
  _list_eval_step = new std::set<int>(list_eval_step);
}
#endif

} // namespace aevol

#endif // AEVOL_EXP_SETUP_H_
