// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//*****************************************************************************




// =================================================================
//                              Includes
// =================================================================
#include <inttypes.h>
#include <cstring>

#include <list>
#include <vector>
#include <memory>

#include <gtest/gtest.h>

#include "ExpManager_7.h"
#include "ExpSetup.h"
#include "fuzzy/FuzzyFactory_7.h"
#include "Individual_7.h"
#include "legacy/biochemistry/Rna.h"
#include "legacy/biochemistry/Protein.h"
#include "macros.h"
#include "MutationParams.h"
#include "Strand.h"

using namespace aevol;

//############################################################################
//                                                                           #
//                         Class IndividualTest                              #
//                                                                           #
//############################################################################
class IndividualTest : public testing::Test
{
 protected:
  virtual void SetUp(void);
  virtual void TearDown(void);

  // We have an version of each individual for each fuzzy set flavor
  std::vector<Individual_7*> indivs1;
  std::vector<Individual_7*> indivs2;
  std::vector<Individual_7*> indivs3;
  std::vector<Individual_7*> indivs4;
};

// ===========================================================================
//                                 Public Methods
// ===========================================================================
void IndividualTest::SetUp(void) {
  // Build ad-hoc genomes
  // (and reverse to test the same things on the lagging strand.):
  //
  // indiv1: (AS + prom + AS + AG + AS + term + AS + prom + AS)
  // indiv2: reverse
  // indiv3: (AS + AG + AS + term + AS + prom + AS)
  // indiv4: reverse
  //
  // AS = Arbitrary Sequence
  // AG = Arbitrary Gene
  // Do not modify the sequences !

  // Define arbitrary sequences
  std::array<std::string, 5> as = {
      "0011",
      "11101",
      "110011",
      "11000",
      "000101"
  };

  // Define an arbitrary gene
  std::string gene = std::string(SHINE_DAL_SEQ) + "0011000100110110010001";

  // Define an arbitrary terminator
  std::string term = "01000001101";

  // Define a few arbitrary promoters
  std::array<std::string, 2> prom = {
      "0101010001110110010110", // dist from consensus: 2 => basal level: 0.6
      "0101011001110010010010"  // dist from consensus: 1 => basal level: 0.8
  };

  // Initialize the experimental setup.
  // These are needed in the GeneticUnit constructors.
  ExpSetup exp_s(nullptr);
  MutationParams params_mut;


  // The following commented code allows to print stuff about rnas and proteins
#if 0
  printf("%" PRId32 " rnas and %" PRId32 " prots\n", indiv4->rna_list().size(),
         indiv4->protein_list().size());

  for (const Rna* rna_node: indiv4->rna_list()) {
    printf("%s rna at pos %" PRId32 " (%f, %" PRId32 ")\n",
           rna_node->strand() == LEADING ? "LEADING" : "LAGGING",
           rna_node->promoter_pos(), rna_node->basal_level(),
           rna_node->transcript_length());
  }

  for (const Protein* protein_node: indiv4->protein_list()) {
    printf("%s protein at pos %" PRId32 " (length: %" PRId32
           ", concentr: %f, nb_rnas: %" PRId32 ")\n",
           protein_node->strand() == LEADING ? "LEADING" : "LAGGING",
           protein_node->shine_dal_pos(), protein_node->length(),
           protein_node->concentration(), protein_node->rna_list().size());
  }
#endif


  // We initialize a version of each individual for each fuzzy set implementation
  for (FuzzyFlavor fuzzyFlavor : {FuzzyFlavor::VECTOR, FuzzyFlavor::DISCRETE_DOUBLE_TABLE}) {
    // Update fuzzy factory // TODO<dpa> Still needed/useful?
    exp_s.set_fuzzy_flavor(fuzzyFlavor);

    FuzzyFactory_7* fuzzy_factory = new FuzzyFactory_7(fuzzyFlavor, 16, 300, 16);

    // Build indiv1
    auto genome1 = as[0] + prom[0] + as[1] + gene + as[2] + term + as[3] + prom[1] + as[4];
    auto indiv1 = Individual_7::make_from_sequence(1.0, fuzzy_factory, genome1.c_str(), genome1.size());

    // Do transcription and translation
    indiv1->start_stop_RNA();
    indiv1->prom_compute_RNA(exp_s);
    indiv1->start_protein();
    indiv1->compute_protein();
    indiv1->translate_protein(exp_s, 1.0);

    indivs1.push_back(indiv1);

    // Build indiv2 (reverse the whole genome of indiv1)
    auto genome2 = indiv1->annotated_chromosome_->dna_->subseq(indiv1->annotated_chromosome_->dna_->length() - 1,
                                                               indiv1->annotated_chromosome_->dna_->length(),
                                                               Strand::LAGGING);
    auto indiv2 = Individual_7::make_from_sequence(1.0, fuzzy_factory, genome2.c_str(), genome2.size());

    // Do transcription and translation
    indiv2->start_stop_RNA();
    indiv2->prom_compute_RNA(exp_s);
    indiv2->start_protein();
    indiv2->compute_protein();
    indiv2->translate_protein(exp_s, 1.0);

    indivs2.push_back(indiv2);

    // Build indiv3
    auto genome3 = as[0] + gene + as[1] + term + as[2] + prom[1] + as[3];
    auto indiv3 = Individual_7::make_from_sequence(1.0, fuzzy_factory, genome3.c_str(), genome3.size());

    // Do transcription and translation
    indiv3->start_stop_RNA();
    indiv3->prom_compute_RNA(exp_s);
    indiv3->start_protein();
    indiv3->compute_protein();
    indiv3->translate_protein(exp_s, 1.0);

    indivs3.push_back(indiv3);

    // Build indiv4 (reverse the whole genome of indiv3)
    auto genome4 = indiv3->annotated_chromosome_->dna_->subseq(70,
                                                               indiv3->annotated_chromosome_->dna_->length(),
                                                               Strand::LAGGING);
    auto indiv4 = Individual_7::make_from_sequence(1.0, fuzzy_factory, genome4.c_str(), genome4.size());

    // Do transcription and translation
    indiv4->start_stop_RNA();
    indiv4->prom_compute_RNA(exp_s);
    indiv4->start_protein();
    indiv4->compute_protein();
    indiv4->translate_protein(exp_s, 1.0);

    indivs4.push_back(indiv4);

    delete fuzzy_factory;
  }
}

void IndividualTest::TearDown() {
  for (auto indiv : indivs1) {
    delete indiv;
  }
}

// For each version of each individual constructed with a different
// fuzzy set implementation, we check that all values are correct.
// We don't need to change the fuzzy set implementation in the
// experimental setup again because it's only used at transcription and
// translation time.

TEST_F(IndividualTest, TestIndiv1) {
  // Check that we have the right number of promoters, terminators etc
  // and at the right positions
  // "right" means those values we have computed by hand

  for (auto fuzzyFlavor = 0 ; fuzzyFlavor <= 1 ; ++fuzzyFlavor) {
    auto indiv = indivs1[fuzzyFlavor];

    // Check genome size
    EXPECT_EQ(109, indiv->annotated_chromosome_->dna_->length());

    // Check RNA list
    auto rna_list = indiv->annotated_chromosome_->rnas();
    EXPECT_EQ(2, rna_list.size());
    const auto* rna = rna_list.front();
    EXPECT_EQ(Strand::LEADING, rna->strand_);
    EXPECT_EQ(4, rna->prom->pos);
    EXPECT_FLOAT_EQ(0.6, rna->e);
    EXPECT_EQ(50, rna->length);
    rna = rna_list.back();
    EXPECT_EQ(Strand::LEADING, rna->strand_);
    EXPECT_EQ(81, rna->prom->pos);
    EXPECT_FLOAT_EQ(0.8, rna->e);
    EXPECT_EQ(82, rna->length);

    // Check protein list
    auto prot_list = indiv->annotated_chromosome_->proteins();
    EXPECT_EQ(2, prot_list.size());

    const auto* prot = prot_list.front();
    EXPECT_EQ(Strand::LEADING, prot->strand_);
    EXPECT_EQ(44, prot->protein_start);
    EXPECT_EQ(4, prot->protein_length);
    EXPECT_FLOAT_EQ(1.4, prot->e);
    EXPECT_EQ(true, prot->is_init_);
    EXPECT_EQ(2, prot->rna_list_.size());

    prot = prot_list.back();
    EXPECT_EQ(Strand::LEADING, prot->strand_);
    EXPECT_EQ(44, prot->protein_start);
    EXPECT_EQ(4, prot->protein_length);
    EXPECT_FLOAT_EQ(0.6, prot->e);
    EXPECT_EQ(false, prot->is_init_);
    EXPECT_EQ(1, prot->rna_list_.size());
  }
}

TEST_F(IndividualTest, TestIndiv2) {
  for (auto fuzzyFlavor = 0 ; fuzzyFlavor <= 1 ; ++fuzzyFlavor) {
    auto indiv = indivs2[fuzzyFlavor];

    // Check genome size
    EXPECT_EQ(109, indiv->annotated_chromosome_->dna_->length());

    // Check RNA list
    auto rna_list = indiv->annotated_chromosome_->rnas();
    EXPECT_EQ(2, rna_list.size());
    const auto* rna = rna_list.front();
    EXPECT_EQ(Strand::LAGGING, rna->strand_);
    EXPECT_EQ(104, rna->prom->pos);
    EXPECT_FLOAT_EQ(0.6, rna->e);
    EXPECT_EQ(50, rna->length);
    rna = rna_list.back();
    EXPECT_EQ(Strand::LAGGING, rna->strand_);
    EXPECT_EQ(27, rna->prom->pos);
    EXPECT_FLOAT_EQ(0.8, rna->e);
    EXPECT_EQ(82, rna->length);

    // Check protein list
    auto prot_list = indiv->annotated_chromosome_->proteins();
    EXPECT_EQ(2, prot_list.size());

    const auto* prot = prot_list.front();
    EXPECT_EQ(Strand::LAGGING, prot->strand_);
    EXPECT_EQ(64, prot->protein_start);
    EXPECT_EQ(4, prot->protein_length);
    EXPECT_FLOAT_EQ(1.4, prot->e);
    EXPECT_EQ(true, prot->is_init_);
    EXPECT_EQ(2, prot->rna_list_.size());

    prot = prot_list.back();
    EXPECT_EQ(Strand::LAGGING, prot->strand_);
    EXPECT_EQ(64, prot->protein_start);
    EXPECT_EQ(4, prot->protein_length);
    EXPECT_FLOAT_EQ(0.6, prot->e);
    EXPECT_EQ(false, prot->is_init_);
    EXPECT_EQ(1, prot->rna_list_.size());
  }
}

TEST_F(IndividualTest, TestIndiv3) {
  for (auto fuzzyFlavor = 0 ; fuzzyFlavor <= 1 ; ++fuzzyFlavor) {
    auto indiv = indivs3[fuzzyFlavor];

    // Check genome size
    EXPECT_EQ(81, indiv->annotated_chromosome_->dna_->length());

    // Check RNA list
    auto rna_list = indiv->annotated_chromosome_->rnas();
    EXPECT_EQ(1, rna_list.size());
    const auto* rna = rna_list.front();
    EXPECT_EQ(Strand::LEADING, rna->strand_);
    EXPECT_EQ(54, rna->prom->pos);
    EXPECT_FLOAT_EQ(0.8, rna->e);
    EXPECT_EQ(42, rna->length);

    // Check protein list
    auto prot_list = indiv->annotated_chromosome_->proteins();
    EXPECT_EQ(1, prot_list.size());

    const auto* prot = prot_list.front();
    EXPECT_EQ(Strand::LEADING, prot->strand_);
    EXPECT_EQ(17, prot->protein_start);
    EXPECT_EQ(4, prot->protein_length);
    EXPECT_FLOAT_EQ(0.8, prot->e);
    EXPECT_EQ(true, prot->is_init_);
    EXPECT_EQ(1, prot->rna_list_.size());
  }
}

TEST_F(IndividualTest, TestIndiv4) {
  for (auto fuzzyFlavor = 0 ; fuzzyFlavor <= 1 ; ++fuzzyFlavor) {
    auto indiv = indivs4[fuzzyFlavor];

    // Check genome size
    EXPECT_EQ(81, indiv->annotated_chromosome_->dna_->length());

    // Check RNA list
    auto rna_list = indiv->annotated_chromosome_->rnas();
    EXPECT_EQ(1, rna_list.size());
    const auto* rna = rna_list.front();
    EXPECT_EQ(Strand::LAGGING, rna->strand_);
    EXPECT_EQ(16, rna->prom->pos);
    EXPECT_FLOAT_EQ(0.8, rna->e);
    EXPECT_EQ(42, rna->length);

    // Check protein list
    auto prot_list = indiv->annotated_chromosome_->proteins();
    EXPECT_EQ(1, prot_list.size());

    const auto* prot = prot_list.front();
    EXPECT_EQ(Strand::LAGGING, prot->strand_);
    EXPECT_EQ(53, prot->protein_start);
    EXPECT_EQ(4, prot->protein_length);
    EXPECT_FLOAT_EQ(0.8, prot->e);
    EXPECT_EQ(true, prot->is_init_);
    EXPECT_EQ(1, prot->rna_list_.size());
  }
}
