// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************

#include "Dna_7.h"

#include <algorithm>

#include "DnaMutator.h"
#include "Individual_7.h"
#include "List_Metadata.h"
#include "legacy/ExpManager.h"
#include "ExpManager_7.h"

#define REDUCTION_FACTOR 16

namespace aevol {

std::unique_ptr<DnaFactory> Dna_7::dna_factory_ = std::make_unique<DnaFactory>(DnaFactory_Policy::ALLOCATE, 0, 0);

Dna_7::Dna_7(size_type length) {
    length_ = length;

    nb_blocks_ = nb_required_blocks(length_);
    posix_memalign((void **) &data_, 64, nb_blocks_ * BLOCK_SIZE * sizeof(char));//new char[nb_blocks_ * BLOCK_SIZE];
    memset(data_, 0, (length_ + 1) * sizeof(char));
}

Dna_7* Dna_7::make_dna(size_t request_size) {
  return dna_factory_->get_dna(request_size);
}

Dna_7* Dna_7::make_from_sequence(const char* seq, size_type len) {
  auto dna = dna_factory_->get_dna(len);
  dna->reset_sequence(seq, len);
  return dna;
}

void Dna_7::release(Dna_7*&& dna) {
  dna_factory_->give_back(dna);
}

void Dna_7::reserve(size_type new_capacity) {
  auto new_nb_blocks = nb_required_blocks(new_capacity);

  if ((new_nb_blocks > nb_blocks_) || (new_nb_blocks < nb_blocks_ / REDUCTION_FACTOR)) {
    nb_blocks_ = new_nb_blocks;
    if (data_ != nullptr) {
      free(data_);
      data_ = nullptr;
    }

    posix_memalign((void **) &data_, 64, nb_blocks_ * BLOCK_SIZE * sizeof(char));
    memset(data_, 0, (new_capacity + 1) * sizeof(char));
  }
}

void Dna_7::reset_sequence(const char* seq, size_type len) {
  reserve(len);
  length_ = len;
  memcpy(data_, seq, (length_+1) * sizeof(char));
}

void Dna_7::set_indiv(Dna_7 * dna, Individual_7 * indiv) {
  reset_sequence(dna->data(), dna->length());

    parent_length_ = dna->length_;
    indiv_ = indiv;
}

void Dna_7::set_indiv(const char* dna, size_type dna_length, Individual_7* indiv) {
  reset_sequence(dna, dna_length);

    parent_length_ = dna_length;

    indiv_ = indiv;
}

void Dna_7::set_indiv(size_type req_length, size_type parent_length, Individual_7 * indiv) {
    reserve(req_length);
    length_ = req_length;

    parent_length_ = parent_length;
    indiv_ = indiv;
}

void Dna_7::set_chrsm(AnnotatedChromosome* ann_chrsm){
  ann_chrsm_ = ann_chrsm;
}

Dna_7::~Dna_7() {
  if (data_ != nullptr) { free(data_); data_=nullptr;}

if (mutation_list.size() > 0) {
    for (auto repl : mutation_list)
        delete repl;
    mutation_list.clear();
}


}

void Dna_7::remove(size_type pos_1, size_type pos_2) {
  assert(pos_1 >= 0 && pos_2 >= pos_1 && pos_2 <= length_);

  // Compute size of new genome
  auto new_length = length_ - (pos_2 - pos_1);

  Dna_7* new_genome = Dna_7::make_dna(new_length);
  new_genome->set_chrsm(ann_chrsm_);
  new_genome->set_indiv(new_length,parent_length_,indiv_);

  // Copy the remaining of the genome in tmp (preceeding and following parts)
  memcpy(new_genome->data_, data_, pos_1 * sizeof(char));
  memcpy(&(new_genome->data_[pos_1]), &data_[pos_2],
         (new_length - pos_1) * sizeof(char));
  new_genome->data_[new_length] = '\0';

  // Replace previous genome with the new one
  char* old_data = data_;
    int new_nb_block = new_genome->nb_blocks_;
    new_genome->nb_blocks_ = nb_blocks_;
    data_ = new_genome->data_;
  new_genome->data_ = old_data;
  new_genome->length_ = length();

  release(std::move(new_genome));

  // Update length data
  length_ = new_length;
  nb_blocks_ = new_nb_block;
}

void Dna_7::insert(size_type pos, const char* seq, size_type seq_length) {
// Insert sequence 'seq' at position 'pos'
  assert(pos >= 0 && pos < length_);

  // If the sequence's length was not provided, compute it
  if (seq_length == -1) {
    seq_length = strlen(seq);
  }

  // Compute size of new genome
  auto new_length = length_ + seq_length;

  Dna_7* new_genome = Dna_7::make_dna(new_length);
  new_genome->set_chrsm(ann_chrsm_);
  new_genome->set_indiv(new_length,parent_length_,indiv_);

  // Build new genome from previous genome and sequence to insert
  memcpy(new_genome->data_, data_, pos * sizeof(char));
  memcpy(&(new_genome->data_[pos]), seq, seq_length * sizeof(char));
  memcpy(&(new_genome->data_[pos+seq_length]), &data_[pos],
         (length_ - pos) * sizeof(char));
  new_genome->data_[new_length] = '\0';

  // Replace the previous genome with the new one
    char* old_data = data_;
    data_ = new_genome->data_;
    int new_nb_block = new_genome->nb_blocks_;
    new_genome->data_ = old_data;
    new_genome->nb_blocks_ = nb_blocks_;
    new_genome->length_ = length();

    release(std::move(new_genome));

  // Update length-related data
  length_ = new_length;
  nb_blocks_ = new_nb_block;
}

void Dna_7::replace(size_type pos, char* seq, size_type seq_length) {
// Invert the sequence between positions 'first' and 'last'
  // Check pos value
  assert(pos >= 0 && pos < length_);

  // If the sequence's length was not provided, compute it
  if (seq_length == -1) {
    seq_length = strlen(seq);
  }

  // Check that the sequence is contiguous
  assert(pos + seq_length <= length_);

  // Perform the replacement
  memcpy(&data_[pos], seq, seq_length * sizeof(char));
}


Mutation* Dna_7::do_switch(size_type pos
#ifdef BASE_4
, char choice
#endif
) {
#ifdef BASE_2
  // Perform the mutation
  if (data_[pos] == '0') data_[pos] = '1';
  else data_[pos] = '0';
#elif BASE_4
  // Perform the mutation
  char current_nucleotide = data_[pos];
  data_[pos] = choice; // generated in range [0; NB_BASE-1[
  // exclude current_nucleotide by shifting generated nucleotide by 1 when equal or above
  if (data_[pos] >= current_nucleotide) data_[pos] += 1;
#endif

  // Remove promoters containing the switched base
  #ifdef __EUKARYOTE
    ann_chrsm_->promoter_list().remove_promoters_around(pos, pos + 1);
  #else
    ann_chrsm_->promoter_list().remove_promoters_around(pos, Utils::mod(pos + 1, length_), length_);
  #endif

  // Look for potential new promoters containing the switched base
  if (length() >= PROM_SIZE)
    ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, pos, Utils::mod(pos + 1, length()));

  #ifdef BASE_2
    return new PointMutation(pos);
  #elif BASE_4
    return new PointMutation(pos, choice);
  #endif
}

Mutation* Dna_7::do_small_insertion(size_type pos, int16_t nb_insert, char* seq) {
  // Remove the promoters that will be broken
  ann_chrsm_->promoter_list().remove_promoters_around(pos, length_);

  insert(pos, seq, nb_insert);

  // Look for new promoters
  if (length_ >= PROM_SIZE) {
    if (length_ - nb_insert < PROM_SIZE) {
      // Special case where the genome was smaller than a promoter before the
      // insertion and greater than (or as big as) a promoter after the
      // insertion.
      // In that case, we must look for new promoters thoroughly on the whole
      // genome using locate_promoters
      ann_chrsm_->promoter_list().locate_promoters(*this);
    }
    else {
      ann_chrsm_->promoter_list().move_all_promoters_after(pos, nb_insert, length_);
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, pos, Utils::mod(pos + nb_insert, length_));
    }
  }

  return new SmallInsertion(pos, nb_insert, seq);
}

Mutation* Dna_7::do_small_deletion(size_type pos, int16_t nb_del) {
  auto old_pos = pos;

  // Remove promoters containing at least one nucleotide from the sequence to  delete
    ann_chrsm_->promoter_list().remove_promoters_around(pos, Utils::mod(pos + nb_del, length_), length_);

  // Do the deletion and update promoter list
  if (pos + nb_del <= length()) { // the deletion does not contain the origin of
    // replication
    // Do the deletion

    remove(pos, pos + nb_del);

    // Update promoter list
    if (length() >= PROM_SIZE) {
      ann_chrsm_->promoter_list().move_all_promoters_after(pos, -nb_del, length_);
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, Utils::mod(pos, length()));
    }
  }
  else { // the deletion contains the origin of replication
    // Do the deletion
    int32_t nb_del_at_pos_0 = nb_del - length() + pos;

    remove(pos, length_);
    remove(0, nb_del_at_pos_0);


    pos -= nb_del_at_pos_0;

    // Update promoter list
    if (length() >= PROM_SIZE) {
      ann_chrsm_->promoter_list().move_all_promoters_after(0, -nb_del_at_pos_0, length_);
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, 0);
    }
  }

  return new SmallDeletion(old_pos, nb_del);
}


Mutation* Dna_7::do_duplication(size_type pos_1, size_type pos_2, size_type pos_3) {
  // Duplicate segment [pos_1; pos_2[ and insert the duplicate before pos_3
  char* duplicate_segment = NULL;

    if (length_ == 1)
    {
        printf("*** genome of size 1 ; duplication not done *** \n");
        return nullptr;
    }

  size_type seg_length;

  if (pos_1 < pos_2) {
    //
    //       pos_1         pos_2                   -> 0-
    //         |             |                   -       -
    // 0--------------------------------->      -         -
    //         ===============                  -         - pos_1
    //           tmp (copy)                      -       -
    //                                             -----      |
    //                                             pos_2    <-'
    //
    seg_length = pos_2 - pos_1;

    posix_memalign((void **)&duplicate_segment,64,(seg_length+1)* sizeof(char));//new char[seg_length + 1];
    memcpy(duplicate_segment, &data_[pos_1], seg_length);
    duplicate_segment[seg_length] = '\0';

  }
  else { // if (pos_1 >= pos_2)
    // The segment to duplicate includes the origin of replication.
    // The copying process will be done in two steps.
    //
    //                                            ,->
    //    pos_2                 pos_1            |      -> 0-
    //      |                     |                   -       - pos_2
    // 0--------------------------------->     pos_1 -         -
    // ======                     =======            -         -
    //  tmp2                        tmp1              -       -
    //                                                  -----
    //
    //
    auto tmp1_len = length() - pos_1;
    auto tmp2_len = pos_2;
    seg_length = tmp1_len + tmp2_len;

    posix_memalign((void **)&duplicate_segment,64,(seg_length+1)* sizeof(char));//new char[seg_length + 1];
    memcpy(duplicate_segment, &data_[pos_1], tmp1_len);     // Copy tmp1
    memcpy(&duplicate_segment[tmp1_len], data_, tmp2_len);  // Copy tmp2
    duplicate_segment[seg_length] = '\0';
  }

  if (seg_length <= 0) { // TODO<dpa> should never happen right?
    free(duplicate_segment);
    return nullptr;
  }

  // Create a copy of the promoters beared by the segment to be duplicated
  // (they will be inserted in the individual's RNA list later)
  PromoterList duplicated_promoters = ann_chrsm_->promoter_list().duplicate_promoters_included_in(pos_1, pos_2, length_);

  ann_chrsm_->promoter_list().remove_promoters_around(pos_3, length_);

  insert(pos_3, duplicate_segment, seg_length);
  
  if (length() >= PROM_SIZE) {
    if (length() - seg_length < PROM_SIZE) {
      // Special case where the genome was smaller than a promoter before
      // the insertion and greater than (or as big as) a promoter after the
      // insertion.
      // In that case, we must look for new promoters thoroughly on the whole
      // genome using locate_promoters
        ann_chrsm_->promoter_list().locate_promoters(*this);
    }
    else {
      ann_chrsm_->promoter_list().move_all_promoters_after(pos_3, seg_length, length_);
      ann_chrsm_->promoter_list().insert_promoters_at(duplicated_promoters, pos_3, length_);
          
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this,pos_3);
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this,pos_3 + seg_length);
    }
  }

  free(duplicate_segment);

  return new Duplication(pos_1, pos_2, pos_3, seg_length);
}

Mutation* Dna_7::do_translocation(size_type pos_1, size_type pos_2, size_type pos_3, size_type pos_4, bool invert) {
  auto pos_min = std::min(pos_1, std::min(pos_2, std::min(pos_3, pos_4)));

  if (not invert) {
    if (pos_min == pos_1) {
      ABCDE_to_ADCBE(pos_1, pos_3, pos_2, pos_4);
    }
    else if (pos_min == pos_2) {
      ABCDE_to_ADCBE(pos_2, pos_4, pos_1, pos_3);
    }
    else if (pos_min == pos_3) {
      ABCDE_to_ADCBE(pos_3, pos_2, pos_4, pos_1);
    }
    else { // if (pos_min == pos_4)
      ABCDE_to_ADCBE(pos_4, pos_1, pos_3, pos_2);
    }
  }
  else { // invert
    if (pos_min == pos_1) {
      ABCDE_to_ADBpCpE(pos_1, pos_3, pos_2, pos_4);
    }
    else if (pos_min == pos_2) {
      ABCDE_to_ADBpCpE(pos_2, pos_4, pos_1, pos_3);
    }
    else if (pos_min == pos_3) {
      ABCDE_to_ACpDpBE(pos_3, pos_2, pos_4, pos_1);
    }
    else { // if (pos_min == pos_4)
      ABCDE_to_ACpDpBE(pos_4, pos_1, pos_3, pos_2);
    }
  }

  auto segment_length = pos_2 - pos_1;

  return new Translocation(pos_1, pos_2, pos_3, pos_4, segment_length, invert);
}

Mutation* Dna_7::do_inversion(size_type pos_1, size_type pos_2) {
// Invert segment going from pos_1 (included) to pos_2 (excluded)
// Exemple : sequence 011101001100 => 110011010001
  if (pos_1 == pos_2) return nullptr; // Invert everything <=> Invert nothing!
  //
  //       pos_1         pos_2                   -> 0-
  //         |             |                   -       -
  // 0--------------------------------->      -         -
  //         ===============                  -         - pos_1
  //           tmp (copy)                      -       -
  //                                             -----      |
  //                                             pos_2    <-'
  //

        if (length_ == 1)
        {
            printf("*** genome of size 1 ; inversion not done *** \n");
            return nullptr;
        }

  auto seg_length = pos_2 - pos_1;

  // Create the inverted sequence
  char* inverted_segment = NULL;
  posix_memalign((void **)&inverted_segment,64,(seg_length+1)* sizeof(char));//new char[seg_length + 1];

#ifdef BASE_2
#if defined(__INTEL_COMPILER) || defined(__INTEL_LLVM_COMPILER)
#pragma omp simd aligned(data_:64) aligned(inverted_segment:64)
#else
#pragma omp simd
#endif
#endif
  for (decltype(seg_length) i = 0; i < seg_length; i++) {
    int32_t j = pos_2 - 1 - i;
    inverted_segment[i] = get_complementary_base(data_[j]);
  }
  inverted_segment[seg_length] = '\0';

  // Remove promoters that included a breakpoint
  ann_chrsm_->promoter_list().remove_promoters_around(pos_1, length_);
  ann_chrsm_->promoter_list().remove_promoters_around(pos_2, length_);

  // Invert the sequence
  replace(pos_1, inverted_segment, seg_length);

  // Update promoter list
  if (length() >= PROM_SIZE) {
    ann_chrsm_->promoter_list().invert_promoters_included_in(pos_1, pos_2);
    ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, pos_1);
    ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, pos_2);
  }

  free(inverted_segment);

  return new Inversion(pos_1, pos_2, seg_length);
}


Mutation* Dna_7::do_deletion(size_type pos_1, size_type pos_2) {
  // Delete segment going from pos_1 (included) to pos_2 (excluded)
    if (length_ == 1)
    {
        printf("*** genome of size 1 ; deletion not done *** \n");
        return nullptr;
    }

  if (pos_1 < pos_2) {
    //
    //       pos_1         pos_2                   -> 0-
    //         |             |                   -       -
    // 0--------------------------------->      -         -
    //         ===============                  -         - pos_1
    //           tmp (copy)                      -       -
    //                                             -----      |
    //                                             pos_2    <-'
    //

    auto segment_length = pos_2 - pos_1;

    // Remove promoters containing at least one nucleotide from the sequence
    // to delete
    ann_chrsm_->promoter_list().remove_promoters_around(pos_1, pos_2, length_);
    // Delete the sequence between pos_1 and pos_2
    remove(pos_1, pos_2);

    // Update promoter list
    if (length() >= PROM_SIZE) {
      ann_chrsm_->promoter_list().move_all_promoters_after(pos_1, -segment_length, length_);
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, pos_1);
    }


  }
  else { // if (pos_1 >= pos_2)
    // The segment to delete includes the origin of replication.
    // The deletion process will be done in two steps.
    //
    //                                            ,->
    //    pos_2                 pos_1            |      -> 0-
    //      |                     |                   -       - pos_2
    // 0--------------------------------->     pos_1 -         -
    // =====                      =======            -         -
    //  tmp2                        tmp1              -       -
    //                                                  -----
    //
    //

    // Remove promoters containing at least one nucleotide from the sequence to delete
    ann_chrsm_->promoter_list().remove_promoters_around(pos_1, pos_2, length_);

    // Delete the sequence between pos_1 and pos_2
    remove(pos_1, length_); // delete tmp1 from genome
    remove(0, pos_2);       // delete tmp2 from genome
    // Update promoter list
    if (length() >= PROM_SIZE) {
      ann_chrsm_->promoter_list().move_all_promoters_after(0, -pos_2, length_);
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, 0);
    }
  }

  return new Deletion(pos_1, pos_2, Utils::mod(pos_2 - pos_1 - 1, length()) + 1);
}

void Dna_7::undergo_mutation(const Mutation& mut) {
  switch (mut.mut_type()) {
    case SWITCH : {
    #ifdef BASE_2
      do_switch(dynamic_cast<const PointMutation&>(mut).pos());
    #elif BASE_4
      const auto& pmut = dynamic_cast<const PointMutation&>(mut);
      // printf("%d -- DO Switch %d : %c\n",AeTime::time(),pmut.pos(), pmut.base());
      do_switch(pmut.pos(), pmut.base());
    #endif
      break;
    }
    case S_INS : {
      const auto& s_ins = dynamic_cast<const SmallInsertion&>(mut);
      // printf("%d -- DO Insert %d Nb %d\n",AeTime::time(),s_ins.length(), s_ins.seq());
      do_small_insertion(s_ins.pos(), s_ins.length(), s_ins.seq());
      break;
    }
    case S_DEL : {
      const auto& s_del = dynamic_cast<const SmallDeletion&>(mut);
      // printf("%d -- DO Delete %d Nb %d\n",AeTime::time(),s_del.pos(), s_del.length());
      do_small_deletion(s_del.pos(), s_del.length());
      break;
    }
    case DUPL : {
      const auto& dupl = dynamic_cast<const Duplication&>(mut);
      // printf("%d -- DO Duplication %d %d %d\n",AeTime::time(),dupl.pos1(), dupl.pos2(), dupl.pos3());
      do_duplication(dupl.pos1(), dupl.pos2(), dupl.pos3());
      break;
    }
    case DEL : {
      const auto& del = dynamic_cast<const Deletion&>(mut);
      // printf("%d -- DO Deletion %d %d\n",AeTime::time(),del.pos1(), del.pos2());

      do_deletion(del.pos1(), del.pos2());
      break;
    }
    case TRANS : {
      const auto& trans = dynamic_cast<const Translocation&>(mut);
      // printf("%d -- DO Translocation %d %d %d %d (%d)\n",AeTime::time(),trans.pos1(), trans.pos2(), trans.pos3(), trans.pos4(),
      //                  trans.invert());
      do_translocation(trans.pos1(), trans.pos2(), trans.pos3(), trans.pos4(),
                       trans.invert());
      break;
    }
    case INV : {
      const auto& inv = dynamic_cast<const Inversion&>(mut);
      // printf("%d -- DO Inversion %d %d\n",AeTime::time(),inv.pos1(), inv.pos2());
      do_inversion(inv.pos1(), inv.pos2());
      break;
    }
    default :
      fprintf(stderr, "ERROR, invalid mutation type in file %s:%d\n",
              __FILE__, __LINE__);
      exit(EXIT_FAILURE);
      break;
  }
}

std::list<Mutation*> Dna_7::apply_mutations(const std::list<MutationEvent*>& mutations_to_apply) {
  std::list<Mutation*> undergone_mutations;

  for (const auto& mut_to_apply: mutations_to_apply) {
    if (mut_to_apply != nullptr) {
      Mutation* undergone_mut = nullptr;
      switch (mut_to_apply->type()) {
        case DO_SWITCH: {
          #ifdef BASE_2
            undergone_mut = do_switch(mut_to_apply->pos_1());
          #elif BASE_4
            undergone_mut = do_switch(mut_to_apply->pos_1(), mut_to_apply->base());
          #endif
          if (undergone_mut) {
            nb_swi_++;
            nb_mut_++;
          }
          break;
        }
        case SMALL_INSERTION: {
          undergone_mut = do_small_insertion(mut_to_apply->pos_1(), mut_to_apply->number(), mut_to_apply->seq());
          if (undergone_mut) {
            nb_indels_++;
            nb_mut_++;
          }
          break;
        }
        case SMALL_DELETION: {
          undergone_mut = do_small_deletion(mut_to_apply->pos_1(), mut_to_apply->number());
          if (undergone_mut) {
            nb_indels_++;
            nb_mut_++;
          }
          break;
        }
        case DUPLICATION: {
          undergone_mut = do_duplication(mut_to_apply->pos_1(), mut_to_apply->pos_2(), mut_to_apply->pos_3());
          if (undergone_mut) {
            nb_large_dupl_++;
            nb_rear_++;
          }
          break;
        }
        case TRANSLOCATION: {
          undergone_mut = do_translocation(mut_to_apply->pos_1(),
                                           mut_to_apply->pos_2(),
                                           mut_to_apply->pos_3(),
                                           mut_to_apply->pos_4(),
                                           mut_to_apply->invert());
          if (undergone_mut) {
            nb_large_trans_++;
            nb_rear_++;
          }
          break;
        }
        case INVERSION: {
          undergone_mut = do_inversion(mut_to_apply->pos_1(), mut_to_apply->pos_2());
          if (undergone_mut) {
            nb_large_inv_++;
            nb_rear_++;
          }
          break;
        }
        case DELETION: {
          undergone_mut = do_deletion(mut_to_apply->pos_1(), mut_to_apply->pos_2());
          if (undergone_mut) {
            nb_large_del_++;
            nb_rear_++;
          }
          break;
        }
      }
      if (undergone_mut) {
        undergone_mutations.push_back(undergone_mut);
      }
    }
  }

  return undergone_mutations;
}


void Dna_7::ABCDE_to_ADCBE(size_type pos_B, size_type pos_C, size_type pos_D, size_type pos_E) {
  // Rearrange the sequence from ABCDE to ADCBE (complex translocation
  // of segment defined between positions pos_B and pos_D)
  //
  // Segments are identified by pos_x values as shown below.
  //
  // TODO(dpa) CHECK THIS !!!
  // WARNING : Segment C includes nucleotide at pos_D // NOTE : WTF???
  //
  //         A      B        C       D       E
  //      |----->=======[>=======>-------[>-------|        =>
  //          pos_B   pos_C    pos_D   pos_E
  //
  //                         |
  //                         V
  //
  //         A      D        C       B        E
  //      |----->-------[>=======>=======[>-------|
  // Compute segment lengths
  auto len_A = pos_B;
  auto len_B = pos_C - pos_B;
  auto len_C = pos_D - pos_C;
  auto len_D = pos_E - pos_D;
  auto len_E = length() - pos_E;
  auto len_AD = len_A + len_D;
  auto len_ADC = len_AD + len_C;
  auto len_ADCB = len_ADC + len_B;

  // Create new sequence
  Dna_7* new_genome = Dna_7::make_dna(length_);

    new_genome->set_chrsm(ann_chrsm_);
    new_genome->set_indiv(length_,parent_length_,indiv_);

  memcpy(new_genome->data_, data_, len_A * sizeof(char));
  memcpy(&(new_genome->data_[len_A]), &data_[pos_D], len_D * sizeof(char));
  memcpy(&(new_genome->data_[len_AD]), &data_[pos_C], len_C * sizeof(char));
  memcpy(&(new_genome->data_[len_ADC]), &data_[pos_B], len_B * sizeof(char));
  memcpy(&(new_genome->data_[len_ADCB]), &data_[pos_E], len_E * sizeof(char));
  new_genome->data_[length_] = '\0';

  // Replace sequence
  // NB : The size of the genome doesn't change. Therefore, we don't nee
  // to update length_ and nb_blocks_
    char* old_data = data_;
    data_ = new_genome->data_;
    int new_nb_block = new_genome->nb_blocks_;
    new_genome->nb_blocks_ = nb_blocks_;

    new_genome->data_ = old_data;
    new_genome->length_ = length();

    nb_blocks_ = new_nb_block;
    release(std::move(new_genome));


  // ========== Update promoter list ==========
  if (length() >= PROM_SIZE) {
    // Remove promoters that include a breakpoint
    ann_chrsm_->promoter_list().remove_promoters_around(pos_B, length_);
    ann_chrsm_->promoter_list().remove_promoters_around(pos_C, length_);
    ann_chrsm_->promoter_list().remove_promoters_around(pos_D, length_);
    ann_chrsm_->promoter_list().remove_promoters_around(pos_E, length_);

    // Create temporary lists for promoters to move and/or invert
    PromoterList promoters_B;
    PromoterList promoters_C;
    PromoterList promoters_D;

    // Extract promoters that are totally included in each segment to be moved
    // and shift them to their new positions
    if (len_B >= PROM_SIZE) {
      promoters_B = ann_chrsm_->promoter_list().extract_promoters_included_in(pos_B, pos_C);
      promoters_B.shift_promoters(len_D + len_C, length());
    }
    if (len_C >= PROM_SIZE) {
      promoters_C = ann_chrsm_->promoter_list().extract_promoters_included_in(pos_C, pos_D);
      promoters_C.shift_promoters(len_D - len_B, length());
    }
    if (len_D >= PROM_SIZE) {
      promoters_D = ann_chrsm_->promoter_list().extract_promoters_included_in(pos_D, pos_E);
      promoters_D.shift_promoters(-len_B - len_C, length());
    }

    // Reinsert the shifted promoters
    ann_chrsm_->promoter_list().insert_promoters(promoters_B);
    ann_chrsm_->promoter_list().insert_promoters(promoters_C);
    ann_chrsm_->promoter_list().insert_promoters(promoters_D);

    // 5) Look for new promoters including a breakpoint
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, len_A);
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, len_AD);
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, len_ADC);
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, len_ADCB);
  }
}

void Dna_7::ABCDE_to_ADBpCpE(size_type pos_B, size_type pos_C, size_type pos_D, size_type pos_E) {
  // Rearrange the sequence from ABCDE to ADBpCpE (complex translocation
  // with inversion of segment defined between positions pos_B and pos_D)
  // Bp (resp Cp) stands for inverted B (resp C)
  //
  // Segments are identified by pos_x values as shown below.
  //
  // TODO(dpa) CHECK THIS !!!
  // WARNING : Segment C includes nucleotide at pos_D // NOTE : WTF???
  //
  //         A      B        C       D        E
  //      |----->=======[>=======>-------<]-------|
  //          pos_B   pos_C    pos_D   pos_E
  //
  //                         |
  //                         V
  //
  //         A      D        Bp      Cp       E
  //      |----->-------<]=======<=======<]-------|
  // Compute segment lengths

  auto len_A = pos_B;
  auto len_B = pos_C - pos_B;
  auto len_C = pos_D - pos_C;
  auto len_D = pos_E - pos_D;
  auto len_E = length() - pos_E;
  auto len_AD = len_A + len_D;
  auto len_ADB = len_AD + len_B;
  auto len_ADBC = len_ADB + len_C;

  // Create new sequence
  Dna_7* new_genome = Dna_7::make_dna(length_);
    new_genome->set_chrsm(ann_chrsm_);
    new_genome->set_indiv(length_,parent_length_,indiv_);

  // Copy segments A and D
  memcpy(new_genome->data_, data_, len_A * sizeof(char));
  memcpy(&(new_genome->data_[len_A]), &data_[pos_D], len_D * sizeof(char));


  // Build Bp and put it in the new genome
  char* inverted_segment;
  posix_memalign((void **)&inverted_segment,64,(len_B+1)* sizeof(char));//new char[len_B + 1];

  for (size_type i = 0, j = pos_C - 1; i < len_B; i++, j--) {
    inverted_segment[i] = get_complementary_base(data_[j]);
  }
  inverted_segment[len_B] = '\0';

  memcpy(&(new_genome->data_[len_AD]), inverted_segment, len_B * sizeof(char));

  free(inverted_segment);


  // Build Cp and put it in the new genome
  posix_memalign((void **)&inverted_segment,64,(len_C+1)* sizeof(char));//new char[len_C + 1];

  for (size_type i = 0, j = pos_D - 1; i < len_C; i++, j--) {
    inverted_segment[i] = get_complementary_base(data_[j]);
  }
  inverted_segment[len_C] = '\0';

  memcpy(&(new_genome->data_[len_ADB]), inverted_segment, len_C * sizeof(char));

  free(inverted_segment);

  // Copy segment E into the new genome
  memcpy(&(new_genome->data_[len_ADBC]), &data_[pos_E], len_E * sizeof(char));
  new_genome->data_[length_] = '\0';


  // Replace sequence
    char* old_data = data_;
    data_ = new_genome->data_;
    int new_nb_block = new_genome->nb_blocks_;
    new_genome->nb_blocks_ = nb_blocks_;

    new_genome->data_ = old_data;
    new_genome->length_ = length();

    nb_blocks_ = new_nb_block;

    release(std::move(new_genome));


  // ========== Update promoter list ==========
  if (length() >= PROM_SIZE) {
    // Remove promoters that include a breakpoint
    ann_chrsm_->promoter_list().remove_promoters_around(pos_B, length_);
    ann_chrsm_->promoter_list().remove_promoters_around(pos_C, length_);
    ann_chrsm_->promoter_list().remove_promoters_around(pos_D, length_);
    ann_chrsm_->promoter_list().remove_promoters_around(pos_E, length_);

    // Create temporary lists for promoters to move and/or invert
    PromoterList promoters_B;
    PromoterList promoters_C;
    PromoterList promoters_D;

    // 2) Extract promoters that are totally included in each segment to be
    //    moved (B, C and D)
    if (len_B >= PROM_SIZE) {
      promoters_B = ann_chrsm_->promoter_list().extract_promoters_included_in(pos_B, pos_C);
    }
    if (len_C >= PROM_SIZE) {
      promoters_C = ann_chrsm_->promoter_list().extract_promoters_included_in(pos_C, pos_D);
    }
    if (len_D >= PROM_SIZE) {
      promoters_D = ann_chrsm_->promoter_list().extract_promoters_included_in(pos_D, pos_E);
    }
    // 3a) Invert promoters of segments B and C
    promoters_B.invert_promoters(pos_B, pos_C);
    promoters_C.invert_promoters(pos_C, pos_D);

    // 3b) Shift these promoters positions
    promoters_B.shift_promoters(len_D, length());
    promoters_C.shift_promoters(len_D, length());
    promoters_D.shift_promoters(-len_B - len_C, length());

    // 4) Reinsert the shifted promoters
    ann_chrsm_->promoter_list().insert_promoters(promoters_C);
    ann_chrsm_->promoter_list().insert_promoters(promoters_B);
    ann_chrsm_->promoter_list().insert_promoters(promoters_D);

    // 5) Look for new promoters including a breakpoint
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, len_A);
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, len_AD);
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, len_ADB);
      ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, len_ADBC);
  }
}

void Dna_7::ABCDE_to_ACpDpBE(size_type pos_B, size_type pos_C, size_type pos_D, size_type pos_E) {
  // Rearrange the sequence from ABCDE to ACpDpBE (complex translocation with
  // inversion of segment defined between positions pos_C and pos_E)
  // Cp (resp Dp) stands for inverted C (resp D)
  //
  // Segments are identified by pos_x values as shown below.
  //
  // TODO(dpa) CHECK THIS !!!
  // WARNING : Segment D includes nucleotide at pos_E // NOTE : WTF???
  //
  //         A      B        C       D       E
  //      |----<]-------->=======[>=======>-------|
  //          pos_B    pos_C    pos_D   pos_E
  //
  //                         |
  //                         V
  //
  //          A       C'      D'       B       E
  //       |-----<]=======>=======<]------->-------|

  // Compute segment lengths
  auto len_A = pos_B;
  auto len_B = pos_C - pos_B;
  auto len_C = pos_D - pos_C;
  auto len_D = pos_E - pos_D;
  auto len_E = length() - pos_E;
  auto len_AC = len_A + len_C;
  auto len_ACD = len_AC + len_D;
  auto len_ACDB = len_ACD + len_B;

  // Create new sequence
  Dna_7* new_genome = Dna_7::make_dna(length_);
    new_genome->set_chrsm(ann_chrsm_);
    new_genome->set_indiv(length_,parent_length_,indiv_);

  // Copy segment A
  memcpy(new_genome->data_, data_, len_A * sizeof(char));


  // Build Cp and put it in the new genome
  char* inverted_segment;
  posix_memalign((void **)&inverted_segment,64,(len_C+1)* sizeof(char));//new char[len_C + 1];

  for (size_type i = 0, j = pos_D - 1; i < len_C; i++, j--) {
    inverted_segment[i] = get_complementary_base(data_[j]);
  }
  inverted_segment[len_C] = '\0';

  memcpy(&(new_genome->data_[len_A]), inverted_segment, len_C * sizeof(char));

  free(inverted_segment);


  // Build Dp and put it in the new genome
  posix_memalign((void **)&inverted_segment,64,(len_D+1)* sizeof(char));//new char[len_D + 1];

  for (size_type i = 0, j = pos_E - 1; i < len_D; i++, j--) {
    inverted_segment[i] = get_complementary_base(data_[j]);
  }
  inverted_segment[len_D] = '\0';

  memcpy(&(new_genome->data_[len_AC]), inverted_segment, len_D * sizeof(char));

  free(inverted_segment);

  // Copy segments B and E
  memcpy(&(new_genome->data_[len_ACD]), &data_[pos_B], len_B * sizeof(char));
  memcpy(&(new_genome->data_[len_ACDB]), &data_[pos_E], len_E * sizeof(char));
    new_genome->data_[length_] = '\0';


  // Replace sequence
    char* old_data = data_;
    data_ = new_genome->data_;

    int new_nb_block = new_genome->nb_blocks_;
    new_genome->nb_blocks_ = nb_blocks_;

    new_genome->data_ = old_data;
    new_genome->length_ = length();

    nb_blocks_ = new_nb_block;
    release(std::move(new_genome));


  // ========== Update promoter list ==========
  // 1) Remove promoters that include a breakpoint
  // 2) Extract promoters that are totally included in each segment to be
  //    moved (B, C and D)
  // 3) Shift (and invert when needed) these promoters positions
  // 4) Reinsert the shifted promoters
  // 5) Look for new promoters including a breakpoint
  if (length() >= PROM_SIZE) {
    // 1) Remove promoters that include a breakpoint
    ann_chrsm_->promoter_list().remove_promoters_around(pos_B, length_);
    ann_chrsm_->promoter_list().remove_promoters_around(pos_C, length_);
    ann_chrsm_->promoter_list().remove_promoters_around(pos_D, length_);
    ann_chrsm_->promoter_list().remove_promoters_around(pos_E, length_);
    // Create temporary lists for promoters to move and/or invert
    PromoterList promoters_B;
    PromoterList promoters_C;
    PromoterList promoters_D;

    // 2) Extract promoters that are totally included in each segment to be
    //    moved (B, C and D)
    if (len_B >= PROM_SIZE) {
      promoters_B = ann_chrsm_->promoter_list().extract_promoters_included_in(pos_B, pos_C);
    }
    if (len_C >= PROM_SIZE) {
      promoters_C = ann_chrsm_->promoter_list().extract_promoters_included_in(pos_C, pos_D);
    }
    if (len_D >= PROM_SIZE) {
      promoters_D = ann_chrsm_->promoter_list().extract_promoters_included_in(pos_D, pos_E);
    }

    // 3a) Invert promoters of segments C and D
    promoters_C.invert_promoters(pos_C, pos_D);
    promoters_D.invert_promoters(pos_D, pos_E);

    // 3b) Shift these promoters positions
    promoters_B.shift_promoters(len_C + len_D, length());
    promoters_C.shift_promoters(-len_B, length());
    promoters_D.shift_promoters(-len_B, length());

    // 4) Reinsert the shifted promoters
    ann_chrsm_->promoter_list().insert_promoters(promoters_B);
    ann_chrsm_->promoter_list().insert_promoters(promoters_D);
    ann_chrsm_->promoter_list().insert_promoters(promoters_C);

    // 5) Look for new promoters including a breakpoint
    ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, len_A);
    ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, len_AC);
    ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, len_ACD);
    ann_chrsm_->promoter_list().look_for_new_promoters_around(*this, len_ACDB);
  }
}
void Dna_7::reset_stat() {
  nb_swi_ = 0;
  nb_indels_= 0;
  nb_mut_= 0;
  nb_large_dupl_= 0;
  nb_large_del_= 0;
  nb_large_trans_= 0;
  nb_large_inv_= 0;
  nb_rear_= 0;
}

/***
 * returns the subsequence of size count beginning at position first (included) on strand strand
 *
 * \param first the position of the first base (included) of the subsequence of interest
 * \param count size of the requested subsequence
 * \param strand strand of interest (note that LAGGING reads "backwards")
 *
 * \return a copy of the subsequence of interest
 */
std::string Dna_7::subseq(size_type first, size_type count, Strand strand) const {
  first = Utils::mod(first, length_);

  if (strand == Strand::LEADING) {
    bool spans_OriC = first + count > length_;

    if (not spans_OriC) {
      return std::string(&(data_[first]), count);
    } else {
      return std::string(&(data_[first])) + std::string(data_, count - (length_ - first));
    }
  } else { // Strand::LAGGING
    auto last = Utils::mod(first + 1 - count, length_);
    bool spans_OriC =  last >= first;

    std::string subseq;
    subseq.reserve(count);

    if (not spans_OriC) {
      std::transform(std::make_reverse_iterator(&(data_[first + 1])),
                     std::make_reverse_iterator(&(data_[last])),
                     std::back_inserter(subseq),
                     get_complementary_base);
      return subseq;
    } else { // spans_OriC
      std::transform(std::make_reverse_iterator(&(data_[first + 1])),
                     std::make_reverse_iterator(&(data_[0])),
                     std::back_inserter(subseq),
                     get_complementary_base);
      auto remains_to_transform = count - subseq.size();
      std::transform(std::make_reverse_iterator(&(data_[length_])),
                     std::make_reverse_iterator(&(data_[length_ - remains_to_transform])),
                     std::back_inserter(subseq),
                     get_complementary_base);
      return subseq;
    }
  }
}

void Dna_7::reset_dna_factory(DnaFactory*&& new_factory) {
  dna_factory_.reset(new_factory);
}

void Dna_7::reduce_dna_factory_memory_usage(const ExpManager_7* exp_m) {
  dna_factory_->reduce_space(exp_m);
}

}
