// ****************************************************************************
//
//          Aevol - An in silico experimental evolution platform
//
// ****************************************************************************
//
// Copyright: See the AUTHORS file provided with the package or <www.aevol.fr>
// Web: http://www.aevol.fr/
// E-mail: See <http://www.aevol.fr/contact/>
// Original Authors : Guillaume Beslon, Carole Knibbe, David Parsons
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ****************************************************************************

#ifndef AEVOL_DNA_7_H
#define AEVOL_DNA_7_H


#include <cstdint>
#include <list>
#include <memory>

#include "Utils.h"
#include "MutationEvent.h"
#include "Mutation.h"
#include "Strand.h"

#ifdef BASE_4
#include "macros.h"
#endif

namespace aevol {

#define BLOCK_SIZE INT32_C(1024)
class DnaFactory;
class ExpManager_7;
class Individual_7;
class AnnotatedChromosome;

class Dna_7 final {
  friend class DnaFactory;

 public:
  using size_type = int32_t;

 protected:
  Dna_7() = delete;
  Dna_7(const Dna_7&) = delete;
  Dna_7(Dna_7&&) = delete;
  Dna_7& operator=(const Dna_7&) = delete;
  Dna_7& operator=(Dna_7&&) = delete;

  Dna_7(size_type length);
  ~Dna_7();

 public:
  static Dna_7* make_dna(size_t request_size);
  static Dna_7* make_from_sequence(const char* seq, size_type len);
  static void release(Dna_7*&& dna);

  void reserve(size_type new_capacity);
  void reset_sequence(const char* seq, size_type len);

    void set_indiv(Dna_7* dna, Individual_7* indiv);
    void set_indiv(Individual_7* indiv);
    void set_indiv(size_type length, size_type parent_length, Individual_7* indiv);
    void set_indiv(const char* dna, size_type dna_length, Individual_7 * indiv);
    void set_chrsm(AnnotatedChromosome* ann_chrsm);
    
    inline char get_lead(size_type pos) { return data_[Utils::mod(pos,length_ )]; }
    inline char get_lag(size_type pos) {
      #ifdef BASE_2
      return data_[Utils::mod(pos,length_)]; 
      #elif BASE_4
      return get_complementary_base(data_[Utils::mod(pos,length_)]);
      #endif
      }

    inline int nb_block() { return nb_blocks_; }

    void undergo_mutation(const Mutation& mut);

    std::list<Mutation*> apply_mutations(const std::list<MutationEvent*>& mutations_to_apply);

    Mutation* do_switch(size_type pos
    #ifdef BASE_4
    , char new_base
    #endif
    );

    Mutation* do_small_insertion(size_type pos, int16_t nb_insert, char* seq);
    Mutation* do_small_deletion(size_type pos, int16_t nb_del);
    Mutation* do_duplication(size_type pos_1, size_type pos_2, size_type pos_3);
    Mutation* do_inversion(size_type pos_1, size_type pos_2);
    Mutation* do_translocation(size_type pos_1, size_type pos_2, size_type pos_3, size_type pos_4, bool invert);
    Mutation* do_deletion(size_type pos_1, size_type pos_2);

    void remove(size_type first, size_type last);
    void insert(size_type pos, const char* seq, size_type seq_length);
    void replace(size_type pos, char* seq, size_type seq_length);

    void reset_stat();

    const char* data() const {return data_;}
    std::string subseq(size_type first, size_type count, Strand strand) const;
    auto length() const {
      return length_;
    }

    char* to_char() {
      return data_;
    }

    auto parent_length() const { return parent_length_; }
    void set_parent_length(size_type parent_length) { parent_length_ = parent_length; }

    static int32_t nb_required_blocks(size_type length);

    void ABCDE_to_ADCBE(size_type pos_B, size_type pos_C, size_type pos_D, size_type pos_E);
    void ABCDE_to_ADBpCpE(size_type pos_B, size_type pos_C, size_type pos_D, size_type pos_E);
    void ABCDE_to_ACpDpBE(size_type pos_B, size_type pos_C, size_type pos_D, size_type pos_E);

    static void reset_dna_factory(DnaFactory*&& new_factory);
    static void reduce_dna_factory_memory_usage(const ExpManager_7* exp_m);

    std::list<MutationEvent*> mutation_list;

    char* data_;
    size_type length_;

    // Stats
    int32_t nb_swi_ = 0;
    int32_t nb_indels_= 0;
    int32_t nb_mut_= 0;

    int32_t nb_large_dupl_= 0;
    int32_t nb_large_del_= 0;
    int32_t nb_large_trans_= 0;
    int32_t nb_large_inv_= 0;
    int32_t nb_rear_= 0;

    Individual_7* indiv_ = nullptr;
    AnnotatedChromosome* ann_chrsm_ = nullptr;

 private:
  static std::unique_ptr<DnaFactory> dna_factory_;

  size_type parent_length_;
  int32_t nb_blocks_;
};

inline void Dna_7::set_indiv(Individual_7* indiv) {
    indiv_ = indiv;
}

inline int32_t Dna_7::nb_required_blocks(size_type length) {
  return length/BLOCK_SIZE + 1;
}
}


#endif
